#include "ImageLoader.h"

#include "ImageImportTga.h"

#ifdef __BCPLUSPLUS__

#else
	#define try 		{
	#define __finally 	}
	#define catch(x) 	}
#endif

//********************************************************
// Externals
//********************************************************

TImageLoader ImageLoader;

//********************************************************
// TTextureFrame
//********************************************************

TTextureFrame::TTextureFrame  ( )
{
        texture    = -1;
        image      = NULL;
}

TTextureFrame::~TTextureFrame ( )
{
        if (texture!=-1) glDeleteTextures(1, &texture);
        if (image) free(image);
        texture = -1;
        image   = NULL;
}

unsigned char * TTextureFrame::GetImage   ( )
{
        return image;
}

GLuint          TTextureFrame::GetTexture ( )
{
        return texture;
}

void TTextureFrame::SetInfo    ( unsigned char * img, GLuint tex )
{
	image = img;
        texture = tex;
}

//********************************************************
// TTexture
//********************************************************

TTexture::TTexture  ( )
{
        filename     = "";
        numoflinks   = 1;
        clone        = false;
        currenttime  = 0;
        currentframe = 0;
        width        = 1;
        height       = 1;
        numofbytes   = 0;
        duration     = 50;
        animloopmode = animloopJumpStartAtEnd;
}

TTexture::~TTexture ( )
{
        ClearAll ( );
}

void TTexture::ClearAll     ( )
{
        for ( unsigned int i = 0; i<frames.size(); i++ )
                delete frames[i];
        frames.clear();
        ResetAnimation  ( );
}

TTextureFrame  * TTexture::AddFrame          ( )
{
	TTextureFrame * tex    = new TTextureFrame();
        endframe = frames.size()-1;
        frames.push_back ( tex );
        return tex;
}

unsigned int TTexture::GetWidth      ( )
{
        return width;
}

unsigned int TTexture::GetHeight     ( )
{
        return height;
}

unsigned int TTexture::GetNumOfBytes ( )
{
        return numofbytes;
}

unsigned int TTexture::GetNumOfFrames  ( )
{
        return frames.size();
}

TTextureFrame * TTexture::GetFrame        ( unsigned int index )
{
        if ( index>=frames.size() ) return NULL;
        return frames[index];
}

unsigned int TTexture::GetCurrentFrameNo ( )
{
        return currentframe;
}

void TTexture::SetCurrentFrameNo ( unsigned int index )
{
        if (index>=frames.size()) return;
        currentframe = index;
        startframe      = index;
        endframe        = index;
        ResetAnimation  ( );
}

unsigned int  TTexture::GetStartFrame     ( )
{
        return startframe;
}

unsigned int  TTexture::GetEndFrame       ( )
{
        return endframe;
}

void TTexture::SetAnimationRange ( unsigned int startf, unsigned int endf )
{
        if (startf>=frames.size()) return;
        if (endf>=frames.size()) return;
        if (startf>endf) return;
        startframe   = startf;
        endframe     = endf;
        currentframe = startframe;
        ResetAnimation  ( );
}

unsigned int     TTexture::GetDuration       ( )
{
        return duration;
}

TAnimLoopMode    TTexture::GetAnimLoopMode   ( )
{
        return animloopmode;
}

void TTexture::ResetAnimation  ( )
{
        currenttime  = 0;
}

void TTexture::SetDuration       ( unsigned int  fduration )
{
        duration = fduration;
}

void TTexture::SetAnimLoopMode   ( TAnimLoopMode lmode )
{
        animloopmode = lmode;
}

void TTexture::AdvanceAnimation( )
{
        if (frames.size()==0) return;
        unsigned int ctime = mGetTickCount();
        if (currenttime == 0)
        {
                currentframe = startframe;
                currenttime  = ctime;
        }
        else
        {
                unsigned int delta = ctime - currenttime;
                while ( delta >= duration )
                {
                        delta       -= duration;
                        currenttime += duration;
                        currentframe++;
                        if ( animloopmode==animloopJumpStartAtEnd &&
                             currentframe>endframe )
                                currentframe = startframe;
                                else
                        if ( animloopmode==animloopStopAtEnd &&
                             currentframe>endframe )
                                currentframe = endframe;
                }
        }
}

TTextureFrame * TTexture::BindTexture   ( bool advanceanim )
{
        if (advanceanim) AdvanceAnimation( );
        TTextureFrame * tex = GetCurrentFrame ( );
        if (tex && tex->texture!=-1)
        {
                glBindTexture  ( GL_TEXTURE_2D, tex->texture );
                return tex;
        }
        else
        return NULL;
}

TTextureFrame * TTexture::GetCurrentFrame ( )
{
        if (frames.size()==0) return NULL;
        return frames[currentframe];
}

bool TTexture::LoadFromFile ( mstring fname, bool allocateimage, bool allocatetexture, bool append )
{
        if (!append)
                ClearAll ( );

        if (!mIsFileExist(fname) )
                return false;

        mstring last3 = mUpperCase( fname.substr( fname.length()-3,3 ) );
        mstring last4 = mUpperCase( fname.substr( fname.length()-4,4 ) );

        if (last3=="M2B")
        {
                TGameFileStream stream;
                TGameMemoryStream memstream;
                if ( stream.OpenFileForRead( fname ) )
                {
                        memstream.SetSize( stream.GetSize() );
                        stream.Read(memstream.GetMemory(), stream.GetSize() );
                        return LoadFromStream ( &memstream, allocateimage, allocatetexture, append );
                }
                else
                return false;
        }
        else
        if (last3=="TGA")
                return ImportTgaFile ( fname, this, allocateimage, allocatetexture, append );

        return false;
}

bool TTexture::SaveToStream ( TGameIOStream * stream )
{
        unsigned int signature = 0x30a925fc;
        int version   = 1;

        stream->Write ( &signature );
        stream->Write ( &version   );

        int i;
        for ( i = 0; i<frames.size(); i++ )
                if ( frames[i]->GetImage ( ) == NULL ) return false;

        stream->Write( &width );
        stream->Write( &height );
        stream->Write( &numofbytes );
        stream->Write( &duration );
        stream->Write( (unsigned char*)(&animloopmode) );

        int frmsize = frames.size();
        stream->Write( &frmsize );

        unsigned char * buff;
        for ( i = 0; i<frames.size(); i++ )
        {
                buff = frames[i]->GetImage ( );
                stream->Write( buff, width*height*numofbytes );
        }
        return true;
}

bool TTexture::LoadFromStream ( TGameIOStream * stream, bool allocateimage, bool allocatetexture, bool append )
{
        unsigned int signature = 0;
        int     version   = 0;

        stream->Read ( &signature );
        stream->Read ( &version   );

        if ( signature != 0x30a925fc || version > 1 )
                return false;

        unsigned int    pwidth;
        unsigned int    pheight;
        unsigned int    pnumofbytes;

        stream->Read( &pwidth );
        stream->Read( &pheight );
        stream->Read( &pnumofbytes );

        if (append && (GetWidth()!=pwidth ||
                        GetHeight()!=pheight ||
                        GetNumOfBytes()!=pnumofbytes))
                return false;

        width = pwidth;
        height = pheight;
        numofbytes = pnumofbytes;

        stream->Read( &duration );
        stream->Read( (unsigned char*)(&animloopmode) );

        int frmsize;
        stream->Read( &frmsize );

        startframe = 0;
        endframe   = 0;

        unsigned char * buff;
        for ( int i = 0; i<frmsize; i++ )
        {
                TTextureFrame  * frm = AddFrame  ( );
                if ( frm )
                {
                        unsigned char * img = (unsigned char *)malloc (width*height*numofbytes);
                        if ( img )
                        {
                                if ( stream->Read( img, width*height*numofbytes ) == width*height*numofbytes )
                                {
                                        GLuint tex = -1;
                                        if (allocatetexture)
                                        {
                                                glGenTextures     ( 1, &tex );
                                                if ( tex != -1 )
                                                {
                                                        glBindTexture     ( GL_TEXTURE_2D, tex );

                                                        if (numofbytes==3)
                                                                gluBuild2DMipmaps ( GL_TEXTURE_2D, numofbytes,
                                                                                        width, height,
                                                                                        GL_RGB, GL_UNSIGNED_BYTE, img );
                                                        else
                                                        if (numofbytes==4)
                                                                gluBuild2DMipmaps ( GL_TEXTURE_2D, numofbytes,
                                                                                        width, height,
                                                                                        GL_RGBA, GL_UNSIGNED_BYTE, img );
                                                }
                                        }

                                        if (!allocateimage)
                                        {
                                                free (img);
                                                img = NULL;
                                        }

                                        frm->SetInfo      ( img, tex );
                                }
                                else
                                {
                                        frm->SetInfo      ( img, -1 );
                                        return false;
                                }
                        }
                        else
                        {
                                frm->SetInfo      ( NULL, -1 );
                                return false;
                        }
                }
                else
                return false;
        }
        return true;
}

bool TTexture::SaveToFile   ( mstring fname )
{
        TGameFileStream stream;
        if ( stream.OpenFileForCreate( fname ) )
                SaveToStream ( &stream );
                else
                return false;
        return true;
}

mstring TTexture::GetFullFilePath   ( )
{
        return mExtractPathOfFile(filename);
}

mstring TTexture::GetFullFileName   ( )
{
        return filename;
}

mstring TTexture::GetOnlyFileName   ( )
{
        return mExtractNameOfFile(filename);
}

void		 TTexture::SetInfo    	   ( 	unsigned int    pwidth,
					        unsigned int    pheight,
					        unsigned int    pnumofbytes )
{
	width	= pwidth;
        height	= pheight;
	numofbytes = pnumofbytes;
}

//********************************************************
// TImageLoader
//********************************************************

TImageLoader::TImageLoader ()
{
}

TImageLoader::~TImageLoader ()
{
        ClearAll();
}

void TImageLoader::ClearAll()
{
        while ( textures.size()>0 )
                DeleteTextureByIndex(0);
}

TTexture  * TImageLoader::LinkImage( mstring fname, bool clone,
                                     bool allocateimage, bool allocatetexture,
                                     mstring rootpath )
{
        mstring yedekf = fname;

        TTexture * tex;
        mstring CurPath =  mGetCurrentDir();
        if (fname.find(":")==-1 && fname.find("\\\\")==-1) fname = rootpath+"\\"+fname;

        mstring yedekf2 = fname;

        // set currrent directory as the model directory
        mSetCurrentDir( mExtractPathOfFile(fname) );
        try
        {
                mstring file = mExtractNameOfFile(fname);
                fname = mGetCurrentDir()+"\\"+mUpperCase(file);

                // Search if the file is already loaded
                if (!clone)
                {
                        int i;
                        for (i=0; i<textures.size(); i++)
                                if (textures[i]->filename==fname && !textures[i]->clone)
                                {
                                        textures[i]->numoflinks++;
                                        mSetCurrentDir( CurPath );
                                        return textures[i];
                                }
                }

                // if not loaded then load it
                tex = new TTexture;
                if (tex)
                {
                        tex->filename    = fname;
                        tex->numoflinks  = 1;
                        tex->clone       = clone;

                        tex->LoadFromFile(file, allocateimage, allocatetexture );

                        textures.push_back(tex);
                }
        }
        __finally
        {
                mSetCurrentDir( CurPath );
        }
        return tex;
}

void  TImageLoader::UnlinkImage (TTexture  * tex)
{
        int i;
        for (i=0; i<textures.size(); i++)
                if (textures[i]==tex)
                {
                        if (textures[i]->numoflinks<=1)
                                DeleteTextureByIndex(i);
                                else
                                textures[i]->numoflinks--;
                        return;
                }
}

void  TImageLoader::DeleteTextureByIndex (int index)
{
        if (index<0 || index>=textures.size()) return;
        delete textures[index];
		textures.erase ( textures.begin()+index );
}

void  TImageLoader::DeleteImage (TTexture * tex)
{
        int i;
        for (i=0; i<textures.size(); i++)
                if (textures[i]==tex)
                {
                        DeleteTextureByIndex(i);
                        return;
                }
}


