//******************************************************************************
// Written by   : Cagatay Undeger
// Date         : 18.12.2003
// Version      : 2.0
//******************************************************************************

#ifndef __GAME_ENVIRONMENT_IMAGE_LOADER__
#define __GAME_ENVIRONMENT_IMAGE_LOADER__

#include "main_functions.h"
#include "io_streams.h"

#include <GL/gl.h>
#include <GL/glu.h>

#include <vector>

//********************************************************
// Classes
//********************************************************

class TTextureFrame;
class TTexture;
class TImageLoader;

//********************************************************
// Externals
//********************************************************

extern TImageLoader ImageLoader;

//********************************************************
// Type Definitions
//********************************************************

typedef enum { animloopJumpStartAtEnd, animloopStopAtEnd }
TAnimLoopMode;

//********************************************************
// TTextureFrame
//********************************************************

class TTextureFrame
{
        friend class TTexture;
        friend class TImageLoader;
protected:
        GLuint          texture;
        unsigned char * image;
public:
        TTextureFrame  ( );
        ~TTextureFrame ( );

        unsigned char * GetImage   ( ); // image data is loaded upside down and B,G,R order
        GLuint          GetTexture ( );

        void		SetInfo    ( unsigned char * img, GLuint tex );
};

//********************************************************
// TTexture
//********************************************************

class TTexture
{
        friend class TImageLoader;
protected:
        mstring         filename;
        unsigned int    numoflinks;
        bool            clone;

        unsigned int    width;
        unsigned int    height;
        unsigned int    numofbytes;
        unsigned int    duration;
        TAnimLoopMode   animloopmode;

        unsigned int    startframe;
        unsigned int    endframe;
        unsigned int    currenttime;
        unsigned int    currentframe;
        std::vector<TTextureFrame*> frames;
public:
        TTexture  ( );
        ~TTexture ( );

        mstring          GetFullFilePath   ( );
        mstring          GetFullFileName   ( );
        mstring          GetOnlyFileName   ( );

        unsigned int     GetWidth          ( );
        unsigned int     GetHeight         ( );
        unsigned int     GetNumOfBytes     ( );
        unsigned int     GetDuration       ( );
        TAnimLoopMode    GetAnimLoopMode   ( );
        unsigned int     GetStartFrame     ( );
        unsigned int     GetEndFrame       ( );

        unsigned int     GetNumOfFrames    ( );
        TTextureFrame  * GetFrame          ( unsigned int index );
        TTextureFrame  * GetCurrentFrame   ( );
        unsigned int     GetCurrentFrameNo ( );
        TTextureFrame  * AddFrame          ( );

        void             SetDuration       ( unsigned int  fduration );
        void             SetAnimLoopMode   ( TAnimLoopMode lmode );
        void             SetCurrentFrameNo ( unsigned int index );
        void             SetAnimationRange ( unsigned int startf, unsigned int endf );

        void             ResetAnimation    ( );
        void             AdvanceAnimation  ( );
        TTextureFrame  * BindTexture       ( bool advanceanim = true );

        void             ClearAll          ( );
        bool             LoadFromFile      ( mstring fname, bool allocateimage = true,
        				     bool allocatetexture = true, bool append = false );
        bool             SaveToFile        ( mstring fname );

        bool             LoadFromStream    ( TGameIOStream * stream, bool allocateimage, bool allocatetexture, bool append );
        bool             SaveToStream      ( TGameIOStream * stream );

        void		 SetInfo    	   ( 	unsigned int    pwidth,
					        unsigned int    pheight,
					        unsigned int    pnumofbytes );
};

//********************************************************
// TImageLoader
//********************************************************

class TImageLoader
{
protected:
        std::vector<TTexture*> textures;
        void DeleteTextureByIndex ( int index );
public:

        TImageLoader();
        virtual ~TImageLoader ();

        TTexture * LinkImage            ( mstring fname, bool clone,
        				  bool allocateimage, bool allocatetexture,
                                          mstring rootpath = ApplicationDirectory ); // bmp, tga, avi
        void       UnlinkImage          ( TTexture * img );

        void       ClearAll             ( );
        void       DeleteImage          ( TTexture * img );
};

#endif
