#include "MeshImport3ds.h"
#include "l3ds.h"

#ifdef __BCPLUSPLUS__

#else
	#define try 		{
	#define __finally 	}
	#define catch(x) 	}
#endif

bool Get3DSMesh (       L3DS * l3ds,
                        TMesh * mesh,
                        bool append_frame_only,
                        unsigned int frame_no,
                        float nearplane,
                        float farplane,
                        void * LoaderParameters )
{
        // source objects
	LMesh     * lpart;
	LVector4    vertex;
        LVector2    mappingcoord;
        LTri        tri;
        LColor3     ambc,difc,spec;
	LMaterial * lmaterial;
        LMap        map;
        LVector3    nor1,nor2,nor3;
        float       shine, alpha;
        mstring     mapfile, bumpfile;

        LColor3 defaultcol;
	LMaterial defaultmaterial;
        defaultcol.r = 1; defaultcol.g = 1; defaultcol.b = 1;
        defaultmaterial.SetDiffuseColor( defaultcol );

        // destination objects
        TMeshPart  * meshpart;
        TMeshFrame * meshframe;
        TVertex3f  * vert3f;
        TVertex2f  * vert2f;
        TColor3f   * col3f;
        TTriangle  * triangle;
        float      * partshine;
        float      * partalpha;
        TMeshMaterial  * material;
        int              materialid;

        int start_vertex_index;
        int start_mapco_index;
        int start_material_index;
        unsigned int index;
	int i,j;

        // get/create frame
        while ( mesh->GetNumOfFrames()<= frame_no ) mesh->AddFrame( );
        meshframe = mesh->GetFrame( frame_no );
        if (!meshframe) return false;

        // add materials
        start_material_index = mesh->GetNumOfMaterials();
        if (!append_frame_only)
        {
                for ( i = 0; i< l3ds->GetMaterialCount(); i++ )
                {
                        lmaterial  = l3ds->GetMaterial ( i );
                        map        = lmaterial->GetTextureMap1 ( );
                        mapfile    = map.mapName;
                        bumpfile   = map.mapName;

                        material = mesh->AddMaterial  ( &index );
                        material->SetDiffuseFileName  ( mapfile );
                }
        }

	for ( i = 0; i < l3ds->GetMeshCount(); i++)     // for each part of model
	{
		lpart =  l3ds->GetMesh(i);
                if (lpart)
		{
                        if (!append_frame_only) // create a new part if appending new mesh
                        {
        			meshpart = mesh->AddPart( );
                                meshpart->SetNearPlane  ( nearplane );
                                meshpart->SetFarPlane   ( farplane  );
                        }
                        else
                        meshpart = NULL;

                        start_vertex_index = meshframe->GetNumOfVertices();
                        start_mapco_index  = meshframe->GetNumOfMappingCoords();

			for (j = 0; j<lpart->GetVertexCount();j++) // add vertices
			{
				vertex     = lpart->GetVertex(j);
				vert3f     = meshframe->AddVertex ( );
                                vert3f->x  = vertex.x;
                                vert3f->y  = vertex.y;
                                vert3f->z  = vertex.z;
			}

			for (j = 0; j<lpart->GetUVCount();j++) // add mapping coords
			{
				mappingcoord = lpart->GetUV(j);
				vert2f       = meshframe->AddMappingCoord ( );
                                vert2f->x    = mappingcoord.x;
                                vert2f->y    = mappingcoord.y;
			}

			for (j=0; j<lpart->GetTriangleCount();j++) // for each triangle
			{
				tri       = lpart->GetTri(j);
                                if ( tri.materialId < l3ds->GetMaterialCount() )
	                                lmaterial = l3ds->GetMaterial(tri.materialId);
                                        else
                                        lmaterial = &defaultmaterial;

                                nor1      = tri.normal[0];
                                nor2      = tri.normal[1];
                                nor3      = tri.normal[2];

                                ambc	  = lmaterial->GetAmbientColor ( );
                                difc	  = lmaterial->GetDiffuseColor ( );
                                spec	  = lmaterial->GetSpecularColor( );
                                shine     = lmaterial->GetShininess( );
                                alpha     = 1.0f - lmaterial->GetTransparency( );

                                materialid = start_material_index + tri.materialId;

                                if (meshpart)
                                        triangle = meshpart->AddTriangle ( );

                                // vertices
                                if (meshpart)
                                {
                                        triangle->vertices[0] = tri.a + start_vertex_index;
                                        triangle->vertices[1] = tri.b + start_vertex_index;
                                        triangle->vertices[2] = tri.c + start_vertex_index;
                                }

                                // mapping coords
                                if (meshpart)
                                {
                                        triangle->mappingcoords[0] = tri.a + start_mapco_index;
                                        triangle->mappingcoords[1] = tri.b + start_mapco_index;
                                        triangle->mappingcoords[2] = tri.c + start_mapco_index;
                                }

                                // normal 1
                                vert3f     = meshframe->AddNormal ( &index );
                                vert3f->x  = nor1.x;
                                vert3f->y  = nor1.y;
                                vert3f->z  = nor1.z;
                                if (meshpart)
                                        triangle->normals[0] = index;

                                // normal 2
                                vert3f     = meshframe->AddNormal ( &index );
                                vert3f->x  = nor2.x;
                                vert3f->y  = nor2.y;
                                vert3f->z  = nor2.z;
                                if (meshpart)
                                        triangle->normals[1] = index;

                                // normal 3
                                vert3f     = meshframe->AddNormal ( &index );
                                vert3f->x  = nor3.x;
                                vert3f->y  = nor3.y;
                                vert3f->z  = nor3.z;
                                if (meshpart)
                                        triangle->normals[2] = index;

                                // diffuse
                                col3f     = meshframe->AddDiffuse ( &index );
                                col3f->r  = difc.r;
                                col3f->g  = difc.g;
                                col3f->b  = difc.b;
                                if (meshpart)
                                        triangle->diffuse = index;

                                // specular
                                col3f     = meshframe->AddSpecular ( &index );
                                col3f->r  = spec.r;
                                col3f->g  = spec.g;
                                col3f->b  = spec.b;
                                if (meshpart)
                                        triangle->specular = index;

                                // shininess
                                partshine  = meshframe->AddShininess ( &index );
                                *partshine = shine;
                                if (meshpart)
                                        triangle->shininess = index;

                                // alpha
                                partalpha  = meshframe->AddAlpha ( &index );
                                *partalpha = alpha;
                                if (meshpart)
                                        triangle->alpha = index;

                                // material id
                                if (meshpart)
                                        triangle->materialID = materialid;
			}
		}
	}
        return true;
}

bool Import3dsFile (    mstring fname,
                        TMesh * mesh,
                        bool append_frame_only,
                        unsigned int frame_no,
                        float nearplane,
                        float farplane,
                        void * LoaderParameters )
{
        bool loadok = false;
        L3DS * l3ds = new L3DS();
        try
        {
                loadok = l3ds->LoadFile(fname);
                if (loadok)
                {
                        loadok = Get3DSMesh( l3ds, mesh, append_frame_only,
                                             frame_no, nearplane, farplane,
                                             LoaderParameters );
                        mesh->UpdateBBox();
                }
        }
        __finally
        {
                delete l3ds;
        }
        return loadok;
}

