#include "MeshLoader.h"
#include "MeshImport3ds.h"

#include <math.h>

#ifdef __BCPLUSPLUS__

#else
	#define try 		{
	#define __finally 	}
	#define catch(x) 	}
#endif

//********************************************************
// Externals
//********************************************************

TMeshLoader MeshLoader;

//********************************************************
// TMeshMaterial
//********************************************************

TMeshMaterial::TMeshMaterial  ( )
{
        mesh                   = NULL;
        name                   = "material";
        materialtype           = 0;
        diffusefile            = "";
        lastdiffusefileloaded  = "";
        lastclonetex	       = false;
        diffusetexture         = NULL;
}

TMeshMaterial::~TMeshMaterial ( )
{
        if (diffusetexture) ImageLoader.UnlinkImage ( diffusetexture );
        diffusetexture = NULL;
}

TMesh         * TMeshMaterial::GetMesh         ( )
{
        return mesh;
}

mstring         TMeshMaterial::GetName         ( )
{
        return name;
}

mstring         TMeshMaterial::GetDiffuseFileName     ( )
{
        return diffusefile;
}

int             TMeshMaterial::GetMaterialType ( )
{
        return materialtype;
}

TTexture      * TMeshMaterial::GetDiffuseTexture      ( )
{
        return diffusetexture;
}

void            TMeshMaterial::SetName         ( mstring pname )
{
        name = pname;
}

void            TMeshMaterial::SetDiffuseFileName     ( mstring fname )
{
        diffusefile = fname;
}

void            TMeshMaterial::SetMaterialType ( int mtype )
{
        materialtype = mtype;
}

void TMeshMaterial::LoadMaterial ( )
{
        if ( lastdiffusefileloaded != diffusefile || lastclonetex != mesh->clonetexture )
        {
                mesh->istransparent        = -1;
                lastdiffusefileloaded   = diffusefile;

                if ( diffusefile != "" )
                {
                        mstring meshpath = mesh->GetFullFilePath();
                        mstring CurPath =  mGetCurrentDir();
                        mSetCurrentDir( meshpath );
                        try
                        {
                                if (diffusetexture)
                                        ImageLoader.UnlinkImage ( diffusetexture );
                                lastclonetex = mesh->clonetexture;

                                diffusetexture = ImageLoader.LinkImage ( diffusefile,
                                                                         lastclonetex,
                                                                         false, true,
                                                                         mGetCurrentDir() );
                        }
                        __finally
                        {
                                mSetCurrentDir( CurPath );
                        }
                }
                else
                {
                        if (diffusetexture)
                                ImageLoader.UnlinkImage ( diffusetexture );
                        lastclonetex   = mesh->clonetexture;
                        diffusetexture = NULL;
                }
        }
}

//********************************************************
// TMeshFrame
//********************************************************

TMeshFrame::TMeshFrame  ( )
{
        mesh        = NULL;
        name        = "frame";
}

TMeshFrame::~TMeshFrame ( )
{
}

TMesh          * TMeshFrame::GetMesh ( )
{
        return mesh;
}

mstring          TMeshFrame::GetName ( )
{
        return name;
}

void             TMeshFrame::SetName ( mstring pname )
{
        name = pname;
}

int              TMeshFrame::GetNumOfVertices  ( )
{
        return vertices.size();
}

TVertex3f *      TMeshFrame::GetVertex         ( unsigned int index )
{
        if (index>=vertices.size()) return NULL;
        return &vertices[index];
}

TVertex3f *      TMeshFrame::AddVertex         ( unsigned int * index )
{
        TVertex3f v;
        if (index) *index = vertices.size();
        vertices.push_back(v);
        return &vertices[vertices.size()-1];
}

void             TMeshFrame::DeleteVertex      ( unsigned int index )
{
        if (index>=vertices.size()) return;
		vertices.erase ( vertices.begin()+index );
}

int              TMeshFrame::GetNumOfNormals   ( )
{
        return normals.size();
}

TVertex3f *      TMeshFrame::GetNormal         ( unsigned int index )
{
        if (index>=normals.size()) return NULL;
        return &normals[index];
}

TVertex3f *      TMeshFrame::AddNormal         ( unsigned int * index )
{
        TVertex3f v;
        if (index) *index = normals.size();
        normals.push_back(v);
        return &normals[normals.size()-1];
}

void             TMeshFrame::DeleteNormal      ( unsigned int index )
{
        if (index>=normals.size()) return;
		normals.erase ( normals.begin()+index );
}

int              TMeshFrame::GetNumOfDiffuses   ( )
{
        return diffuses.size();
}

TColor3f *       TMeshFrame::GetDiffuse          ( unsigned int index )
{
        if (index>=diffuses.size()) return NULL;
        return &diffuses[index];
}

TColor3f *       TMeshFrame::AddDiffuse         ( unsigned int * index )
{
        TColor3f v;
        if (index) *index = diffuses.size();
        diffuses.push_back(v);
        return &diffuses[diffuses.size()-1];
}

void             TMeshFrame::DeleteDiffuse      ( unsigned int index )
{
        if (index>=diffuses.size()) return;
		diffuses.erase ( diffuses.begin()+index );
}

int              TMeshFrame::GetNumOfSpeculars    ( )
{
        return speculars.size();
}

TColor3f *       TMeshFrame::GetSpecular          ( unsigned int index )
{
        if (index>=speculars.size()) return NULL;
        return &speculars[index];
}

TColor3f *       TMeshFrame::AddSpecular         ( unsigned int * index )
{
        TColor3f v;
        if (index) *index = speculars.size();
        speculars.push_back(v);
        return &speculars[speculars.size()-1];
}

void             TMeshFrame::DeleteSpecular      ( unsigned int index )
{
        if (index>=speculars.size()) return;
		speculars.erase ( speculars.begin()+index );
}

int              TMeshFrame::GetNumOfShininess  ( )
{
        return shininess.size();
}

float *       TMeshFrame::GetShininess         ( unsigned int index )
{
        if (index>=shininess.size()) return NULL;
        return &shininess[index];
}

float *       TMeshFrame::AddShininess       ( unsigned int * index )
{
        float v;
        if (index) *index = shininess.size();
        shininess.push_back(v);
        return &shininess[shininess.size()-1];
}

void             TMeshFrame::DeleteShininess      ( unsigned int index )
{
        if (index>=shininess.size()) return;
		shininess.erase ( shininess.begin()+index );
}

int              TMeshFrame::GetNumOfAlphas  ( )
{
        return alphas.size();
}

float *       TMeshFrame::GetAlpha      ( unsigned int index )
{
        if (index>=alphas.size()) return NULL;
        return &alphas[index];
}

float *       TMeshFrame::AddAlpha       ( unsigned int * index )
{
        float v;
        if (index) *index = alphas.size();
        alphas.push_back(v);
        return &alphas[alphas.size()-1];
}

void             TMeshFrame::DeleteAlpha      ( unsigned int index )
{
        if (index>=alphas.size()) return;
		alphas.erase ( alphas.begin()+index );
}

int              TMeshFrame::GetNumOfMappingCoords ( )
{
        return mappingcoords.size();
}

TVertex2f *      TMeshFrame::GetMappingCoord       ( unsigned int index )
{
        if (index>=mappingcoords.size()) return NULL;
        return &mappingcoords[index];
}

TVertex2f *      TMeshFrame::AddMappingCoord       ( unsigned int * index )
{
        TVertex2f v;
        if (index) *index = mappingcoords.size();
        mappingcoords.push_back(v);
        return &mappingcoords[mappingcoords.size()-1];
}

void             TMeshFrame::DeleteMappingCoord    ( unsigned int index )
{
        if (index>=mappingcoords.size()) return;
		mappingcoords.erase ( mappingcoords.begin()+index );
}

//********************************************************
// TMeshBoundingBox
//********************************************************

TMeshBoundingBox::TMeshBoundingBox ()
{
        centerx = 0;
        centery = 0;
        centerz = 0;
        sizex   = 0;
        sizey   = 0;
        sizez   = 0;
        sizediagonal   = 0;
        sizexydiagonal = 0;

        minx = 0;
        miny = 0;
        minz = 0;
        maxx = 0;
        maxy = 0;
        maxz = 0;
}

//********************************************************
// TTriangle
//********************************************************

TTriangle::TTriangle ()
{
        int i;
        for ( i = 0; i<3; i++)
        {
                vertices [i] = -1;
                normals  [i] = -1;
                mappingcoords [i] = -1;
        }
        diffuse   = -1;
        specular  = -1;
	shininess = -1;
	alpha     = -1;
        materialID= -1;
}

//********************************************************
// TMeshPart
//********************************************************

TMeshPart::TMeshPart  ( )
{
        mesh        = NULL;
        name        = "part";
        nearplane   = 0;
        farplane    = 3000;
        hbillboard  = false;
        pbillboard  = false;
        rbillboard  = false;
        doublefaced = false;
        visiblealphathreshold = 0;
        emission    = TColor3f(0,0,0);
        visible     = true;
        alpha       = 1;
}

TMeshPart::~TMeshPart ( )
{
}

TMesh          * TMeshPart::GetMesh ( )
{
        return mesh;
}

mstring          TMeshPart::GetName ( )
{
        return name;
}

float            TMeshPart::GetNearplane   ( )
{
        return nearplane;
}
float            TMeshPart::GetFarPlane    ( )
{
        return farplane;
}

bool             TMeshPart::GetHBillboard ( )
{
        return hbillboard;
}

bool             TMeshPart::GetPBillboard ( )
{
        return pbillboard;
}

bool             TMeshPart::GetRBillboard ( )
{
        return rbillboard;
}

TColor3f TMeshPart::GetEmission    ( )
{
        return emission;
}

bool TMeshPart::GetDoubleFaced    ( )
{
        return doublefaced;
}

float TMeshPart::GetAlpha       ( )
{
        return alpha;
}

bool TMeshPart::GetVisible ( )
{
        return visible;
}

TMeshBoundingBox * TMeshPart::GetBBox  ( )
{
        return &bbox;
}

void TMeshPart::UpdateBBox ( )
{
        float minx = 0, miny = 0, minz = 0, maxx = 0, maxy = 0, maxz = 0;
        bool first = true;
        TVertex3f va,vb,vc;
        TTriangle * tri;
        int a,b,c;
        unsigned int t, f, numt, numf, numv;
        numt = triangles.size();
        numf = mesh->frames.size();
        for ( t = 0; t<numt; t++ )      // for each triangle
        {
                tri = &triangles[t];
                a   = tri->vertices[0];
                b   = tri->vertices[1];
                c   = tri->vertices[2];
                if (a!=-1 && b!=-1 && c!=-1)
                {
                        for ( f = 0; f<numf; f++ )      // for each frame of triangle
                        {
                                numv = mesh->frames[f]->vertices.size();
                                if (a<numv && b<numv && c<numv)
                                {
                                        va = mesh->frames[f]->vertices[a];
                                        vb = mesh->frames[f]->vertices[b];
                                        vc = mesh->frames[f]->vertices[c];

                                        if ( first || (va.x<minx) ) minx = va.x;
                                        if ( first || (va.y<miny) ) miny = va.y;
                                        if ( first || (va.z<minz) ) minz = va.z;
                                        if ( (vb.x<minx) ) minx = vb.x;
                                        if ( (vb.y<miny) ) miny = vb.y;
                                        if ( (vb.z<minz) ) minz = vb.z;
                                        if ( (vc.x<minx) ) minx = vc.x;
                                        if ( (vc.y<miny) ) miny = vc.y;
                                        if ( (vc.z<minz) ) minz = vc.z;

                                        if ( first || (va.x>maxx) ) maxx = va.x;
                                        if ( first || (va.y>maxy) ) maxy = va.y;
                                        if ( first || (va.z>maxz) ) maxz = va.z;
                                        if ( (vb.x>maxx) ) maxx = vb.x;
                                        if ( (vb.y>maxy) ) maxy = vb.y;
                                        if ( (vb.z>maxz) ) maxz = vb.z;
                                        if ( (vc.x>maxx) ) maxx = vc.x;
                                        if ( (vc.y>maxy) ) maxy = vc.y;
                                        if ( (vc.z>maxz) ) maxz = vc.z;

                                        first = false;
                                }
                        }
                }
        }

        if ( hbillboard )
        {
                // rotate around c (0,0,0)
                // |----------|
                // |       |  |
                // |-------c--|
                // |       |  |
                // |       |  |
                // |----------|

                double minsizex = minx<0 ? fabs(minx) : 0;
                double minsizey = miny<0 ? fabs(miny) : 0;
                double maxsizex = maxx>0 ? maxx : 0;
                double maxsizey = maxy>0 ? maxy : 0;

                double sizex1y1 = minsizex*minsizex + minsizey*minsizey;
                if ( sizex1y1<0 ) sizex1y1 = 0;
                sizex1y1 = sqrt(sizex1y1);

                double sizex2y1 = maxsizex*maxsizex + minsizey*minsizey;
                if ( sizex2y1<0 ) sizex2y1 = 0;
                sizex2y1 = sqrt(sizex2y1);

                double sizex2y2 = maxsizex*maxsizex + maxsizey*maxsizey;
                if ( sizex2y2<0 ) sizex2y2 = 0;
                sizex2y2 = sqrt(sizex2y2);

                double sizex1y2 = minsizex*minsizex + maxsizey*maxsizey;
                if ( sizex1y2<0 ) sizex1y2 = 0;
                sizex1y2 = sqrt(sizex1y2);

                double   maxr = MaximumOf ( MaximumOf (sizex1y1,sizex2y1), MaximumOf (sizex1y2,sizex2y2) );

                minx = -maxr;
                maxx =  maxr;
                miny = -maxr;
                maxy =  maxr;
        }
        if ( pbillboard )
        {
                double minsizez = minz<0 ? fabs(minz) : 0;
                double minsizey = miny<0 ? fabs(miny) : 0;
                double maxsizez = maxz>0 ? maxz : 0;
                double maxsizey = maxy>0 ? maxy : 0;

                double sizez1y1 = minsizez*minsizez + minsizey*minsizey;
                if ( sizez1y1<0 ) sizez1y1 = 0;
                sizez1y1 = sqrt(sizez1y1);

                double sizez2y1 = maxsizez*maxsizez + minsizey*minsizey;
                if ( sizez2y1<0 ) sizez2y1 = 0;
                sizez2y1 = sqrt(sizez2y1);

                double sizez2y2 = maxsizez*maxsizez + maxsizey*maxsizey;
                if ( sizez2y2<0 ) sizez2y2 = 0;
                sizez2y2 = sqrt(sizez2y2);

                double sizez1y2 = minsizez*minsizez + maxsizey*maxsizey;
                if ( sizez1y2<0 ) sizez1y2 = 0;
                sizez1y2 = sqrt(sizez1y2);

                double   maxr = MaximumOf ( MaximumOf (sizez1y1,sizez2y1), MaximumOf (sizez1y2,sizez2y2) );

                minz = -maxr;
                maxz =  maxr;
                miny = -maxr;
                maxy =  maxr;
        }
        if ( rbillboard )
        {
                double minsizez = minz<0 ? fabs(minz) : 0;
                double minsizex = minx<0 ? fabs(minx) : 0;
                double maxsizez = maxz>0 ? maxz : 0;
                double maxsizex = maxx>0 ? maxx : 0;

                double sizez1x1 = minsizez*minsizez + minsizex*minsizex;
                if ( sizez1x1<0 ) sizez1x1 = 0;
                sizez1x1 = sqrt(sizez1x1);

                double sizez2x1 = maxsizez*maxsizez + minsizex*minsizex;
                if ( sizez2x1<0 ) sizez2x1 = 0;
                sizez2x1 = sqrt(sizez2x1);

                double sizez2x2 = maxsizez*maxsizez + maxsizex*maxsizex;
                if ( sizez2x2<0 ) sizez2x2 = 0;
                sizez2x2 = sqrt(sizez2x2);

                double sizez1x2 = minsizez*minsizez + maxsizex*maxsizex;
                if ( sizez1x2<0 ) sizez1x2 = 0;
                sizez1x2 = sqrt(sizez1x2);

                double   maxr = MaximumOf ( MaximumOf (sizez1x1,sizez2x1), MaximumOf (sizez1x2,sizez2x2) );

                minz = -maxr;
                maxz =  maxr;
                minx = -maxr;
                maxx =  maxr;
        }

        bbox.minx = minx;
        bbox.miny = miny;
        bbox.minz = minz;
        bbox.maxx = maxx;
        bbox.maxy = maxy;
        bbox.maxz = maxz;

        bbox.centerx = (minx+maxx)/2;
        bbox.centery = (miny+maxy)/2;
        bbox.centerz = (minz+maxz)/2;

        bbox.sizex   = fabs(minx-maxx);
        bbox.sizey   = fabs(miny-maxy);
        bbox.sizez   = fabs(minz-maxz);

        double   r = bbox.sizex*bbox.sizex + bbox.sizey*bbox.sizey;
        if (r<0) r = 0;
        bbox.sizexydiagonal = sqrt(r);

        r += bbox.sizez*bbox.sizez;
        if (r<0) r = 0;
        bbox.sizediagonal = sqrt(r);
}

void             TMeshPart::SetName        ( mstring pname )
{
        name = pname;
}

void             TMeshPart::SetNearPlane   ( float np )
{
        nearplane = np;
}

void             TMeshPart::SetFarPlane    ( float fp )
{
        farplane = fp;
}

void             TMeshPart::SetHBillboard ( bool billb )
{
        hbillboard = billb;
}

void             TMeshPart::SetPBillboard ( bool billb )
{
        pbillboard = billb;
}

void             TMeshPart::SetRBillboard ( bool billb )
{
        rbillboard = billb;
}

void TMeshPart::SetEmission    ( TColor3f emis )
{
        emission = emis;
}

void TMeshPart::SetDoubleFaced  ( bool douface )
{
        doublefaced = douface;
}

void             TMeshPart::SetAlpha       ( float alph )
{
        alpha = alph;
}

float TMeshPart::GetVisibleAlphaThreshold ( )
{
        return visiblealphathreshold;
}

void             TMeshPart::SetVisibleAlphaThreshold ( float vist )
{
        if ( vist<0 ) vist = 0;
        if ( vist>1 ) vist = 1;
        visiblealphathreshold = vist;
}

void             TMeshPart::SetVisible     ( bool vis )
{
        visible = vis;
}

int              TMeshPart::GetNumOfTriangles ( )
{
        return triangles.size();
}

TTriangle *      TMeshPart::GetTriangle       ( unsigned int index )
{
        if ( index>=triangles.size() ) return NULL;
        return &triangles[index];
}

TTriangle *      TMeshPart::AddTriangle       ( unsigned int * index )
{
        TTriangle v;
        if (index) *index = triangles.size();
        triangles.push_back(v);
        return &triangles[triangles.size()-1];
}

void             TMeshPart::DeleteTriangle    ( unsigned int index )
{
        if (index>=triangles.size()) return;
		triangles.erase ( triangles.begin()+index );
}

bool TMeshPart::IsTransparent ( )
{
        int t,f;
        TMeshMaterial * material;
        TTriangle     * tri;
        int trialpha;
        int numtris = triangles.size();
        for ( t = 0; t<numtris; t++ )
        {
                tri = &triangles[t];

                if ( tri->materialID>=0 &&
                     tri->materialID < mesh->materials.size() )
                        material = mesh->materials[ tri->materialID ];
                        else
                        material = NULL;

                // Has texture with alpha channel?
                if ( material  &&
                     material->diffusetexture &&
                     material->diffusetexture->GetNumOfFrames()>0 &&
                     material->diffusetexture->GetNumOfBytes()>3 )
                     return true;

                // Has transparent surfaces?
                trialpha = tri->alpha;
                if ( trialpha >= 0 )
                {
                        for ( f = 0; f<mesh->frames.size(); f++ )
                        if ( trialpha < mesh->frames[f]->alphas.size() &&
                             mesh->frames[f]->alphas[trialpha] < 1.0 )
                             return true;
                }
        }
        return false;
}

//********************************************************
// TMesh
//********************************************************

TMesh::TMesh  ( )
{
        filename     = "";
        numoflinks   = 1;
        clone        = false;
        clonetexture = false;
        currentframe = 0;
        targetframe  = 0;
        frametransition = 0;
        smoothframetransitions = true;
        alpha        = 1;
        data         = NULL;
        istransparent= false;
}

TMesh::~TMesh ( )
{
        ClearAll ( );
}

void TMesh::ClearAll     ( )
{
		unsigned int i;
        for ( i = 0; i<frames.size(); i++ )
                delete frames[i];

        for ( i = 0; i<materials.size(); i++ )
                delete materials[i];

        for ( i = 0; i<parts.size(); i++ )
                delete parts[i];

        frames.clear();
        materials.clear();
        parts.clear();

        currentframe    = 0;
        targetframe     = 0;
        frametransition = 0;
}

unsigned int TMesh::GetNumOfFrames  ( )
{
        return frames.size();
}

TMeshFrame * TMesh::GetFrame        ( unsigned int index )
{
        if ( index>=frames.size() ) return NULL;
        return frames[index];
}

unsigned int TMesh::GetCurrentFrameNo ( )
{
        return currentframe;
}

unsigned int TMesh::GetTargetFrameNo ( )
{
        return targetframe;
}

void TMesh::SetCurrentFrameNo ( unsigned int index )
{
        if (index>=frames.size()) return;
        currentframe = index;
}

void TMesh::SetTargetFrameNo ( unsigned int index )
{
        if (index>=frames.size()) return;
        targetframe = index;
}

float TMesh::GetFrameTransition( )
{
        return frametransition;
}

bool TMesh::GetSmoothFrameTransitions( )
{
        return smoothframetransitions;
}

void  TMesh::SetFrameTransition( float trns )
{
        if (trns<0) trns = 0;
        if (trns>1) trns = 1;
        frametransition = trns;
}

void  TMesh::SetSmoothFrameTransitions( bool smt )
{
        smoothframetransitions = smt;
}

int              TMesh::GetNumOfTriangles ( )
{
        int tot = 0;
        for ( int i = 0; i<parts.size(); i++ )
                tot += parts[i]->GetNumOfTriangles ( );
        return tot;
}

TMeshFrame * TMesh::GetCurrentFrame ( )
{
        if ( frames.size() == 0 ) return NULL;
        if ( currentframe>=frames.size() ) currentframe = frames.size()-1;
        return frames[currentframe];
}

TMeshFrame * TMesh::GetTargetFrame ( )
{
        if ( frames.size() == 0 ) return NULL;
        if ( targetframe>=frames.size() ) targetframe = frames.size()-1;
        return frames[targetframe];
}

bool TMesh::LoadFromFile ( mstring fname,
                           bool append_frame_only,
                           unsigned int frame_no,
                           float nearplane,
                           float farplane,
                           void * LoaderParameters )
{
        if (!mIsFileExist(fname) )
                return false;

        istransparent = -1;

        mstring last3 = mUpperCase( fname.substr( fname.length()-3,3 ) );
        mstring last4 = mUpperCase( fname.substr( fname.length()-4,4 ) );

        if (last3=="M3B")
        {
                TGameFileStream stream;
                TGameMemoryStream memstream;
                if ( stream.OpenFileForRead( fname ) )
                {
                        memstream.SetSize( stream.GetSize() );
                        stream.Read( memstream.GetMemory(), stream.GetSize() );
                        return LoadFromStream ( &memstream, append_frame_only, frame_no, nearplane, farplane, LoaderParameters );;
                }
                else
                return false;
        }
        else
        if (last3=="3DS")
                return Import3dsFile ( fname, this, append_frame_only,
                                       frame_no, nearplane, farplane,
                                       LoaderParameters );

        return false;
}

bool TMesh::SaveToStream ( TGameIOStream * stream )
{
        unsigned int signature = 0x40a925fc;
        int version   = 3;

        stream->Write ( &signature );
        stream->Write ( &version   );

        // general info
        stream->Write( &smoothframetransitions );
        stream->Write( &alpha );

        // bbox info
        stream->Write( &bbox.centerx );
        stream->Write( &bbox.centery );
        stream->Write( &bbox.centerz );
        stream->Write( &bbox.sizex );
        stream->Write( &bbox.sizey );
        stream->Write( &bbox.sizez );
        stream->Write( &bbox.sizexydiagonal );
        stream->Write( &bbox.sizediagonal );
        stream->Write( &bbox.minx );
        stream->Write( &bbox.miny );
        stream->Write( &bbox.minz );
        stream->Write( &bbox.maxx );
        stream->Write( &bbox.maxy );
        stream->Write( &bbox.maxz );

        int i,j,numof;

        int numframes = frames.size();
        int numparts = parts.size();
        int nummaterials = materials.size();

        stream->Write( &numframes );
        stream->Write( &numparts );
        stream->Write( &nummaterials );

        // frames
        for ( i = 0; i<numframes; i++ )
        {
                stream->WriteString( &frames[i]->name );

                numof = frames[i]->vertices.size();
                stream->Write ( &numof );
                for ( j = 0; j<numof; j++ )
                {
                        stream->Write ( &frames[i]->vertices[j].x );
                        stream->Write ( &frames[i]->vertices[j].y );
                        stream->Write ( &frames[i]->vertices[j].z );
                }

                numof = frames[i]->normals.size();
                stream->Write ( &numof );
                for ( j = 0; j<numof; j++ )
                {
                        stream->Write ( &frames[i]->normals[j].x );
                        stream->Write ( &frames[i]->normals[j].y );
                        stream->Write ( &frames[i]->normals[j].z );
                }

                numof = frames[i]->mappingcoords.size();
                stream->Write ( &numof );
                for ( j = 0; j<numof; j++ )
                {
                        stream->Write ( &frames[i]->mappingcoords[j].x );
                        stream->Write ( &frames[i]->mappingcoords[j].y );
                }

                numof = frames[i]->diffuses.size();
                stream->Write ( &numof );
                for ( j = 0; j<numof; j++ )
                {
                        stream->Write ( &frames[i]->diffuses[j].r );
                        stream->Write ( &frames[i]->diffuses[j].g );
                        stream->Write ( &frames[i]->diffuses[j].b );
                }

                numof = frames[i]->speculars.size();
                stream->Write ( &numof );
                for ( j = 0; j<numof; j++ )
                {
                        stream->Write ( &frames[i]->speculars[j].r );
                        stream->Write ( &frames[i]->speculars[j].g );
                        stream->Write ( &frames[i]->speculars[j].b );
                }

                numof = frames[i]->shininess.size();
                stream->Write ( &numof );
                for ( j = 0; j<numof; j++ )
                {
                        stream->Write ( &frames[i]->shininess[j] );
                }

                numof = frames[i]->alphas.size();
                stream->Write ( &numof );
                for ( j = 0; j<numof; j++ )
                {
                        stream->Write ( &frames[i]->alphas[j] );
                }
        }

        // materials
        for ( i = 0; i<nummaterials; i++ )
        {
                stream->WriteString( &materials[i]->name );
                stream->WriteString( &materials[i]->diffusefile );
                stream->Write ( &materials[i]->materialtype );
        }

        // parts
        for ( i = 0; i<numparts; i++ )
        {
                stream->WriteString( &parts[i]->name );

                stream->Write ( &parts[i]->nearplane );
                stream->Write ( &parts[i]->farplane );
                stream->Write ( &parts[i]->hbillboard );
                stream->Write ( &parts[i]->pbillboard );
                stream->Write ( &parts[i]->rbillboard );
                stream->Write ( &parts[i]->emission.r );
                stream->Write ( &parts[i]->emission.g );
                stream->Write ( &parts[i]->emission.b );
                stream->Write ( &parts[i]->doublefaced );
                stream->Write ( &parts[i]->visible );
                stream->Write ( &parts[i]->alpha );
                stream->Write ( &parts[i]->visiblealphathreshold );

                // bbox info
                stream->Write( &parts[i]->bbox.centerx );
                stream->Write( &parts[i]->bbox.centery );
                stream->Write( &parts[i]->bbox.centerz );
                stream->Write( &parts[i]->bbox.sizex );
                stream->Write( &parts[i]->bbox.sizey );
                stream->Write( &parts[i]->bbox.sizez );
                stream->Write( &parts[i]->bbox.sizexydiagonal );
                stream->Write( &parts[i]->bbox.sizediagonal );
                stream->Write( &parts[i]->bbox.minx );
                stream->Write( &parts[i]->bbox.miny );
                stream->Write( &parts[i]->bbox.minz );
                stream->Write( &parts[i]->bbox.maxx );
                stream->Write( &parts[i]->bbox.maxy );
                stream->Write( &parts[i]->bbox.maxz );

                numof = parts[i]->triangles.size();
                stream->Write ( &numof );
                for ( j = 0; j<numof; j++ )
                {
                        stream->Write ( &parts[i]->triangles[j].vertices[0] );
                        stream->Write ( &parts[i]->triangles[j].vertices[1] );
                        stream->Write ( &parts[i]->triangles[j].vertices[2] );
                        stream->Write ( &parts[i]->triangles[j].normals[0] );
                        stream->Write ( &parts[i]->triangles[j].normals[1] );
                        stream->Write ( &parts[i]->triangles[j].normals[2] );
                        stream->Write ( &parts[i]->triangles[j].mappingcoords[0] );
                        stream->Write ( &parts[i]->triangles[j].mappingcoords[1] );
                        stream->Write ( &parts[i]->triangles[j].mappingcoords[2] );
                        stream->Write ( &parts[i]->triangles[j].diffuse );
                        stream->Write ( &parts[i]->triangles[j].specular );
                        stream->Write ( &parts[i]->triangles[j].shininess );
                        stream->Write ( &parts[i]->triangles[j].alpha );
                        stream->Write ( &parts[i]->triangles[j].materialID );
                }
        }

        return true;
}

bool TMesh::LoadFromStream (    TGameIOStream * stream,
                                bool append_frame_only,
                                unsigned int frame_no,
                                float nearplane,
                                float farplane,
                                void * LoaderParameters )
{
        unsigned int signature = 0;
        int     version   = 0;

        stream->Read ( &signature );
        stream->Read ( &version   );

        if ( signature != 0x40a925fc )
                return false;

        istransparent   = -1;
        currentframe    = 0;
        targetframe     = 0;
        frametransition = 0;

        // general info
        unsigned int    duration;
        char            animloopmode;
        if ( version <= 2 ) stream->Read( &duration );
        stream->Read( &smoothframetransitions );
        stream->Read( &alpha );
        if ( version <= 2 ) stream->Read( &animloopmode );

        // bbox info
        stream->Read( &bbox.centerx );
        stream->Read( &bbox.centery );
        stream->Read( &bbox.centerz );
        stream->Read( &bbox.sizex );
        stream->Read( &bbox.sizey );
        stream->Read( &bbox.sizez );
        stream->Read( &bbox.sizexydiagonal );
        stream->Read( &bbox.sizediagonal );
        stream->Read( &bbox.minx );
        stream->Read( &bbox.miny );
        stream->Read( &bbox.minz );
        stream->Read( &bbox.maxx );
        stream->Read( &bbox.maxy );
        stream->Read( &bbox.maxz );

        int numframes    = 0;
        int numparts     = 0;
        int nummaterials = 0;

        stream->Read( &numframes );
        stream->Read( &numparts );
        stream->Read( &nummaterials );

        // get/create frames
        while ( GetNumOfFrames() < frame_no + numframes )
                AddFrame( );

        int i,j,v,numof;

        TMeshPart  * meshpart;
        TMeshFrame * meshframe;
        TVertex3f  * vert3f;
        TVertex2f  * vert2f;
        TColor3f   * col3f;
        TTriangle  * triangle;
        float      * shininess;
        float      * alpha;
        TMeshMaterial  * material;
        int              materialid;

        int start_vertex_index = 0;
        int start_norm_index   = 0;
        int start_mapco_index  = 0;
        int start_diff_index   = 0;
        int start_spec_index   = 0;
        int start_shin_index   = 0;
        int start_alph_index   = 0;
        int start_material_index = GetNumOfMaterials();

        // frames
        for ( i = 0; i<numframes; i++ )
        {
                meshframe = GetFrame( frame_no + i );
                if (meshframe)
                {
                        if ( i == 0 )
                        {
                                start_vertex_index = meshframe->GetNumOfVertices();
                                start_norm_index   = meshframe->GetNumOfNormals();
                                start_mapco_index  = meshframe->GetNumOfMappingCoords();
                                start_diff_index   = meshframe->GetNumOfDiffuses();
                                start_spec_index   = meshframe->GetNumOfSpeculars();
                                start_shin_index   = meshframe->GetNumOfShininess();
                                start_alph_index   = meshframe->GetNumOfAlphas();
                        }

                        stream->ReadString( &meshframe->name );

                        // add vertices
                        stream->Read ( &numof );
                        for ( v = 0; v < numof; v++ )
                        {
				vert3f     = meshframe->AddVertex ( );
                                stream->Read ( &vert3f->x );
                                stream->Read ( &vert3f->y );
                                stream->Read ( &vert3f->z );
                        }

                        // add normals
                        stream->Read ( &numof );
                        for ( v = 0; v < numof; v++ )
                        {
				vert3f     = meshframe->AddNormal ( );
                                stream->Read ( &vert3f->x );
                                stream->Read ( &vert3f->y );
                                stream->Read ( &vert3f->z );
                        }

                        // add mapping coords
                        stream->Read ( &numof );
                        for ( v = 0; v < numof; v++ )
                        {
				vert2f    = meshframe->AddMappingCoord ( );
                                stream->Read ( &vert2f->x );
                                stream->Read ( &vert2f->y );
                        }

                        // add diffuses
                        stream->Read ( &numof );
                        for ( v = 0; v < numof; v++ )
                        {
                                col3f    = meshframe->AddDiffuse ( );
                                stream->Read ( &col3f->r );
                                stream->Read ( &col3f->g );
                                stream->Read ( &col3f->b );
                        }

                        // add speculars
                        stream->Read ( &numof );
                        for ( v = 0; v < numof; v++ )
                        {
                                col3f    = meshframe->AddSpecular ( );
                                stream->Read ( &col3f->r );
                                stream->Read ( &col3f->g );
                                stream->Read ( &col3f->b );
                        }

                        // add shininess
                        stream->Read ( &numof );
                        for ( v = 0; v < numof; v++ )
                        {
                                shininess = meshframe->AddShininess ( );
                                stream->Read ( shininess );
                        }

                        // add alphas
                        stream->Read ( &numof );
                        for ( v = 0; v < numof; v++ )
                        {
                                alpha = meshframe->AddAlpha ( );
                                stream->Read ( alpha );
                        }
                }
        }

        // create mesh part
        if (!append_frame_only) // create a new part if appending new mesh
        {
                // add materials
                for ( i = 0; i<nummaterials; i++ )
                {
                        TMeshMaterial * material = AddMaterial ( );
                        stream->ReadString( &material->name );
                        stream->ReadString( &material->diffusefile );
                        stream->Read      ( &material->materialtype );
                }

                // add parts
                for ( i = 0; i<numparts; i++ )
                {
                        meshpart = AddPart( );

                        stream->ReadString( &meshpart->name );

                        stream->Read ( &meshpart->nearplane );
                        stream->Read ( &meshpart->farplane );
                        stream->Read ( &meshpart->hbillboard );
                        stream->Read ( &meshpart->pbillboard );
                        stream->Read ( &meshpart->rbillboard );
                        if ( version>1 )
                        {
                                stream->Read ( &meshpart->emission.r );
                                stream->Read ( &meshpart->emission.g );
                                stream->Read ( &meshpart->emission.b );
                        }
                        else
                        {
                                stream->Read ( &meshpart->emission.r );
                                meshpart->emission.g = meshpart->emission.r;
                                meshpart->emission.b = meshpart->emission.r;
                        }
                        stream->Read ( &meshpart->doublefaced );
                        stream->Read ( &meshpart->visible );
                        stream->Read ( &meshpart->alpha );
                        stream->Read ( &meshpart->visiblealphathreshold );

                        // bbox info
                        stream->Read ( &meshpart->bbox.centerx );
                        stream->Read ( &meshpart->bbox.centery );
                        stream->Read ( &meshpart->bbox.centerz );
                        stream->Read ( &meshpart->bbox.sizex );
                        stream->Read ( &meshpart->bbox.sizey );
                        stream->Read ( &meshpart->bbox.sizez );
                        stream->Read ( &meshpart->bbox.sizexydiagonal );
                        stream->Read ( &meshpart->bbox.sizediagonal );
                        stream->Read ( &meshpart->bbox.minx );
                        stream->Read ( &meshpart->bbox.miny );
                        stream->Read ( &meshpart->bbox.minz );
                        stream->Read ( &meshpart->bbox.maxx );
                        stream->Read ( &meshpart->bbox.maxy );
                        stream->Read ( &meshpart->bbox.maxz );

                        stream->Read ( &numof );
                        for ( j = 0; j<numof; j++ )
                        {
                                triangle = meshpart->AddTriangle ( );

                                stream->Read ( &triangle->vertices[0] ); triangle->vertices[0] += start_vertex_index;
                                stream->Read ( &triangle->vertices[1] ); triangle->vertices[1] += start_vertex_index;
                                stream->Read ( &triangle->vertices[2] ); triangle->vertices[2] += start_vertex_index;

                                stream->Read ( &triangle->normals [0] ); triangle->normals[0] += start_norm_index;
                                stream->Read ( &triangle->normals [1] ); triangle->normals[1] += start_norm_index;
                                stream->Read ( &triangle->normals [2] ); triangle->normals[2] += start_norm_index;

                                stream->Read ( &triangle->mappingcoords[0] ); triangle->mappingcoords[0] += start_mapco_index;
                                stream->Read ( &triangle->mappingcoords[1] ); triangle->mappingcoords[1] += start_mapco_index;
                                stream->Read ( &triangle->mappingcoords[2] ); triangle->mappingcoords[2] += start_mapco_index;

                                stream->Read ( &triangle->diffuse); triangle->diffuse += start_diff_index;
                                stream->Read ( &triangle->specular); triangle->specular += start_spec_index;
                                stream->Read ( &triangle->shininess); triangle->shininess += start_shin_index;
                                stream->Read ( &triangle->alpha); triangle->alpha += start_alph_index;

                                stream->Read ( &triangle->materialID ); triangle->materialID += start_material_index;
                        }
                }
        }
        else
        {
                UpdateBBox();
        }

        return true;
}

bool TMesh::SaveToFile   ( mstring fname )
{
        TGameFileStream stream;
        if ( stream.OpenFileForCreate( fname ) )
                SaveToStream ( &stream );
                else
                return false;
        return true;
}

bool TMesh::IsTransparent ( )
{
        if (alpha<1)
                return true;
                else
        if (istransparent == 1)
                return true;
                else
        if (istransparent == 0)
                return false;

        int numparts = parts.size();
        for ( int i = 0; i<numparts; i++ )
                if ( parts[i]->IsTransparent() )
                {
                        istransparent = 1;
                        return true;
                }

        istransparent = 0;
        return false;
}

void TMesh::UpdateBBoxFromSubParts  ( )
{
        bool first = true;
        float minx = 0, miny = 0, minz = 0, maxx = 0, maxy = 0, maxz = 0;

        for ( int i = 0; i<parts.size(); i++ )
        {
                if ( first || parts[i]->bbox.minx<minx ) minx = parts[i]->bbox.minx;
                if ( first || parts[i]->bbox.miny<miny ) miny = parts[i]->bbox.miny;
                if ( first || parts[i]->bbox.minz<minz ) minz = parts[i]->bbox.minz;
                if ( first || parts[i]->bbox.maxx>maxx ) maxx = parts[i]->bbox.maxx;
                if ( first || parts[i]->bbox.maxy>maxy ) maxy = parts[i]->bbox.maxy;
                if ( first || parts[i]->bbox.maxz>maxz ) maxz = parts[i]->bbox.maxz;
                first = false;
        }

        bbox.minx = minx;
        bbox.miny = miny;
        bbox.minz = minz;
        bbox.maxx = maxx;
        bbox.maxy = maxy;
        bbox.maxz = maxz;

        bbox.centerx = (minx+maxx)/2;
        bbox.centery = (miny+maxy)/2;
        bbox.centerz = (minz+maxz)/2;

        bbox.sizex   = fabs(minx-maxx);
        bbox.sizey   = fabs(miny-maxy);
        bbox.sizez   = fabs(minz-maxz);

        double   r = bbox.sizex*bbox.sizex + bbox.sizey*bbox.sizey;
        if (r<0) r = 0;
        bbox.sizexydiagonal = sqrt(r);

        r += bbox.sizez*bbox.sizez;
        if (r<0) r = 0;
        bbox.sizediagonal = sqrt(r);
}

void TMesh::UpdateBBox        ( )
{
        for ( int i = 0; i<parts.size(); i++ )
                parts[i]->UpdateBBox();
        UpdateBBoxFromSubParts  ( );
}

TMeshBoundingBox * TMesh::GetBBox ( )
{
        return &bbox;
}

float TMesh::GetAlpha          ( )
{
        return alpha;
}

void *  TMesh::GetData           ( )
{
        return data;
}
void TMesh::SetAlpha          ( float palpha )
{
        alpha         = palpha;
}

void TMesh::SetData           ( void * pdata )
{
        data = pdata;
}

unsigned int TMesh::GetNumOfParts     ( )
{
        return parts.size();
}

TMeshPart *  TMesh::GetPart           ( unsigned int index )
{
        if (index>=parts.size()) return NULL;
        return parts[index];
}

TMeshPart *  TMesh::AddPart           ( unsigned int * index )
{
        TMeshPart * p = new TMeshPart();
        p->mesh = this;
        if (index) *index = parts.size();
        parts.push_back(p);
        return p;
}

void  TMesh::DeletePart        ( unsigned int index )
{
        if (index>=parts.size()) return;
        delete parts[index];
		parts.erase ( parts.begin()+index );
}

unsigned int  TMesh::GetNumOfMaterials ( )
{
        return materials.size();
}

TMeshMaterial  *  TMesh::GetMaterial       ( unsigned int index )
{
        if (index>=materials.size()) return NULL;
        return materials[index];
}

TMeshMaterial  *  TMesh::AddMaterial       ( unsigned int * index )
{
        TMeshMaterial * p = new TMeshMaterial();
        p->mesh = this;
        if (index) *index = materials.size();
        materials.push_back(p);
        return p;
}

void   TMesh::DeleteMaterial    ( unsigned int index )
{
        if (index>=materials.size()) return;
        delete materials[index];
		materials.erase ( materials.begin()+index );
}

TMeshFrame * TMesh::AddFrame          ( unsigned int * index )
{
        TMeshFrame * p = new TMeshFrame();
        p->mesh = this;
        if (index) *index = frames.size();
        frames.push_back(p);
        return p;
}

void TMesh::DeleteFrame       ( unsigned int index )
{
        if (index>=frames.size()) return;
        delete frames[index];
		frames.erase ( frames.begin()+index );
}

mstring TMesh::GetFullFilePath   ( )
{
        return mExtractPathOfFile(filename);
}

mstring TMesh::GetFullFileName   ( )
{
        return filename;
}

mstring TMesh::GetOnlyFileName   ( )
{
        return mExtractNameOfFile(filename);
}

void TMesh::SetFileName       ( mstring fname )
{
        filename = fname;
}

void TMesh::FitInto ( bool proportional, float minx, float miny, float minz, float maxx, float maxy, float maxz )
{
        float tmp;
        if ( minx>maxx ) { tmp = minx; minx = maxx; maxx = tmp; }
        if ( miny>maxy ) { tmp = miny; miny = maxy; maxy = tmp; }
        if ( minz>maxz ) { tmp = minz; minz = maxz; maxz = tmp; }

        float sizex   = maxx-minx;
        float sizey   = maxy-miny;
        float sizez   = maxz-minz;

        float centerx = (minx+maxx)/2;
        float centery = (miny+maxy)/2;
        float centerz = (minz+maxz)/2;

        float scalex  = 1;
        float scaley  = 1;
        float scalez  = 1;

        if ( bbox.sizex>0 ) scalex = sizex/bbox.sizex; else scalex = 999999999;
        if ( bbox.sizey>0 ) scaley = sizey/bbox.sizey; else scaley = 999999999;
        if ( bbox.sizez>0 ) scalez = sizez/bbox.sizez; else scalez = 999999999;

        if ( proportional )
        {
                if ( scalex<scaley )
                {
                        if ( scalex<scalez )
                        {
                                scaley = scalex;
                                scalez = scalex;
                        }
                        else
                        {
                                scalex = scalez;
                                scaley = scalez;
                        }
                }
                else
                if ( scaley<scalez )
                {
                        scalex = scaley;
                        scalez = scaley;
                }
                else
                {
                        scalex = scalez;
                        scaley = scalez;
                }
        }

        UpdateBBox ( );

        for ( int f = 0; f<frames.size(); f++ )
        {
                for ( int v = 0; v<frames[f]->vertices.size(); v++ )
                {
                        frames[f]->vertices[v].x = (frames[f]->vertices[v].x - bbox.centerx)*scalex + centerx;
                        frames[f]->vertices[v].y = (frames[f]->vertices[v].y - bbox.centery)*scaley + centery;
                        frames[f]->vertices[v].z = (frames[f]->vertices[v].z - bbox.centerz)*scalez + centerz;
                }
        }

        UpdateBBox ( );
}

//********************************************************
// TMeshLoader
//********************************************************

TMeshLoader::TMeshLoader ()
{
}

TMeshLoader::~TMeshLoader ()
{
        ClearAll();
}

void TMeshLoader::ClearAll()
{
        while ( meshes.size()>0 )
                DeleteMeshByIndex(0);
}

TMesh  * TMeshLoader::LinkMesh( mstring fname, bool clone, bool clone_texture,
                                float nearplane, float farplane,
                                void * LoaderParameters,
                                mstring rootpath )
{
        TMesh * mesh;
        mstring CurPath =  mGetCurrentDir();
        if (fname.find(":")==-1 && fname.find("\\\\")==-1) fname = rootpath+"\\"+fname;

        // set currrent directory as the model directory
        mSetCurrentDir( mExtractPathOfFile(fname) );
        try
        {
                mstring file = mExtractNameOfFile(fname);
                fname = mGetCurrentDir()+"\\"+mUpperCase(file);

                // Search if the file is already loaded
                if (!clone)
                {
                        int i;
                        for (i=0; i<meshes.size(); i++)
                                if (meshes[i]->filename==fname && !meshes[i]->clone)
                                {
                                        meshes[i]->numoflinks++;
                                        mSetCurrentDir( CurPath );
                                        return meshes[i];
                                }
                }

                // if not loaded then load it
                mesh = new TMesh;
                if (mesh)
                {
                                mesh->filename     = fname;
                                mesh->numoflinks   = 1;
                                mesh->clone        = clone;
                                mesh->clonetexture = clone_texture;

                                mesh->LoadFromFile( file,
                                                    false, 0, nearplane, farplane,
                                                    LoaderParameters );

                                meshes.push_back(mesh);
                }
        }
        __finally
        {
                mSetCurrentDir( CurPath );
        }
        return mesh;
}

void  TMeshLoader::UnlinkMesh (TMesh  * mesh)
{
        int i;
        for (i=0; i<meshes.size(); i++)
                if (meshes[i]==mesh)
                {
                        if (meshes[i]->numoflinks<=1)
                                DeleteMeshByIndex(i);
                                else
                                meshes[i]->numoflinks--;
                        return;
                }
}

void  TMeshLoader::DeleteMeshByIndex (int index)
{
        if (index<0 || index>=meshes.size()) return;
        delete meshes[index];
		meshes.erase ( meshes.begin()+index );
}

void  TMeshLoader::DeleteMesh (TMesh * mesh)
{
        int i;
        for (i=0; i<meshes.size(); i++)
                if (meshes[i]==mesh)
                {
                        DeleteMeshByIndex(i);
                        return;
                }
}

void  TMeshLoader::DetachMesh           ( TMesh * mesh )
{
        int i;
        for (i=0; i<meshes.size(); i++)
                if (meshes[i]==mesh)
                {
					meshes.erase ( meshes.begin()+i );
                        return;
                }
}


