#include "MeshLoader.h"

void TMesh::LoadMaterials  ( )
{
        for ( int i = 0; i<materials.size(); i++ )
                materials[i]->LoadMaterial();
}

void TMesh::AdvanceMaterialAnims ( )
{
        for ( int m = 0; m< materials.size(); m++ )
        {
                if ( materials[m] && materials[m]->diffusetexture )
                        materials[m]->diffusetexture->AdvanceAnimation ( );
        }
}

void TMeshPart::Render            ( float camh, float camp, float camr,
                                    float objh, float objp, float objr,
                                    float camrange,
                                    float globalalpha,
                                    bool  hasanytransparent,
                                    int  &prevtexture_enabled,
                                    TTexture * &prevtexture,
                                     bool notexture,
                                     bool usefixedcolor,
                                     TColor3f fixedcolor,
                                     bool disablebackfaceculling )
{
        // current and target frames
        TMeshFrame * sframe;
        TMeshFrame * eframe;

        bool  has_billboard;
        float emis[4] = {1,1,1,1};
        float spec[4] = {1,1,1,1};
        float diff[4] = {1,1,1,1};
        float allonediff[4] = {1,1,1,1};
        TTriangle  * tri;
        TMeshMaterial * material;
        TColor3f  * diffuse, * specular;
        int t, numtris, snumvertices, snumnormals, snummappings;
        int snumalphas, enumvertices, enumnormals, enummappings, enumalphas;
        float partalpha, shininess;
        float frametransition;
        bool smoothframetransitions;

        TVertex3f * sv3f, * ev3f;
        TVertex2f * sv2f, * ev2f;
        float vx1, vy1, vz1, vx2, vy2, vz2, vx3, vy3, vz3;
        float nx1, ny1, nz1, nx2, ny2, nz2, nx3, ny3, nz3;
        float mx1, my1, mx2, my2, mx3, my3;

        if ( usefixedcolor )
        {
                allonediff[0] = fixedcolor.r;
                allonediff[1] = fixedcolor.g;
                allonediff[2] = fixedcolor.b;
        }
        else
        {
                allonediff[0] = 1.0;
                allonediff[1] = 1.0;
                allonediff[2] = 1.0;
        }

        if ( camrange >= nearplane &&
             camrange < farplane &&
             visible &&
             alpha>0 )
        {
                // set alpha threshold of part
                glAlphaFunc( GL_GREATER, visiblealphathreshold * globalalpha );

                // compute alpha value of part
                float globalpartalpha = globalalpha * alpha;

                // set depth mask of part
                if ( hasanytransparent )
                {
                        if ( IsTransparent() )
                                glDepthMask  ( 0 );
                                else
                                glDepthMask  ( 1 );
                }
                else
                glDepthMask  ( 1 );

                // set face culling of part
                if ( disablebackfaceculling || doublefaced )
                        glDisable ( GL_CULL_FACE );
                        else
                        glEnable  ( GL_CULL_FACE );

                // get current and destination frame
                sframe = mesh->GetCurrentFrame();
                if ( !sframe ) return;
                eframe = mesh->GetTargetFrame();
                if ( !eframe ) return;

                // get frame transition
                frametransition = mesh->frametransition;

                // is frame transitions are smooth
                smoothframetransitions = mesh->smoothframetransitions;

                // set billboard configuration
                if ( hbillboard || pbillboard || rbillboard )
                {
                        glPushMatrix ( );

                        glRotatef    ( -objr, 0, 1, 0 );
                        glRotatef    ( -objp, 1, 0, 0 );
                        glRotatef    (  objh, 0, 0, 1 );

                        if (!hbillboard) glRotatef ( -objh, 0, 0, 1 );
                        if (!pbillboard) glRotatef (  objp, 1, 0, 0 );
                        if (!rbillboard) glRotatef (  objr, 0, 1, 0 );

                        if (hbillboard) glRotatef  ( -camh,    0, 0, 1 );
                        if (pbillboard) glRotatef  (  camp,    1, 0, 0 );
                        if (rbillboard) glRotatef  (  camr,    0, 1, 0 );
                        has_billboard = true;
                }
                else
                {
                        has_billboard = false;
                }

                // set emission
                emis[0] = emission.r;
                emis[1] = emission.g;
                emis[2] = emission.b;
                emis[3] = 1;
                glMaterialfv( GL_FRONT_AND_BACK,GL_EMISSION, emis );

                // set specular
                spec[0] = 0;
                spec[1] = 0;
                spec[2] = 0;
                glMaterialfv( GL_FRONT_AND_BACK,GL_SPECULAR,spec );

                // set shininess
                shininess = 0;
                glMaterialf ( GL_FRONT_AND_BACK,GL_SHININESS, shininess );

                //////////////////////////////////////////////////////
                // render triangles
                //////////////////////////////////////////////////////
                glBegin ( GL_TRIANGLES );
                numtris = triangles.size();
                snumvertices = sframe->vertices.size();
                snumnormals  = sframe->normals.size();
                snummappings = sframe->mappingcoords.size();
                snumalphas   = sframe->alphas.size();
                enumvertices = eframe->vertices.size();
                enumnormals  = eframe->normals.size();
                enummappings = eframe->mappingcoords.size();
                enumalphas   = eframe->alphas.size();
                for ( t = 0; t<numtris; t++ )
                {
                        tri = &triangles[t];
                        if (    tri->vertices[0]<snumvertices &&
                                tri->vertices[1]<snumvertices &&
                                tri->vertices[2]<snumvertices &&
                                tri->normals [0]<snumnormals &&
                                tri->normals [1]<snumnormals &&
                                tri->normals [2]<snumnormals &&
                                tri->mappingcoords [0]<snummappings &&
                                tri->mappingcoords [1]<snummappings &&
                                tri->mappingcoords [2]<snummappings &&
                                tri->alpha<snumalphas &&
                                ( frametransition == 0 ||
                                  (
                                    tri->vertices[0]<enumvertices &&
                                    tri->vertices[1]<enumvertices &&
                                    tri->vertices[2]<enumvertices &&
                                    tri->normals [0]<enumnormals &&
                                    tri->normals [1]<enumnormals &&
                                    tri->normals [2]<enumnormals &&
                                    tri->mappingcoords [0]<enummappings &&
                                    tri->mappingcoords [1]<enummappings &&
                                    tri->mappingcoords [2]<enummappings &&
                                    tri->alpha<enumalphas
                                  )
                                )
                           )
                        {
                                if ( !smoothframetransitions ||
                                     frametransition == 0 ||
                                     eframe == sframe )
                                {
                                        // vertex 1
                                        sv3f = &sframe->vertices[ tri->vertices[0] ];
                                        vx1 = sv3f->x;
                                        vy1 = sv3f->y;
                                        vz1 = sv3f->z;
                                        // vertex 2
                                        sv3f = &sframe->vertices[ tri->vertices[1] ];
                                        vx2 = sv3f->x;
                                        vy2 = sv3f->y;
                                        vz2 = sv3f->z;
                                        // vertex 3
                                        sv3f = &sframe->vertices[ tri->vertices[2] ];
                                        vx3 = sv3f->x;
                                        vy3 = sv3f->y;
                                        vz3 = sv3f->z;
                                        // normal 1
                                        sv3f = &sframe->normals[ tri->normals[0] ];
                                        nx1 = sv3f->x;
                                        ny1 = sv3f->y;
                                        nz1 = sv3f->z;
                                        // normal 2
                                        sv3f = &sframe->normals[ tri->normals[1] ];
                                        nx2 = sv3f->x;
                                        ny2 = sv3f->y;
                                        nz2 = sv3f->z;
                                        // normal 3
                                        sv3f = &sframe->normals[ tri->normals[2] ];
                                        nx3 = sv3f->x;
                                        ny3 = sv3f->y;
                                        nz3 = sv3f->z;
                                }
                                else
                                {
                                        // vertex 1
                                        sv3f = &sframe->vertices[ tri->vertices[0] ];
                                        ev3f = &eframe->vertices[ tri->vertices[0] ];
                                        vx1 = sv3f->x + (ev3f->x-sv3f->x)*frametransition;
                                        vy1 = sv3f->y + (ev3f->y-sv3f->y)*frametransition;
                                        vz1 = sv3f->z + (ev3f->z-sv3f->z)*frametransition;
                                        // vertex 2
                                        sv3f = &sframe->vertices[ tri->vertices[1] ];
                                        ev3f = &eframe->vertices[ tri->vertices[1] ];
                                        vx2 = sv3f->x + (ev3f->x-sv3f->x)*frametransition;
                                        vy2 = sv3f->y + (ev3f->y-sv3f->y)*frametransition;
                                        vz2 = sv3f->z + (ev3f->z-sv3f->z)*frametransition;
                                        // vertex 3
                                        sv3f = &sframe->vertices[ tri->vertices[2] ];
                                        ev3f = &eframe->vertices[ tri->vertices[2] ];
                                        vx3 = sv3f->x + (ev3f->x-sv3f->x)*frametransition;
                                        vy3 = sv3f->y + (ev3f->y-sv3f->y)*frametransition;
                                        vz3 = sv3f->z + (ev3f->z-sv3f->z)*frametransition;
                                        // normal 1
                                        sv3f = &sframe->normals[ tri->normals[0] ];
                                        ev3f = &eframe->normals[ tri->normals[0] ];
                                        nx1 = sv3f->x + (ev3f->x-sv3f->x)*frametransition;
                                        ny1 = sv3f->y + (ev3f->y-sv3f->y)*frametransition;
                                        nz1 = sv3f->z + (ev3f->z-sv3f->z)*frametransition;
                                        // normal 2
                                        sv3f = &sframe->normals[ tri->normals[1] ];
                                        ev3f = &eframe->normals[ tri->normals[1] ];
                                        nx2 = sv3f->x + (ev3f->x-sv3f->x)*frametransition;
                                        ny2 = sv3f->y + (ev3f->y-sv3f->y)*frametransition;
                                        nz2 = sv3f->z + (ev3f->z-sv3f->z)*frametransition;
                                        // normal 3
                                        sv3f = &sframe->normals[ tri->normals[2] ];
                                        ev3f = &eframe->normals[ tri->normals[2] ];
                                        nx3 = sv3f->x + (ev3f->x-sv3f->x)*frametransition;
                                        ny3 = sv3f->y + (ev3f->y-sv3f->y)*frametransition;
                                        nz3 = sv3f->z + (ev3f->z-sv3f->z)*frametransition;
                                }

                                // mapping coords 1
                                sv2f = &sframe->mappingcoords[ tri->mappingcoords[0] ];
                                mx1 = sv2f->x;
                                my1 = sv2f->y;
                                // mapping coords 2
                                sv2f = &sframe->mappingcoords[ tri->mappingcoords[1] ];
                                mx2 = sv2f->x;
                                my2 = sv2f->y;
                                // mapping coords 3
                                sv2f = &sframe->mappingcoords[ tri->mappingcoords[2] ];
                                mx3 = sv2f->x;
                                my3 = sv2f->y;

                                // get diffuse
                                diffuse = &sframe->diffuses[ tri->diffuse ];
                                // get alpha
                                partalpha = sframe->alphas[ tri->alpha ] * globalpartalpha;
                                // get specular
                                specular = &sframe->speculars[ tri->specular ];
                                // get shininess
                                shininess = sframe->shininess[ tri->shininess ];

                                // get material
                                if ( tri->materialID >= 0 &&
                                     tri->materialID < mesh->materials.size() )
                                        material = mesh->materials[ tri->materialID ];
                                        else
                                        material = NULL;

                                // set texture
                                if ( !notexture && material  &&
                                     material->diffusetexture &&
                                     material->diffusetexture->GetNumOfFrames()>0 )
                                {
                                        // if has texture
                                        if (prevtexture != material->diffusetexture)
                                        {
                                                glEnd ( );

                                                if (prevtexture_enabled != 1)
                                                {
                                                        prevtexture_enabled = 1;
                                                        glEnable ( GL_TEXTURE_2D );
                                                }
                                                prevtexture = material->diffusetexture;
                                                material->diffusetexture->BindTexture( false );
                                                glTexParameteri( GL_TEXTURE_2D,  GL_TEXTURE_WRAP_S, GL_REPEAT );
                                                glTexParameteri( GL_TEXTURE_2D,  GL_TEXTURE_WRAP_T, GL_REPEAT );

                                                glBegin ( GL_TRIANGLES );
                                        }

                                        // set color & alpha
                                        glMaterialfv ( GL_FRONT_AND_BACK, GL_AMBIENT, allonediff );
                                        glMaterialfv ( GL_FRONT_AND_BACK, GL_DIFFUSE, allonediff );
                                        glColor4f    ( allonediff[0],allonediff[1],allonediff[2], partalpha );
                                }
                                else
                                {
                                        // if does not have texture
                                        if (prevtexture_enabled != 0)
                                        {
                                                glEnd ( );

                                                prevtexture_enabled = 0;
                                                glDisable ( GL_TEXTURE_2D );

                                                glBegin ( GL_TRIANGLES );
                                        }
                                        prevtexture = NULL;

                                        // set color & alpha
                                        if ( usefixedcolor )
                                        {
                                                glMaterialfv ( GL_FRONT_AND_BACK, GL_AMBIENT, allonediff );
                                                glMaterialfv ( GL_FRONT_AND_BACK, GL_DIFFUSE, allonediff );
                                                glColor4f    ( allonediff[0],allonediff[1],allonediff[2], partalpha );
                                        }
                                        else
                                        {
                                                diff[0] = diffuse->r;
                                                diff[1] = diffuse->g;
                                                diff[2] = diffuse->b;
                                                glMaterialfv ( GL_FRONT_AND_BACK, GL_AMBIENT, diff );
                                                glMaterialfv ( GL_FRONT_AND_BACK, GL_DIFFUSE, diff );
                                                glColor4f    ( diff[0], diff[1], diff[2], partalpha );
                                        }
                                }

                                /*
                                // set specular
                                spec[0] = specular->r;
                                spec[1] = specular->g;
                                spec[2] = specular->b;
                                glMaterialfv( GL_FRONT_AND_BACK,GL_SPECULAR,spec );

                                // set shininess
                                glMaterialf ( GL_FRONT_AND_BACK,GL_SHININESS, shininess );
                                */

                                // set vertex 1
                                glTexCoord2d ( mx1, my1 );
                                glNormal3f   ( nx1, ny1, nz1 );
                                glVertex3f   ( vx1, vy1, vz1 );

                                // set vertex 2
                                glTexCoord2d ( mx2, my2 );
                                glNormal3f   ( nx2, ny2, nz2 );
                                glVertex3f   ( vx2, vy2, vz2 );

                                // set vertex 3
                                glTexCoord2d ( mx3, my3 );
                                glNormal3f   ( nx3, ny3, nz3 );
                                glVertex3f   ( vx3, vy3, vz3 );
                        }
                }
                glEnd   ( );

                if (has_billboard)
                        glPopMatrix ( );
        }
}

void TMesh::Render ( float camh, float camp, float camr, float objh, float objp, float objr,
                     float camrange, float globalalpha,
                     int  &prevtexture_enabled,
                     TTexture * &prevtexture,
                     bool notexture,
                     bool usefixedcolor,
                     TColor3f fixedcolor,
                     bool disablebackfaceculling )
{
        // store GL states
        GLboolean enabledCullFace = glIsEnabled   ( GL_CULL_FACE );
        GLboolean enabledText2D   = glIsEnabled   ( GL_TEXTURE_2D );
        GLboolean enabledBlend    = glIsEnabled   ( GL_BLEND );
        GLboolean enabledBack     = glIsEnabled   ( GL_BACK );
        GLboolean enabledColMat   = glIsEnabled   ( GL_COLOR_MATERIAL );
        GLboolean depthmask;        glGetBooleanv ( GL_DEPTH_WRITEMASK, &depthmask );

        // set constant GL state
        glEnable        ( GL_BLEND );
        glEnable        ( GL_COLOR_MATERIAL );
        glColorMaterial ( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE );
        glCullFace      ( GL_BACK  );
        glBlendFunc     ( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );

        // load & advance material and frame animations
        LoadMaterials        ( );
        AdvanceMaterialAnims ( );

        // render parts
        bool hasanytransparent   = IsTransparent ( );
        for ( int p = 0; p<parts.size(); p++ )
                parts[p]->Render ( camh, camp, camr, objh, objp, objr,
                                   camrange,
                                   globalalpha*alpha,
                                   hasanytransparent,
                                   prevtexture_enabled,
                                   prevtexture,
                                   notexture,
                                   usefixedcolor,
                                   fixedcolor,
                                   disablebackfaceculling );

        // restore GL states
        if ( enabledCullFace )
                glEnable ( GL_CULL_FACE );
                else
                glDisable ( GL_CULL_FACE );
        if ( enabledText2D )
                glEnable ( GL_TEXTURE_2D );
                else
                glDisable ( GL_TEXTURE_2D );
        if ( enabledBlend )
                glEnable ( GL_BLEND );
                else
                glDisable ( GL_BLEND );
        if ( enabledBack )
                glEnable ( GL_BACK );
                else
                glDisable ( GL_BACK );
        if ( enabledColMat )
                glEnable ( GL_COLOR_MATERIAL );
                else
                glDisable ( GL_COLOR_MATERIAL );
        if ( depthmask )
                glDepthMask ( 1 );
                else
                glDepthMask ( 0 );
        glBindTexture  ( GL_TEXTURE_2D, 0 );
        glAlphaFunc    ( GL_ALWAYS, 0 );
        glColor4f      ( 1, 1, 1, 1 );

        // not restored GL states
        /*
        glColorMaterial ( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE );

        glBlendFunc  ( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );

        glMaterialfv ( GL_FRONT_AND_BACK, GL_EMISSION, emis );
        glMaterialfv ( GL_FRONT_AND_BACK, GL_SPECULAR, spec );
        glMaterialf  ( GL_FRONT_AND_BACK, GL_SHININESS, shininess );
        glMaterialfv ( GL_FRONT_AND_BACK, GL_AMBIENT, allonediff );
        glMaterialfv ( GL_FRONT_AND_BACK, GL_DIFFUSE, allonediff );

        glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT );
        glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT );
        */
}

