#include "camera_controller.h"

#include <math.h>

//******************************************************************************
// TCameraController
//******************************************************************************

TCameraController::TCameraController ()
{
	cameramotion   = cameramotionNone;

	leftfunction   = mousefunc3dRotate;
	middlefunction = mousefunc3dPan;
	rightfunction  = mousefunc3dForwardBackward;

        mousebuttons [0] = false;
        mousebuttons [1] = false;
        mousebuttons [2] = false;

	camx	= 0;
	camy	= 0;
        camz	= 0;
        camh	= 0;
        camp	= 0;
        camr	= 0;
        camvfov = 50;
        camfarplane = 3000;
        speed       = 0;

	prevtickcount = -1;
	prevfrmtime   = 0;
        zerotime      = 0;

        panspeed 	= 0.3;
        rotatespeed 	= 0.1;
        accelerationpersec = 100;
        rotationpersec     = 0;

        screenwidth	= 1;
        screenheight	= 1;

	data			= NULL;

        LeftDownEvent		= NULL;
        RightDownEvent		= NULL;
        MiddleDownEvent		= NULL;
        MoveEvent		= NULL;
        LeftMoveEvent		= NULL;
        RightMoveEvent		= NULL;
        MiddleMoveEvent		= NULL;
        LeftUpEvent		= NULL;
	RightUpEvent		= NULL;
        MiddleUpEvent		= NULL;
        SetCaptureControlEvent	= NULL;
}

TCameraController::~TCameraController()
{
}

// ------------------------------ Public Functions ---------------------------

float TCameraController::GetAccelerationPerSec ( )
{
        return accelerationpersec;
}

float TCameraController::GetRotationalSpeed ( )
{
        return rotationpersec;
}

float TCameraController::GetPanPerPixel 	    ( )
{
        return panspeed;
}

float TCameraController::GetRotatePerPixel     ( )
{
        return rotatespeed;
}

void   TCameraController::SetPanPerPixel 	    ( float spd )
{
	panspeed = spd;
}

void   TCameraController::SetRotatePerPixel     ( float spd )
{
	rotatespeed = spd;
}

void   TCameraController::SetAccelerationPerSec( float spd )
{
	accelerationpersec = spd;
}

void   TCameraController::SetRotationalSpeed( float spd )
{
	rotationpersec = spd;
}

void TCameraController::SetScreenSize  ( int wd, int hd )
{
	if ( wd<1 )  wd = 1;
	if ( hd<1 )  hd = 1;
        screenwidth  = wd;
        screenheight = hd;
}

void TCameraController::SetMouseFunction ( int Button, TMouseFunction3DMode mmode )
{
	switch ( Button )
        {
        case 0: leftfunction = mmode; break;
        case 1: rightfunction = mmode; break;
        case 2: middlefunction = mmode; break;
        }
}

float TCameraController::GetCamX    ( )
{
	return camx;
}

float TCameraController::GetCamY    ( )
{
	return camy;
}

float TCameraController::GetCamZ ( )
{
	return camz;
}

float  TCameraController::GetCamH      ( )
{
	return camh;
}

float  TCameraController::GetCamP      ( )
{
	return camp;
}

float  TCameraController::GetCamR      ( )
{
	return camr;
}

float  TCameraController::GetCamVFov   ( )
{
	return camvfov;
}

float  TCameraController::GetCamFarPlane   ( )
{
	return camfarplane;
}

float  TCameraController::GetSpeed   ( )
{
	return speed;
}

float  TCameraController::GetCamHFov   ( )
{
	float  camhfov = 2 * atan ( screenwidth*tan( camvfov/180*PI/2 )/screenheight ) ;
        return camhfov*180/PI;
}

void   TCameraController::SetCamCoord       ( float cx, float cy, float cz )
{
	camx = cx;
        camy = cy;
        camz   = cz;
}

void   TCameraController::SetCamOrientation ( float ch, float cp, float cr )
{
	camh = fmod(ch,360);
        camp = fmod(cp,360);
        camr = fmod(cr,360);
}

void   TCameraController::SetCamVFov        ( float cvfov )
{
	if ( cvfov<1   ) cvfov = 1;
	if ( cvfov>135 ) cvfov = 135;
	camvfov = cvfov;
}

void   TCameraController::SetCamFarPlane    ( float cfarplane )
{
        if ( cfarplane<1 ) cfarplane = 1;
	camfarplane = cfarplane;
}

void   TCameraController::SetSpeed    ( float spd )
{
	speed = spd;
}

int    TCameraController::GetLastMouseX   ( )
{
	return lastmousex;
}

int    TCameraController::GetLastMouseY   ( )
{
	return lastmousey;
}

void TCameraController::OnMouseDown( int Button, bool ShiftDown, bool AltDown, bool CtrlDown, bool DoubleClicked, int X, int Y )
{
        SetCaptureControl();

        mousebuttons [Button]   = true;

	firstmousex  [Button]   = X;
        firstmousey  [Button]   = Y;
        lastmousex              = X;
        lastmousey              = Y;

        if (Button==0)
                OnLeftDown   ( X, Y );
                else
        if (Button==1)
                OnRightDown  ( X, Y );
                else
        if (Button==2)
                OnMiddleDown ( X, Y );
}

void TCameraController::OnMouseUp  ( int Button, bool ShiftDown, bool AltDown, bool CtrlDown, int X, int Y)
{
        mousebuttons [Button] = false;

        if ( !mousebuttons [0] && !mousebuttons [1] && !mousebuttons [2] )
	        UnsetCaptureControl();

        int DX  	= X-lastmousex;
        int DY  	= Y-lastmousey;
        lastmousex      = X;
        lastmousey      = Y;

        if (Button==0)
                OnLeftUp   ( firstmousex[Button], firstmousey[Button], DX, DY, X, Y );
                else
        if (Button==1)
                OnRightUp   ( firstmousex[Button], firstmousey[Button], DX, DY, X, Y );
                else
        if (Button==2)
                OnMiddleUp   ( firstmousex[Button], firstmousey[Button], DX, DY, X, Y );
}

void TCameraController::OnMouseMove( bool ShiftDown, bool AltDown, bool CtrlDown, int X, int Y )
{
        int DX  	= X-lastmousex;
        int DY  	= Y-lastmousey;
        lastmousex      = X;
        lastmousey      = Y;

	OnMove   ( DX, DY, X, Y );

        if (mousebuttons[0])
                OnLeftMove   ( firstmousex[0], firstmousey[0], DX, DY, X, Y );

        if (mousebuttons[1])
                OnRightMove  ( firstmousex[1], firstmousey[1], DX, DY, X, Y );

        if (mousebuttons[2])
                OnMiddleMove ( firstmousex[2], firstmousey[2], DX, DY, X, Y );
}

void TCameraController::OnAbnormalMouseUp( )
{
        mousebuttons [0] = false;
        mousebuttons [1] = false;
        mousebuttons [2] = false;

        UnsetCaptureControl();
}

// ------------------------------ General On Mouse Downs ---------------------------

void TCameraController::OnLeftDown  ( int X, int Y )
{
	switch ( leftfunction )
        {
		case mousefunc3dPan:
                	break;
		case mousefunc3dRotate:
                	break;
		case mousefunc3dForwardBackward:
                	break;
		case mousefunc3dAccelerate:
                	break;
		case mousefunc3dDeaccelerate:
                	break;
		case mousefunc3dZeroSpeed:
                	DoDownZeroSpeed ( 0, X, Y );
                	break;
        }
        if ( LeftDownEvent )
       		LeftDownEvent ( data, X, Y );
}

void TCameraController::OnRightDown ( int X, int Y )
{
	switch ( rightfunction )
        {
		case mousefunc3dPan:
                	break;
		case mousefunc3dRotate:
                	break;
		case mousefunc3dForwardBackward:
                	break;
		case mousefunc3dAccelerate:
                	break;
		case mousefunc3dDeaccelerate:
                	break;
		case mousefunc3dZeroSpeed:
                	DoDownZeroSpeed ( 1, X, Y );
                	break;
        }
        if ( RightDownEvent )
       		RightDownEvent ( data, X, Y );
}

void TCameraController::OnMiddleDown( int X, int Y )
{
	switch ( middlefunction )
        {
		case mousefunc3dPan:
                	break;
		case mousefunc3dRotate:
                	break;
		case mousefunc3dForwardBackward:
                	break;
		case mousefunc3dAccelerate:
                	break;
		case mousefunc3dDeaccelerate:
                	break;
		case mousefunc3dZeroSpeed:
                	DoDownZeroSpeed ( 2, X, Y );
                	break;
        }
        if ( MiddleDownEvent )
       		MiddleDownEvent ( data, X, Y );
}

// ------------------------------ General On Mouse Moves ---------------------------

void TCameraController::OnMove      ( int dX, int dY, int X, int Y )
{
        if ( MoveEvent )
       		MoveEvent ( data, dX, dY, X, Y );
}

void TCameraController::OnLeftMove  ( int startX, int startY, int dX, int dY, int X, int Y )
{
	switch ( leftfunction )
        {
		case mousefunc3dPan:
                	DoMovePan ( 0, startX, startY, dX, dY, X, Y );
                	break;
		case mousefunc3dRotate:
                	DoMoveRotate ( 0, startX, startY, dX, dY, X, Y );
                	break;
		case mousefunc3dForwardBackward:
                	DoMoveForwardBackward ( 0, startX, startY, dX, dY, X, Y );
                	break;
		case mousefunc3dAccelerate:
                	break;
		case mousefunc3dDeaccelerate:
                	break;
		case mousefunc3dZeroSpeed:
                	break;
        }
        if ( LeftMoveEvent )
       		LeftMoveEvent ( data, startX, startY, dX, dY, X, Y );
}

void TCameraController::OnRightMove ( int startX, int startY, int dX, int dY, int X, int Y )
{
	switch ( rightfunction )
        {
		case mousefunc3dPan:
                	DoMovePan ( 1, startX, startY, dX, dY, X, Y );
                	break;
		case mousefunc3dRotate:
                	DoMoveRotate ( 1, startX, startY, dX, dY, X, Y );
                	break;
		case mousefunc3dForwardBackward:
                	DoMoveForwardBackward ( 1, startX, startY, dX, dY, X, Y );
                	break;
		case mousefunc3dAccelerate:
                	break;
		case mousefunc3dDeaccelerate:
                	break;
		case mousefunc3dZeroSpeed:
                	break;
        }
        if ( RightMoveEvent )
       		RightMoveEvent ( data, startX, startY, dX, dY, X, Y );
}

void TCameraController::OnMiddleMove( int startX, int startY, int dX, int dY, int X, int Y )
{
	switch ( middlefunction )
        {
		case mousefunc3dPan:
                	DoMovePan ( 2, startX, startY, dX, dY, X, Y );
                	break;
		case mousefunc3dRotate:
                	DoMoveRotate ( 2, startX, startY, dX, dY, X, Y );
                	break;
		case mousefunc3dForwardBackward:
                	DoMoveForwardBackward ( 2, startX, startY, dX, dY, X, Y );
                	break;
		case mousefunc3dAccelerate:
                	break;
		case mousefunc3dDeaccelerate:
                	break;
		case mousefunc3dZeroSpeed:
                	break;
        }
        if ( MiddleMoveEvent )
       		MiddleMoveEvent ( data, startX, startY, dX, dY, X, Y );
}

// ------------------------------ General On Mouse Ups ---------------------------

void TCameraController::OnLeftUp    ( int startX, int startY, int dX, int dY, int X, int Y )
{
	switch ( leftfunction )
        {
		case mousefunc3dPan:
                	break;
		case mousefunc3dRotate:
                	break;
		case mousefunc3dForwardBackward:
                	break;
		case mousefunc3dAccelerate:
                	break;
		case mousefunc3dDeaccelerate:
                	break;
		case mousefunc3dZeroSpeed:
                	break;
        }
        if ( LeftUpEvent )
       		LeftUpEvent ( data, startX, startY, dX, dY, X, Y );
}

void TCameraController::OnRightUp   ( int startX, int startY, int dX, int dY, int X, int Y )
{
	switch ( rightfunction )
        {
		case mousefunc3dPan:
                	break;
		case mousefunc3dRotate:
                	break;
		case mousefunc3dForwardBackward:
                	break;
		case mousefunc3dAccelerate:
                	break;
		case mousefunc3dDeaccelerate:
                	break;
		case mousefunc3dZeroSpeed:
                	break;
        }
        if ( RightUpEvent )
       		RightUpEvent ( data, startX, startY, dX, dY, X, Y );
}

void TCameraController::OnMiddleUp  ( int startX, int startY, int dX, int dY, int X, int Y )
{
	switch ( middlefunction )
        {
		case mousefunc3dPan:
                	break;
		case mousefunc3dRotate:
                	break;
		case mousefunc3dForwardBackward:
                	break;
		case mousefunc3dAccelerate:
                	break;
		case mousefunc3dDeaccelerate:
                	break;
		case mousefunc3dZeroSpeed:
                	break;
        }
        if ( MiddleUpEvent )
       		MiddleUpEvent ( data, startX, startY, dX, dY, X, Y );
}

// ------------------------------ Window Mouse Capture Functions ---------------------------

void TCameraController::SetCaptureControl  ()
{
        if ( SetCaptureControlEvent )
       		SetCaptureControlEvent ( data, true );
}

void TCameraController::UnsetCaptureControl()
{
        if ( SetCaptureControlEvent )
       		SetCaptureControlEvent ( data, false );
}

// ------------------------------ Mouse Pan Function ---------------------------

void TCameraController::DoMovePan ( int button, int startX, int startY, int dX, int dY, int X, int Y )
{
	MoveCamera  ( 0, -dX*panspeed, dY*panspeed );
}

void TCameraController::MoveCamera ( double ForwardInMeter, double RightInMeter, double UpInMeter )
{
	double dx = RightInMeter;
        double dy = ForwardInMeter;
        double dz = UpInMeter;

        VectorOp.Rotate_Y (  camr, &dx, &dz );
        VectorOp.Rotate_X (  camp, &dy, &dz );
        VectorOp.Rotate_Z ( -camh, &dx, &dy );

        SetCamCoord ( camx+dx, camy+dy, camz+dz );
}

// ------------------------------ Mouse Rotate Function ---------------------------

void TCameraController::DoMoveRotate ( int button, int startX, int startY, int dX, int dY, int X, int Y )
{
        SetCamOrientation ( camh - rotatespeed*dX, camp + rotatespeed*dY, camr );
}

// ------------------------------ Mouse ForwardBackward Function ---------------------------

void TCameraController::DoMoveForwardBackward ( int button, int startX, int startY, int dX, int dY, int X, int Y )
{
	MoveCamera          ( -dY*panspeed, 0, 0 );
        SetCamOrientation   ( camh + rotatespeed*dX, camp, camr );
}

// ------------------------------ Reset Speed Functions ------------------------

void TCameraController::DoDownZeroSpeed ( int button, int X, int Y )
{
	speed    = 0;
        zerotime = mGetTickCount();
}

// ------------------------------ Motion Functions ------------------------

TCameraMotionMode TCameraController::GetCameraMotion     ( )
{
	return cameramotion;
}

void   		  TCameraController::SetCameraMotion     ( TCameraMotionMode cmotion )
{
	cameramotion = cmotion;
        zerotime     = mGetTickCount();
}

void   TCameraController::AdvanceCameraMotion ( )
{
	long	tcount = mGetTickCount();
	if ( prevtickcount == -1 ) prevtickcount = tcount;
        float   tm     = (tcount - prevtickcount)/1000.0;	// conver to sec

        tm = 0.4*tm + 0.6*prevfrmtime;
        prevfrmtime = tm;

        prevtickcount  = tcount;

	if (    ( leftfunction == mousefunc3dAccelerate && mousebuttons[0] ) ||
        	( rightfunction == mousefunc3dAccelerate && mousebuttons[1] ) ||
        	( middlefunction == mousefunc3dAccelerate && mousebuttons[2] ) )
        {
	        speed += accelerationpersec*tm*1000.0/3600.0;
        }

	if (    ( leftfunction == mousefunc3dDeaccelerate && mousebuttons[0] ) ||
        	( rightfunction == mousefunc3dDeaccelerate && mousebuttons[1] ) ||
        	( middlefunction == mousefunc3dDeaccelerate && mousebuttons[2] ) )
        {
	        speed -= accelerationpersec*tm*1000.0/3600.0;
        }

        float hrotpersec, protpersec;

	double dx,dy,dz,rxy,ch;

        switch ( cameramotion )
        {
	case cameramotionFly:
                if (    ( leftfunction == mousefunc3dZeroSpeed && mousebuttons[0] ) ||
                        ( rightfunction == mousefunc3dZeroSpeed && mousebuttons[1] ) ||
                        ( middlefunction == mousefunc3dZeroSpeed && mousebuttons[2] ) ||
                        ( tcount - zerotime )<1500 )
                	break;
                hrotpersec = (lastmousex -  (screenwidth/2))*0.15;
                protpersec = ((screenheight/2) - lastmousey)*0.20;
        	SetCamOrientation ( GetCamH()+hrotpersec*tm, GetCamP()+protpersec*tm, GetCamR() );
        	break;
        }

        if ( speed != 0 )
		MoveCamera ( speed*tm*1000.0/3600.0, 0, 0 );

	if ( cameramotion == cameramotionNone && rotationpersec != 0 )
        	SetCamOrientation ( GetCamH()+rotationpersec*tm*1000.0/3600.0, GetCamP(), GetCamR() );
}

