#ifndef __GAME_ENVIRONMENT_SCREEN_CONTROLLER__
#define __GAME_ENVIRONMENT_SCREEN_CONTROLLER__

#include "main_functions.h"
#include "vector_operations.h"

#include <vector>

//******************************************************************************
// Classes
//******************************************************************************

class TCameraController;

//******************************************************************************
// Type Definitions
//******************************************************************************

typedef enum
{
	cameramotionNone, cameramotionFly
}
TCameraMotionMode;

typedef enum
{
	mousefunc3dNone, mousefunc3dPan, mousefunc3dRotate, mousefunc3dForwardBackward,
        mousefunc3dAccelerate, mousefunc3dDeaccelerate, mousefunc3dZeroSpeed
}
TMouseFunction3DMode;

//******************************************************************************
// Callback Function Definitions
//******************************************************************************

typedef void (*TMouse2Param3DProc)        ( void * data, int X, int Y );
typedef void (*TMouse4Param3DProc)        ( void * data, int dX, int dY, int X, int Y );
typedef void (*TMouse6Param3DProc)        ( void * data, int startX, int startY, int dX, int dY, int X, int Y );
typedef void (*TSetCaptureControl3DProc)  ( void * data, bool setcap );

//******************************************************************************
// TScreenController3D
//******************************************************************************

class TCameraController
{
protected:
	TMouseFunction3DMode leftfunction;
	TMouseFunction3DMode rightfunction;
	TMouseFunction3DMode middlefunction;

        TCameraMotionMode  cameramotion;

        bool mousebuttons  [3];	// 0 = left, 1 = right, 2 = middle
        int  firstmousex   [3];
        int  firstmousey   [3];
        int  lastmousex;
        int  lastmousey;

        int  screenwidth;
        int  screenheight;

        long   prevtickcount;
        float  prevfrmtime;
        long   zerotime;

        float speed;
        float panspeed;
        float rotatespeed;
        float accelerationpersec;
        float rotationpersec;

	float camx;
	float camy;
        float camz;
        float camh;
        float camp;
        float camr;
        float camvfov;
        float camfarplane;

	TMouseFunction3DMode   nowoperation;
        int		       nowbutton;

        // mouse buttons
	virtual void OnMove      ( int dX, int dY, int X, int Y );

	virtual void OnLeftDown  ( int X, int Y );
	virtual void OnRightDown ( int X, int Y );
	virtual void OnMiddleDown( int X, int Y );

	virtual void OnLeftMove  ( int startX, int startY, int dX, int dY, int X, int Y );
	virtual void OnRightMove ( int startX, int startY, int dX, int dY, int X, int Y );
	virtual void OnMiddleMove( int startX, int startY, int dX, int dY, int X, int Y );

	virtual void OnLeftUp    ( int startX, int startY, int dX, int dY, int X, int Y );
	virtual void OnRightUp   ( int startX, int startY, int dX, int dY, int X, int Y );
	virtual void OnMiddleUp  ( int startX, int startY, int dX, int dY, int X, int Y );

        virtual void SetCaptureControl  ();
        virtual void UnsetCaptureControl();

        // mouse functions
	virtual void DoMovePan       ( int button, int startX, int startY, int dX, int dY, int X, int Y );
	virtual void DoMoveRotate    ( int button, int startX, int startY, int dX, int dY, int X, int Y );
	virtual void DoMoveForwardBackward ( int button, int startX, int startY, int dX, int dY, int X, int Y );
	virtual void DoDownZeroSpeed       ( int button, int X, int Y );

public:
	TCameraController ( );
        virtual ~TCameraController( );

        // camera information

	float  GetCamX      ( );
	float  GetCamY      ( );
        float  GetCamZ      ( );

        float  GetCamH      ( );
        float  GetCamP      ( );
        float  GetCamR      ( );

        float  GetCamVFov   ( );
        float  GetCamHFov   ( );

        float  GetCamFarPlane ( );

	float  GetSpeed ( );                    // km/saat
        float  GetRotationalSpeed ( );	        // derece/sn

        virtual void   SetCamCoord       ( float cx, float cy, float cz );
        virtual void   SetCamOrientation ( float ch, float cp, float cr );
        virtual void   SetCamVFov        ( float cvfov );
        virtual void   SetCamFarPlane    ( float cfarplane );

        virtual void   SetSpeed		 ( float spd );	// km/saat
        virtual void   SetRotationalSpeed( float spd ); // derece/sn

	virtual void   MoveCamera 	 ( double ForwardInMeter, double RightInMeter, double UpInMeter );

	TCameraMotionMode virtual GetCameraMotion     ( );
	void   		  virtual SetCameraMotion     ( TCameraMotionMode cmotion );
        void   		  virtual AdvanceCameraMotion ( );

        // screen & mouse information

        int    GetLastMouseX   ( );
        int    GetLastMouseY   ( );

        float  GetPanPerPixel 	    ( );
        float  GetRotatePerPixel    ( );
        float  GetAccelerationPerSec( );	// km/saat

        virtual void   SetPanPerPixel 	    ( float spd );
        virtual void   SetRotatePerPixel    ( float spd );
        virtual void   SetAccelerationPerSec( float spd ); // km/saat

	virtual void SetScreenSize  	 ( int wd, int hd );

	// Button > 0 = left, 1 = right, 2 = middle
	virtual void SetMouseFunction   ( int Button, TMouseFunction3DMode mmode );
	virtual void OnMouseDown        ( int Button, bool ShiftDown, bool AltDown, bool CtrlDown, bool DoubleClicked, int X, int Y );
	virtual void OnMouseUp          ( int Button, bool ShiftDown, bool AltDown, bool CtrlDown, int X, int Y);
	virtual void OnMouseMove        ( bool ShiftDown, bool AltDown, bool CtrlDown, int X, int Y );
        virtual void OnAbnormalMouseUp  ( ); // call immediate after a dialog box is shown inside a mouse event

        // callback functions
        void * 			        data;

        TMouse2Param3DProc	        LeftDownEvent;
        TMouse2Param3DProc	        RightDownEvent;
        TMouse2Param3DProc	        MiddleDownEvent;

        TMouse4Param3DProc	        MoveEvent;
        TMouse6Param3DProc	        LeftMoveEvent;
        TMouse6Param3DProc	        RightMoveEvent;
        TMouse6Param3DProc	        MiddleMoveEvent;

        TMouse6Param3DProc	        LeftUpEvent;
        TMouse6Param3DProc	        RightUpEvent;
        TMouse6Param3DProc	        MiddleUpEvent;

        TSetCaptureControl3DProc	SetCaptureControlEvent;
};

#endif
