//---------------------------------------------------------------------------
#ifndef MuhGLPanelH
#define MuhGLPanelH
//---------------------------------------------------------------------------

#include <windows.hpp>
#include <Classes.hpp>
#include <float.h>
#include <syncobjs.hpp>

#include <gl/gl.h>
#include <gl/glu.h>

enum TMuhGLFinalDrawingType {foGLNone, foGLFlush, foGLFinish};

//---------------------------------------------------------------------------

class TMuhGLPixelFormat : public TPersistent
{
protected:
        int fNumOfColorBits;
        int fNumOfAccumulationBits;
        int fNumOfDepthBits;
        int fNumOfStencilBits;
public:
        // constructor - destructor
        __fastcall virtual TMuhGLPixelFormat() : TPersistent(),
                fNumOfColorBits (24),
                fNumOfDepthBits (32),
                fNumOfAccumulationBits (8),
                fNumOfStencilBits (8)
                {
                }
        __fastcall virtual ~TMuhGLPixelFormat()
                {
                }
__published:
        __property int NumOfColorBits         = {read=fNumOfColorBits,write=fNumOfColorBits,default=24};
        __property int NumOfDepthBits         = {read=fNumOfDepthBits,write=fNumOfDepthBits,default=32};
        __property int NumOfAccumulationBits  = {read=fNumOfAccumulationBits,write=fNumOfAccumulationBits,default=8};
        __property int NumOfStencilBits       = {read=fNumOfStencilBits,write=fNumOfStencilBits,default=8};
};

//---------------------------------------------------------------------------

class TMuhGLProperties : public TPersistent
{
protected:
        float 			fFrameRate;
        float 		     	fAspectRatio;
        bool  			fClearColorBuffer;
        bool  			fClearDepthBuffer;
        bool  			fSwapBuffersOn;
        TColor 			fClearColor;
        TMuhGLFinalDrawingType 	fFinalOperation;

public:
        // constructor - destructor
        __fastcall virtual TMuhGLProperties() : TPersistent(),
                    fClearColor                 (0x00E79545),
                    fClearColorBuffer           (true),
                    fClearDepthBuffer           (true),
                    fFinalOperation             (foGLFlush),
                    fSwapBuffersOn              (true) { }
        __fastcall virtual ~TMuhGLProperties() { }
__published:
        __property bool 			ClearColorBuffer = {read=fClearColorBuffer,write=fClearColorBuffer,default=true};
        __property bool 			ClearDepthBuffer = {read=fClearDepthBuffer,write=fClearDepthBuffer,default=true};
        __property bool 			DoSwapBuffers    = {read=fSwapBuffersOn,write=fSwapBuffersOn,default=true};
        __property TMuhGLFinalDrawingType 	FinalOperation   = {read=fFinalOperation,write=fFinalOperation,default=foGLFlush};
        __property TColor 			ClearColor       = {read=fClearColor,write=fClearColor,default=clBlack};
};

//---------------------------------------------------------------------------

class PACKAGE TMuhGLPanel : public TCustomPanel
{
protected:
        // properties
        bool  fInitialized;
        bool  fFirstDraw;
        float fFrameRate;
        float fAspectRatio;

        // events
        TNotifyEvent fOnInitialize;
        TNotifyEvent fOnFirstDraw;
        TNotifyEvent fOnPanelResize;
        TNotifyEvent fOnDisplay;
        TNotifyEvent fOnFinalize;

        // persistent
        TMuhGLPixelFormat  * fPixelFormat;
        TMuhGLProperties * fGLProperties;

        // system
        long TickCount;
        HDC hdc;

        float dx, dy, dz;

	__property OnResize ;

        // functions
        void __fastcall MyOnResize(System::TObject* Sender);
        void __fastcall SetPixelFormatDescriptor(unsigned char alpha);

        void __fastcall InitializeGL    ( );
        void __fastcall FinalizeGL      ( );
	void __fastcall MakeCurrent     ( );
        void __fastcall UnMakeCurrent	( );

	virtual void __fastcall Paint(void)
        {
	    if(ComponentState.Contains(csDesigning))
            	TCustomPanel::Paint();
                else
        	RenderFrame	( );
        };

	void __fastcall myWMEraseBkgnd(TWMEraseBkgnd &Message)
	{
	    if(!ComponentState.Contains(csDesigning))
		Message.Result   = 1;
	}

	BEGIN_MESSAGE_MAP
		MESSAGE_HANDLER( WM_ERASEBKGND, TWMEraseBkgnd, myWMEraseBkgnd )
	END_MESSAGE_MAP(TCustomPanel)

public:
        // property
        __property float FrameRate   = {read=fFrameRate};
        __property float AspectRatio = {read=fAspectRatio};

        __property MouseCapture;

        // constructor - destructor
        __fastcall virtual TMuhGLPanel ( TComponent* Owner );
        __fastcall virtual ~TMuhGLPanel( );

        void __fastcall PreRendering   ( int vleft = 0, int vright = 0, int vwidth = 0, int vheight = 0 );
        void __fastcall PostRendering  ( );

        void __fastcall RenderFrame    ( );

        bool IsKeyDown( int key) { return (GetAsyncKeyState(key) < 0 ); }

__published:
	__property Align ;
	__property DragCursor ;
	__property DragKind ;
	__property DragMode ;
	__property PopupMenu ;
	__property ShowHint ;
	__property Visible ;

	__property OnCanResize ;
	__property OnClick ;
	__property OnConstrainedResize ;
	__property OnDockDrop ;
	__property OnDockOver ;
	__property OnDblClick ;
	__property OnDragDrop ;
	__property OnDragOver ;
	__property OnEndDock ;
	__property OnEndDrag ;
	__property OnEnter ;
	__property OnExit ;
	__property OnGetSiteInfo ;
	__property OnMouseDown ;
	__property OnMouseMove ;
	__property OnMouseUp ;
	__property OnStartDock ;
	__property OnStartDrag ;
	__property OnUnDock ;

        __property ControlStyle;

        // Properties
        __property bool Initialized       = {read=fInitialized};

        // Persistent
	__property TMuhGLPixelFormat  * GLPixelFormat = {read=fPixelFormat, write=fPixelFormat};
	__property TMuhGLProperties * GLProperties  = {read=fGLProperties, write=fGLProperties};

        // Events
        __property TNotifyEvent OnGLInitialize   = {read=fOnInitialize,write=fOnInitialize,default=NULL};
        __property TNotifyEvent OnGLFirstDraw    = {read=fOnFirstDraw,write=fOnFirstDraw,default=NULL};
        __property TNotifyEvent OnGLResize   	 = {read=fOnPanelResize,write=fOnPanelResize,default=NULL};
        __property TNotifyEvent OnGLDisplay      = {read=fOnDisplay,write=fOnDisplay,default=NULL};
	__property TNotifyEvent OnGLFinalize     = {read=fOnFinalize,write=fOnFinalize,default=NULL};
};
//---------------------------------------------------------------------------
#endif
