#ifndef __GAME_ENVIRONMENT_VISUALIZER__
#define __GAME_ENVIRONMENT_VISUALIZER__

#include "standard_objects.h"

//******************************************************************************
// Classes
//******************************************************************************

class TEnvironmentVisualizer;

//******************************************************************************
// TEnvironmentSurfaceVisualizer3D
//******************************************************************************

class TEnvironmentVisualizer
{
protected:
        float  camx, camy, camz, camh, camp, camr;
        float  camhfov, camvfov, camfarplane;

        bool   clearback;
        float  clearColorR, clearColorG, clearColorB;

        unsigned int               lasttime;
        float                      frametime;  // sec

        float                      sunheading, sunpitch;

        bool                       fillpoly;
        bool                       smoothshade;
        bool                       gllighting;

        int 			   widthpixels;
        int 			   heightpixels;

public:
	         TEnvironmentVisualizer ();
        virtual ~TEnvironmentVisualizer ();

        // screen information
        int     GetScreenWidth       ( );
        void	SetScreenWidth	     ( int sw   );

        int	GetScreenHeight      ( );
        void	SetScreenHeight      ( int sh   );

        void    GetBackgroundColor   ( float &r, float &g, float &b );
        void    SetBackgroundColor   ( float r,  float g,  float b  );

        // visualizer setting
        bool    GetFilledTriangles   ( );
        void 	SetFilledTriangles   ( bool val );

        bool    GetLightingOn        ( );
        void 	SetLightingOn        ( bool val );

        // sun state
        float   GetSunHeading        ( );
        void    SetSunHeading        ( float sh );

        float   GetSunPitch          ( );
        void    SetSunPitch          ( float sp );

        // rendering scene
        virtual void InitializeRenderingPers  ( float x, float y, float z, float h, float p, float r,
        			                float vfov, float nearplane, float farplane,
                                                bool clearbackground = true,
                                                bool cleardepthbuffer = true );

	virtual void InitializeLights 	   ( );

	virtual void RenderObjects 	   ( TGameObjectLayer * layer );

        virtual void FinalizeRendering     ( );

        // rendering post query
	float  GetFrameTime 	     ( ); // sec
};

#endif
