#include "graphics_utilities.h"

#ifdef __BCPLUSPLUS__
        #include <Graphics.hpp>
#else
	#define try 		{
	#define __finally 	}
	#define catch(x) 	}
	#define Exception exception
#endif


#include <GL/gl.h>
#include <GL/glu.h>

//******************************************************************************

GLuint  (__stdcall *glCreateShader)	       (GLenum shaderType) = NULL;
void    (__stdcall *glShaderSource)	       (GLuint shader, int numOfStrings, const char **strings, int *lenOfStrings) = NULL;
void    (__stdcall *glCompileShader)	       (GLuint shader) = NULL;
GLuint  (__stdcall *glCreateProgram)	       (void) = NULL;
void    (__stdcall *glAttachShader)	       (GLuint program, GLuint shader) = NULL;
void    (__stdcall *glLinkProgram)	       (GLuint program) = NULL;
void    (__stdcall *glUseProgram)	       (GLuint prog) = NULL;
void    (__stdcall *glDetachShader)	       (GLuint program, GLuint shader) = NULL;
void    (__stdcall *glDeleteShader)	       (GLuint id) = NULL;
void    (__stdcall *glDeleteProgram)	       (GLuint id) = NULL;
GLint   (__stdcall *glGetUniformLocation)      (GLuint program, const char *name) = NULL;
void    (__stdcall *glUniform1f)	       (GLint location, GLfloat v0) = NULL;

//******************************************************************************
// Extern Instances
//******************************************************************************

TOpenGLUtilities 	OpenGLUtils;

//******************************************************************************
// TOpenGLUtilities
//******************************************************************************

static void * get_proc_address(const char *name)
{
	#ifdef _WIN32
		return (void *)wglGetProcAddress(name);
	#else
		return (void *)glXGetProcAddress((const GLubyte *)name);
	#endif
}

TOpenGLUtilities::TOpenGLUtilities()
{
        shadersInitialized =    -1;
}

bool TOpenGLUtilities::InitializeShaders()
{
        if ( shadersInitialized != -1 ) return shadersInitialized == 1;

	glCreateShader =
		(GLuint (__stdcall *)(GLenum shaderType))get_proc_address("glCreateShader");

	glShaderSource =
		(void (__stdcall *)(GLuint shader, int numOfStrings, const char **strings, int *lenOfStrings))get_proc_address("glShaderSource");

	glCompileShader =
		(void (__stdcall *)(GLuint shader))get_proc_address("glCompileShader");

	glCreateProgram =
		(GLuint (__stdcall *)(void))get_proc_address("glCreateProgram");

	glAttachShader =
		(void (__stdcall *)(GLuint program, GLuint shader))get_proc_address("glAttachShader");

	glLinkProgram =
		(void (__stdcall *)(GLuint program))get_proc_address("glLinkProgram");

	glUseProgram =
		(void (__stdcall *)(GLuint prog))get_proc_address("glUseProgram");

	glDetachShader =
		(void (__stdcall *)(GLuint program, GLuint shader))get_proc_address("glDetachShader");

	glDeleteShader =
		(void (__stdcall *)(GLuint id))get_proc_address("glDeleteShader");

	glDeleteProgram =
		(void (__stdcall *)(GLuint id))get_proc_address("glDeleteProgram");

	glGetUniformLocation =
		(GLint (__stdcall *)(GLuint program, const char *name))get_proc_address("glGetUniformLocation");

	glUniform1f =
		(void (__stdcall *)(GLint location, GLfloat v0))get_proc_address("glUniform1f");

        shadersInitialized =   (glCreateShader &&
                                glShaderSource &&
                                glCompileShader &&
                                glCreateProgram &&
                                glAttachShader &&
                                glLinkProgram &&
                                glUseProgram &&
                                glDetachShader &&
                                glDeleteShader &&
                                glDeleteProgram &&
                                glGetUniformLocation &&
                                glUniform1f) ? 1 : 0;

        return shadersInitialized == 1;
}

bool    TOpenGLUtilities::IsShadersInitialized ( )
{
        return shadersInitialized == 1;
}

GLuint  TOpenGLUtilities::CreateShader		(GLenum shaderType)
{
        if ( glCreateShader )
                return glCreateShader(shaderType);
		else
			return 0;
}

void            TOpenGLUtilities::ShaderSource		( unsigned int shader, int numOfStrings,
                                          const char **strings, int *lenOfStrings )
{
        if ( glShaderSource )
                glShaderSource(shader, numOfStrings, strings, lenOfStrings);
}

void            TOpenGLUtilities::CompileShader		( unsigned int shader )
{
        if ( glCompileShader )
                glCompileShader(shader);
}

unsigned int    TOpenGLUtilities::CreateProgram  ( )
{
        if ( glCreateProgram )
                return glCreateProgram();
		else
			return 0;
}

void            TOpenGLUtilities::AttachShader		( unsigned int program, unsigned int shader )
{
        if ( glAttachShader )
                glAttachShader(program, shader);
}

void            TOpenGLUtilities::LinkProgram		( unsigned int program )
{
        if ( glLinkProgram )
                glLinkProgram(program);
}

void            TOpenGLUtilities::UseProgram		( unsigned int program )
{
        if ( glUseProgram )
                glUseProgram(program);
}

void            TOpenGLUtilities::DetachShader		( unsigned int program, unsigned int shader )
{
        if ( glDetachShader )
                glDetachShader(program,shader);
}

void            TOpenGLUtilities::DeleteShader		( unsigned int id )
{
        if ( glDeleteShader )
                glDeleteShader(id);
}

void            TOpenGLUtilities::DeleteProgram		( unsigned int id )
{
        if ( glDeleteProgram )
                glDeleteProgram(id);
}

unsigned int    TOpenGLUtilities::GetUniformLocation	( unsigned int program, const char *name )
{
        if ( glGetUniformLocation )
                return glGetUniformLocation(program,name);
		else
			return 0;
}

void            TOpenGLUtilities::Uniform1f		( int location, float v0 )
{
        if ( glUniform1f )
                glUniform1f(location,v0);
}


TOpenGLUtilities::~TOpenGLUtilities()
{
	for ( int i = 0; i<fonts.size(); i++ )
        	delete fonts[i];
        fonts.clear();
}

TGLFontType * TOpenGLUtilities::LoadFont( mstring fontname, int start, int count,
                     float quality, float extrude, bool bold )
{
	TGLFontType * t = new TGLFontType;
	t->fontStart    =  0;
        t->fontCount    =  0;
	t->fontGLBase   = -1;
        fonts.push_back ( t );

        LOGFONT lf;

        lf.lfHeight = 1000;
        lf.lfWidth  = 0;
        lf.lfEscapement = 0;
        lf.lfOrientation = 0;
        lf.lfWeight = bold ? FW_BOLD : FW_NORMAL;
        lf.lfItalic = false;
        lf.lfUnderline = false;
        lf.lfStrikeOut = false;
        lf.lfCharSet = DEFAULT_CHARSET;
        lf.lfOutPrecision = OUT_DEFAULT_PRECIS;
        lf.lfQuality = CLIP_DEFAULT_PRECIS;
        lf.lfPitchAndFamily = FF_DONTCARE|DEFAULT_PITCH;
        strcpy ( lf.lfFaceName, fontname.c_str() );

        HFONT font = CreateFontIndirect(&lf);
        try
        {
                HDC   hdc  = wglGetCurrentDC();
		HFONT oldfont = (HFONT)SelectObject(hdc, font);
                try
                {
                        t->fontStart  = start;
                        t->fontCount  = count;
                        t->fontGLBase = glGenLists(count);
                        wglUseFontOutlines(	hdc, start , count, t->fontGLBase,
                                                quality,   // This is the sampling tolerance. Higher values create less detailed outlines.
						extrude,   // This is the extrusion depth.
						WGL_FONT_POLYGONS, NULL);
                        int i,w;
                        for ( i = 0; i<255; i++ )
                        {
				GetCharWidth32( hdc, char(i), char(i), &w );
                                t->fontwidths[i] = w/1000.0;
                        }
                }
                __finally
                {
		        SelectObject(hdc, oldfont);
                }
        }
        __finally
        {
                DeleteObject(font);
        }

        return t;
}

void TOpenGLUtilities::DeleteFont    ( TGLFontType * font )
{
	if ( !font ) return;

  	if (font->fontGLBase != -1)
		glDeleteLists( font->fontGLBase, font->fontCount );

	for ( int i = 0; i<fonts.size(); i++ )
        	if ( fonts[i] == font )
                {
					fonts.erase( fonts.begin()+i );
                        break;
                }
        delete font;
}

float TOpenGLUtilities::GetStringWidth( TGLFontType * font, mstring text, float size )
{
	float t = 0;
	char * str = (char*)text.c_str();
        for ( int i = 0; i<text.size(); i++ )
        	t += font->fontwidths[ (unsigned char)str[i] ];
        return t*size;
}

float TOpenGLUtilities::GetStringTopHeight( float size )
{
	return size*0.6666;
}

float TOpenGLUtilities::GetStringBottomHeight( float size )
{
	return size*0.3333;
}

void TOpenGLUtilities::DrawString( TGLFontType * font, mstring text, float size )
{
	mstring targetstr = text;
	if (font->fontStart>0)
  	{
        	for (int i = 0; i< text.size(); i++)
        		targetstr[i] -= font->fontStart;
  	}
  	if (font->fontGLBase != -1)
  	{
  		glListBase(font->fontGLBase);
  		glPushMatrix();
        		glScalef    ( size, size, size );
        		glCallLists ( targetstr.size(), GL_UNSIGNED_BYTE, targetstr.c_str() );
  		glPopMatrix();
  	}
}

void TOpenGLUtilities::ConfigTexture       ( bool    repeat,
                                           int   minmag_filter,
                                           int   env_mode
                                         )
{
        glTexEnvi      ( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, env_mode );
        glTexParameteri( GL_TEXTURE_2D,  GL_TEXTURE_MIN_FILTER, minmag_filter );
        glTexParameteri( GL_TEXTURE_2D,  GL_TEXTURE_MAG_FILTER, minmag_filter );

        if (repeat)
        {
                glTexParameteri( GL_TEXTURE_2D,  GL_TEXTURE_WRAP_S, GL_REPEAT );
                glTexParameteri( GL_TEXTURE_2D,  GL_TEXTURE_WRAP_T, GL_REPEAT );
        }
        else
        {
                glTexParameteri( GL_TEXTURE_2D,  GL_TEXTURE_WRAP_S, GL_CLAMP );
                glTexParameteri( GL_TEXTURE_2D,  GL_TEXTURE_WRAP_T, GL_CLAMP );
        }
}

void TOpenGLUtilities::EnableTextureProjection  ( int mode,
                                                float proj_s1,
                                                float proj_s2,
                                                float proj_s3,
                                                float proj_s4,
                                                float proj_t1,
                                                float proj_t2,
                                                float proj_t3,
                                                float proj_t4
                                         )
{
        glEnable(GL_TEXTURE_GEN_S);
        glEnable(GL_TEXTURE_GEN_T);

        glTexGeni ( GL_S, GL_TEXTURE_GEN_MODE, mode );
        glTexGeni ( GL_T, GL_TEXTURE_GEN_MODE, mode );

        GLfloat s_vector[4] = { proj_s1, proj_s2, proj_s3, proj_s4 };
        GLfloat t_vector[4] = { proj_t1, proj_t2, proj_t3, proj_t4 };

        glTexGenfv( GL_S, GL_OBJECT_PLANE, s_vector );
        glTexGenfv( GL_T, GL_OBJECT_PLANE, t_vector );
}

void TOpenGLUtilities::ConfigureTextureMatrix     ( float proj_scalex,
                                                  float proj_scaley,
                                                  float proj_anglex,
                                                  float proj_angley,
                                                  float proj_anglez,
                                                  float proj_offsetx,
                                                  float proj_offsety,
                                                  float proj_offsetz,
                                                  float local_anglex,
                                                  float local_angley,
                                                  float local_anglez,
                                                  float local_offsetx,
                                                  float local_offsety,
                                                  float local_offsetz
                                                 )
{
        glMatrixMode(GL_TEXTURE);
                glLoadIdentity();
                if ( proj_scalex!=0 || proj_scalex!=0 )
                        glScalef     ( proj_scalex , proj_scaley , 0 );
                if ( local_anglex!=0 || local_angley!=0 || local_anglez!=0 )
                {
                        glRotatef    ( local_angley , 0,1,0 );
                        glRotatef    ( local_anglex , 1,0,0 );
                        glRotatef    ( local_anglez , 0,0,1 );
                }
                if ( local_offsetx!=0 || local_offsetx!=0 || local_offsetx!=0 )
                        glTranslatef ( local_offsetx , local_offsety, local_offsetz );
                glRotatef    ( proj_angley , 0,1,0 );
                glRotatef    ( proj_anglex , 1,0,0 );
                glRotatef    ( proj_anglez , 0,0,1 );
                glTranslatef ( proj_offsetx, proj_offsety, proj_offsetz );
        glMatrixMode(GL_MODELVIEW);
}

void TOpenGLUtilities::DisableTextureProjection ( )
{
        glDisable(GL_TEXTURE_GEN_S);
        glDisable(GL_TEXTURE_GEN_T);

        glMatrixMode(GL_TEXTURE);
                glLoadIdentity();
        glMatrixMode(GL_MODELVIEW);
}

