#ifndef __GAME_ENVIRONMENT_GRAPHICS_UTILITIES__
#define __GAME_ENVIRONMENT_GRAPHICS_UTILITIES__

#include "main_functions.h"

#include <vector>

//******************************************************************************
// Classes
//******************************************************************************

class TOpenGLUtilities;
class TGLFontType;

//******************************************************************************
// Definitions
//******************************************************************************

#define GL_FRAGMENT_SHADER	0x8B30
#define GL_VERTEX_SHADER	0x8B31

//******************************************************************************
// Type Definitions
//******************************************************************************

class TGLFontType
{
public:
        int 	fontStart;
        int 	fontCount;
        int 	fontGLBase;
        float   fontwidths[255];
};

//******************************************************************************
// Extern Instances
//******************************************************************************

extern TOpenGLUtilities 	OpenGLUtils;

//******************************************************************************
// TOpenGLUtilities
//******************************************************************************

class TOpenGLUtilities
{
protected:
        int shadersInitialized;
	std::vector<TGLFontType*> fonts;
public:
        TOpenGLUtilities  ( );
        ~TOpenGLUtilities ( );

        // font utilities
	TGLFontType * 	LoadFont    ( mstring fontname, int start, int count,
                     		      float quality, float extrude, bool bold );
        void		DeleteFont  ( TGLFontType * font );

	float GetStringWidth	    ( TGLFontType * font, mstring text, float size );
	float GetStringTopHeight    ( float size );
	float GetStringBottomHeight ( float size );

	void  DrawString 	    ( TGLFontType * font, mstring text, float size );

        // texture configuration utilities
        void ConfigTexture        	( bool    repeat        = true,
                                          int     minmag_filter   = 0x2703 /*GL_LINEAR_MIPMAP_LINEAR*/,
                                          int     env_mode        = 0x2100 /*GL_MODULATE*/ );

        void EnableTextureProjection    ( int mode            = 0x2401 /*GL_OBJECT_LINEAR*/,
                                          float proj_s1       = 1,
                                          float proj_s2       = 0,
                                          float proj_s3       = 0,
                                          float proj_s4       = 0,
                                          float proj_t1       = 0,
                                          float proj_t2       = 1,
                                          float proj_t3       = 0,
                                          float proj_t4       = 0
                                        );

        void ConfigureTextureMatrix     ( float proj_scalex,
                                          float proj_scaley,
                                          float proj_anglex,
                                          float proj_angley,
                                          float proj_anglez,
                                          float proj_offsetx,
                                          float proj_offsety,
                                          float proj_offsetz,
                                          float local_anglex,
                                          float local_angley,
                                          float local_anglez,
                                          float local_offsetx,
                                          float local_offsety,
                                          float local_offsetz
                                        );

        void DisableTextureProjection ( );

        bool            InitializeShaders       ( );
        bool            IsShadersInitialized    ( );
        
        unsigned int    CreateShader		( unsigned int shaderType );
        void            ShaderSource		( unsigned int shader, int numOfStrings,
                                                  const char **strings, int *lenOfStrings );
        void            CompileShader		( unsigned int shader );
        unsigned int    CreateProgram		( );
        void            AttachShader		( unsigned int program, unsigned int shader );
        void            LinkProgram		( unsigned int program );
        void            UseProgram		( unsigned int program );
        void            DetachShader		( unsigned int program, unsigned int shader );
        void            DeleteShader		( unsigned int id );
        void            DeleteProgram		( unsigned int id );
        unsigned int    GetUniformLocation	( unsigned int program, const char *name );
        void            Uniform1f		( int location, float v0 );
};

#endif

