#include "io_streams.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <io.h>

#ifdef __BCPLUSPLUS__

#else
	#define try     __try
#endif

//******************************************************************************
// TGameIOStream
//******************************************************************************

TGameIOStream::TGameIOStream()
{
}

TGameIOStream::~TGameIOStream()
{
}

bool TGameIOStream::Crop ( unsigned int start, unsigned long cropsize )
{
	unsigned long  filesize = GetSize();

        if (start>=filesize || start+cropsize>filesize) return false;

        unsigned long truncTo   = filesize - cropsize;
        unsigned long movestart = start + cropsize;
        unsigned long movesize  = filesize - start - cropsize;
        unsigned char * movedata = (unsigned char * )malloc ( movesize );

        try
        {

                if (movesize>0 && movedata)
                {
                        Seek_Set   ( movestart );
                        Read       ( movedata, movesize );
                        Seek_Set   ( start );
                        Write      ( movedata, movesize );
                }
                SetSize ( truncTo );
        }
        __finally
        {
                if (movedata) free (movedata);
        }

        return true;
}

bool	TGameIOStream:: Read  ( bool           * val )
{
        return Read ( val, sizeof(bool) ) == sizeof(bool);
}

bool	TGameIOStream:: Read  ( unsigned char  * val )
{
        return Read ( val, sizeof(unsigned char) ) == sizeof(unsigned char);
}

bool  TGameIOStream:: Read  ( char   * val )
{
        return Read ( val, sizeof(char) ) == sizeof(char);
}

bool TGameIOStream:: Read  ( unsigned short * val )
{
        return Read ( val, sizeof(unsigned short) ) == sizeof(unsigned short);
}

bool	TGameIOStream:: Read  ( short  * val )
{
        return Read ( val, sizeof(short) ) == sizeof(short);
}

bool TGameIOStream:: Read  ( unsigned int   * val )
{
        return Read ( val, sizeof(unsigned int) ) == sizeof(unsigned int);
}

bool  TGameIOStream:: Read  ( int    * val )
{
        return Read ( val, sizeof(int) ) == sizeof(int);
}

bool TGameIOStream:: Read  ( unsigned long   * val )
{
        return Read ( val, sizeof(unsigned long) ) == sizeof(unsigned long);
}

bool  TGameIOStream:: Read  ( long   * val )
{
        return Read ( val, sizeof(long) ) == sizeof(long);
}

bool TGameIOStream:: Read  ( __int64    * val )
{
        return Read ( val, sizeof(__int64) ) == sizeof(__int64);
}

bool TGameIOStream:: Read  ( float  * val )
{
        return Read ( val, sizeof(float) ) == sizeof(float);
}

bool TGameIOStream:: Read  ( double * val )
{
        return Read ( val, sizeof(double) ) == sizeof(double);
}

bool TGameIOStream:: Read  ( long double    * val )
{
        return Read ( val, sizeof(long double) ) == sizeof(long double);
}

bool TGameIOStream:: ReadString  ( mstring * val )
{
	unsigned int size = 0;
        if ( Read ( &size, sizeof(unsigned int) ) != sizeof(unsigned int) )
        	return false;
        val->resize( size );
		char dummy;
		if ( size>0 )
		{
	        if ( Read ( (void*)val->c_str(), size+1 ) != size+1 )
		    	return false;
		}
		else
		{
	        if ( Read ( &dummy, 1 ) != 1 )
		    	return false;
		}
	return true;
}

bool TGameIOStream:: Write ( bool   * val )
{
        return Write ( val, sizeof(bool) ) == sizeof(bool);
}

bool TGameIOStream:: Write ( unsigned char  * val )
{
        return Write ( val, sizeof(unsigned char) ) == sizeof(unsigned char);
}

bool TGameIOStream:: Write ( char   * val )
{
        return Write ( val, sizeof(char) ) == sizeof(char);
}

bool TGameIOStream:: Write ( unsigned short * val )
{
        return Write ( val, sizeof(unsigned short) ) == sizeof(unsigned short);
}

bool TGameIOStream:: Write ( short  * val )
{
        return Write ( val, sizeof(short) ) == sizeof(short);
}

bool TGameIOStream:: Write ( unsigned int   * val )
{
        return Write ( val, sizeof(unsigned int) ) == sizeof(unsigned int);
}

bool TGameIOStream:: Write ( int    * val )
{
        return Write ( val, sizeof(int) ) == sizeof(int);
}

bool TGameIOStream:: Write ( unsigned long   * val )
{
        return Write ( val, sizeof(unsigned long) ) == sizeof(unsigned long);
}

bool TGameIOStream:: Write ( long    * val )
{
        return Write ( val, sizeof(long) ) == sizeof(long);
}

bool TGameIOStream:: Write ( __int64    * val )
{
        return Write ( val, sizeof(__int64) ) == sizeof(__int64);
}

bool TGameIOStream:: Write ( float  * val )
{
        return Write ( val, sizeof(float) ) == sizeof(float);
}

bool TGameIOStream:: Write ( double * val )
{
        return Write ( val, sizeof(double) ) == sizeof(double);
}

bool TGameIOStream:: Write ( long double    * val )
{
        return Write ( val, sizeof(long double) ) == sizeof(long double);
}

bool TGameIOStream:: WriteString ( mstring* val )
{
	unsigned int size = val->length();
        if ( Write ( &size, sizeof(unsigned int) ) != sizeof(unsigned int) )
        	return false;
        return Write ( (void*)val->c_str(), size+1 ) == size+1;
}

//******************************************************************************
// TGameFileStream
//******************************************************************************

TGameFileStream::TGameFileStream ( )
{
        hFile = -1;
}

TGameFileStream::~TGameFileStream()
{
        if (hFile != -1) CloseFile();
}

bool TGameFileStream::OpenFileForRead(mstring fname)
{
        if (hFile != -1) CloseFile();
        hFile = open(fname.c_str(),O_RDONLY | O_BINARY);
        if (hFile == -1) return false; else return true;
}

bool TGameFileStream::OpenFileForCreate(mstring fname)
{
        if (hFile != -1) CloseFile();
        hFile = open(fname.c_str(),O_CREAT | O_TRUNC | O_RDWR | O_BINARY, S_IREAD | S_IWRITE);
        if (hFile == -1) return false; else return true;
}

bool TGameFileStream::OpenFileForModify(mstring fname)
{
        if (hFile != -1) CloseFile();
        hFile = open(fname.c_str(), O_CREAT | O_RDWR | O_BINARY, S_IREAD | S_IWRITE);
        if (hFile == -1) return false; else return true;
}

bool TGameFileStream::IsOpened ()
{
	return (hFile != -1);
}

void TGameFileStream::CloseFile()
{
        if (hFile!=-1) close(hFile);
        hFile = -1;
}

unsigned long TGameFileStream::GetSize     ( )
{
	#ifdef __linux__
                long prevpos = tell ( hFile );

                lseek( hFile, 0, SEEK_END);
                long filesize = tell ( hFile );

                lseek( hFile, prevpos, SEEK_SET);
        	return filesize;
        #else
        	return filelength (hFile);
        #endif
}

bool TGameFileStream::SetSize     ( unsigned long size )
{
	return chsize (hFile, size ) == 0;
}

unsigned long TGameFileStream::GetPos   ( )
{
        return tell ( hFile );
}

unsigned long TGameFileStream::Seek_Cur(unsigned long offset)
{
        return lseek( hFile,offset,SEEK_CUR);
}

unsigned long TGameFileStream::Seek_End(unsigned long offset)
{
        return lseek( hFile,offset,SEEK_END);
}

unsigned long TGameFileStream::Seek_Set(unsigned long offset)
{
        return lseek( hFile,offset,SEEK_SET);
}

unsigned long  TGameFileStream::Read  ( void *buf, unsigned long len )
{
        return read( hFile,buf,len);
}

unsigned long  TGameFileStream::Write ( void *buf, unsigned long len )
{
        return write( hFile, buf, len);
}

unsigned long TGameFileStream::ReadWithoutNewLine  ( void *buf, unsigned long len )
{

        long numread;
        long totalread = 0;
        while ( totalread<len )
        {
		numread = read( hFile, &((unsigned char*)buf)[totalread], len-totalread );
                if ( numread<=0 ) return totalread;

                for ( int i=totalread; i<totalread+numread; i++ )
                if  ( ((unsigned char*)buf)[i] == '\n' || ((unsigned char*)buf)[i] == '\r' )
                {
	                for ( int i2=i+1; i2<totalread+numread; i2++ )
                        	((unsigned char*)buf)[i2-1] = ((unsigned char*)buf)[i2];
                        i--;
                        numread--;
                }
                totalread+=numread;
        }
	return len;
}

bool	TGameFileStream::Read  ( bool           * val ) { return TGameIOStream::Read (val); };
bool	TGameFileStream::Read  ( unsigned char  * val ) { return TGameIOStream::Read (val); };
bool	TGameFileStream::Read  ( char           * val ) { return TGameIOStream::Read (val); };
bool	TGameFileStream::Read  ( unsigned short * val ) { return TGameIOStream::Read (val); };
bool	TGameFileStream::Read  ( short          * val ) { return TGameIOStream::Read (val); };
bool	TGameFileStream::Read  ( unsigned int   * val ) { return TGameIOStream::Read (val); };
bool	TGameFileStream::Read  ( int            * val ) { return TGameIOStream::Read (val); };
bool	TGameFileStream::Read  ( __int64        * val ) { return TGameIOStream::Read (val); };
bool	TGameFileStream::Read  ( unsigned long  * val ) { return TGameIOStream::Read (val); };
bool	TGameFileStream::Read  ( long           * val ) { return TGameIOStream::Read (val); };
bool	TGameFileStream::Read  ( float          * val ) { return TGameIOStream::Read (val); };
bool	TGameFileStream::Read  ( double         * val ) { return TGameIOStream::Read (val); };
bool	TGameFileStream::Read  ( long double    * val ) { return TGameIOStream::Read (val); };
bool	TGameFileStream::ReadString  ( mstring  * val ) { return TGameIOStream::ReadString (val); };

bool	TGameFileStream::Write ( bool           * val ) { return TGameIOStream::Write (val); };
bool	TGameFileStream::Write ( unsigned char  * val ) { return TGameIOStream::Write (val); };
bool	TGameFileStream::Write ( char           * val ) { return TGameIOStream::Write (val); };
bool	TGameFileStream::Write ( unsigned short * val ) { return TGameIOStream::Write (val); };
bool	TGameFileStream::Write ( short          * val ) { return TGameIOStream::Write (val); };
bool	TGameFileStream::Write ( unsigned int   * val ) { return TGameIOStream::Write (val); };
bool	TGameFileStream::Write ( int            * val ) { return TGameIOStream::Write (val); };
bool	TGameFileStream::Write ( __int64        * val ) { return TGameIOStream::Write (val); };
bool	TGameFileStream::Write ( unsigned long  * val ) { return TGameIOStream::Write (val); };
bool	TGameFileStream::Write ( long           * val ) { return TGameIOStream::Write (val); };
bool	TGameFileStream::Write ( float          * val ) { return TGameIOStream::Write (val); };
bool	TGameFileStream::Write ( double         * val ) { return TGameIOStream::Write (val); };
bool	TGameFileStream::Write ( long double    * val ) { return TGameIOStream::Write (val); };
bool	TGameFileStream::WriteString ( mstring  * val ) { return TGameIOStream::WriteString (val); };

//******************************************************************************
// TGameMemoryStream
//******************************************************************************

TGameMemoryStream::TGameMemoryStream ( )
{
        mem  = NULL;
        pos  = 0;
        size = 0;
}

TGameMemoryStream::~TGameMemoryStream()
{
        if (mem) free(mem);
}

void * TGameMemoryStream::GetMemory ( )
{
        return mem;
}

unsigned long TGameMemoryStream::GetSize     ( )
{
	return size;
}

bool TGameMemoryStream::SetSize     ( unsigned long nsize )
{
	mem = (unsigned char*)realloc ( mem, nsize );
        size = nsize;
	return mem;
}

unsigned long TGameMemoryStream::GetPos   ( )
{
        return pos;
}

unsigned long TGameMemoryStream::Seek_Cur(unsigned long offset)
{
        pos += offset;
		return pos;
}

unsigned long TGameMemoryStream::Seek_End(unsigned long offset)
{
        pos = size - offset;
		return pos;
}

unsigned long TGameMemoryStream::Seek_Set(unsigned long offset)
{
        pos = offset;
		return pos;
}

unsigned long  TGameMemoryStream::Read  ( void *buf, unsigned long len )
{
        if ( pos>=size ) return 0;
        if ( size-pos < len ) len = size-pos;

        mCopyMemory ( buf, &mem[pos], len );
        pos += len;
        return len;
}

unsigned long  TGameMemoryStream::Write ( void *buf, unsigned long len )
{
        unsigned int newsize = pos + len;
        if ( newsize > size ) SetSize ( newsize );
        mCopyMemory ( &mem[pos], buf, len );
        pos += len;
        return len;
}

bool	TGameMemoryStream::Read  ( bool           * val ) { return TGameIOStream::Read (val); };
bool	TGameMemoryStream::Read  ( unsigned char  * val ) { return TGameIOStream::Read (val); };
bool	TGameMemoryStream::Read  ( char           * val ) { return TGameIOStream::Read (val); };
bool	TGameMemoryStream::Read  ( unsigned short * val ) { return TGameIOStream::Read (val); };
bool	TGameMemoryStream::Read  ( short          * val ) { return TGameIOStream::Read (val); };
bool	TGameMemoryStream::Read  ( unsigned int   * val ) { return TGameIOStream::Read (val); };
bool	TGameMemoryStream::Read  ( int            * val ) { return TGameIOStream::Read (val); };
bool	TGameMemoryStream::Read  ( __int64        * val ) { return TGameIOStream::Read (val); };
bool	TGameMemoryStream::Read  ( unsigned long  * val ) { return TGameIOStream::Read (val); };
bool	TGameMemoryStream::Read  ( long           * val ) { return TGameIOStream::Read (val); };
bool	TGameMemoryStream::Read  ( float          * val ) { return TGameIOStream::Read (val); };
bool	TGameMemoryStream::Read  ( double         * val ) { return TGameIOStream::Read (val); };
bool	TGameMemoryStream::Read  ( long double    * val ) { return TGameIOStream::Read (val); };
bool	TGameMemoryStream::ReadString  ( mstring  * val ) { return TGameIOStream::ReadString (val); };

bool	TGameMemoryStream::Write ( bool           * val ) { return TGameIOStream::Write (val); };
bool	TGameMemoryStream::Write ( unsigned char  * val ) { return TGameIOStream::Write (val); };
bool	TGameMemoryStream::Write ( char           * val ) { return TGameIOStream::Write (val); };
bool	TGameMemoryStream::Write ( unsigned short * val ) { return TGameIOStream::Write (val); };
bool	TGameMemoryStream::Write ( short          * val ) { return TGameIOStream::Write (val); };
bool	TGameMemoryStream::Write ( unsigned int   * val ) { return TGameIOStream::Write (val); };
bool	TGameMemoryStream::Write ( int            * val ) { return TGameIOStream::Write (val); };
bool	TGameMemoryStream::Write ( __int64        * val ) { return TGameIOStream::Write (val); };
bool	TGameMemoryStream::Write ( unsigned long  * val ) { return TGameIOStream::Write (val); };
bool	TGameMemoryStream::Write ( long           * val ) { return TGameIOStream::Write (val); };
bool	TGameMemoryStream::Write ( float          * val ) { return TGameIOStream::Write (val); };
bool	TGameMemoryStream::Write ( double         * val ) { return TGameIOStream::Write (val); };
bool	TGameMemoryStream::Write ( long double    * val ) { return TGameIOStream::Write (val); };
bool	TGameMemoryStream::WriteString ( mstring  * val ) { return TGameIOStream::WriteString (val); };

