#ifndef __GAME_ENVIRONMENT_IO_STREAMS__
#define __GAME_ENVIRONMENT_IO_STREAMS__

#include "main_functions.h"

//******************************************************************************
// Classes
//******************************************************************************

class TGameIOStream;
class TGameFileStream;
class TGameMemoryStream;

//******************************************************************************
// TGameIOStream
//******************************************************************************

class  TGameIOStream
{
public:
        TGameIOStream();
        virtual ~TGameIOStream();

	unsigned long virtual GetSize  ( ) = 0;
	bool virtual          SetSize  ( unsigned long size ) = 0;

        unsigned long virtual GetPos   ( ) = 0;
        unsigned long virtual Seek_Cur ( unsigned long offset ) = 0;
        unsigned long virtual Seek_End ( unsigned long offset ) = 0;
        unsigned long virtual Seek_Set ( unsigned long offset ) = 0;

        unsigned long virtual Read     ( void *buf, unsigned long len ) = 0;
        unsigned long virtual Write    ( void *buf, unsigned long len ) = 0;

	bool virtual          Crop     ( unsigned int start, unsigned long cropsize );

        bool virtual Read  ( bool           * val );
        bool virtual Read  ( unsigned char  * val );
        bool virtual Read  ( char           * val );
        bool virtual Read  ( unsigned short * val );
        bool virtual Read  ( short          * val );
        bool virtual Read  ( unsigned int   * val );
        bool virtual Read  ( int            * val );
        bool virtual Read  ( __int64        * val );
        bool virtual Read  ( unsigned long  * val );
        bool virtual Read  ( long           * val );
        bool virtual Read  ( float          * val );
        bool virtual Read  ( double         * val );
        bool virtual Read  ( long double    * val );
        bool virtual ReadString  ( mstring  * val );

        bool virtual Write ( bool           * val );
        bool virtual Write ( unsigned char  * val );
        bool virtual Write ( char           * val );
        bool virtual Write ( unsigned short * val );
        bool virtual Write ( short          * val );
        bool virtual Write ( unsigned int   * val );
        bool virtual Write ( int            * val );
        bool virtual Write ( __int64        * val );
        bool virtual Write ( unsigned long  * val );
        bool virtual Write ( long           * val );
        bool virtual Write ( float          * val );
        bool virtual Write ( double         * val );
        bool virtual Write ( long double    * val );
        bool virtual WriteString ( mstring  * val );
};

//******************************************************************************
// TGameFileStream
//******************************************************************************

class  TGameFileStream : public TGameIOStream
{
protected:
        int hFile;
public:
        TGameFileStream();
        virtual ~TGameFileStream();

        bool OpenFileForRead   (mstring fname);	// if file exists, it is open for read
        bool OpenFileForCreate (mstring fname);	// file is always created and reset
        bool OpenFileForModify (mstring fname);	// if file not exists, it is created

        bool IsOpened ();

        void CloseFile();

	unsigned long virtual GetSize     ( );
	bool virtual          SetSize     ( unsigned long size   );

        unsigned long virtual GetPos      ( );
        unsigned long virtual Seek_Cur    ( unsigned long offset );
        unsigned long virtual Seek_End    ( unsigned long offset );
        unsigned long virtual Seek_Set    ( unsigned long offset );

        bool virtual Read  ( bool           * val );
        bool virtual Read  ( unsigned char  * val );
        bool virtual Read  ( char           * val );
        bool virtual Read  ( unsigned short * val );
        bool virtual Read  ( short          * val );
        bool virtual Read  ( unsigned int   * val );
        bool virtual Read  ( int            * val );
        bool virtual Read  ( __int64        * val );
        bool virtual Read  ( unsigned long  * val );
        bool virtual Read  ( long           * val );
        bool virtual Read  ( float          * val );
        bool virtual Read  ( double         * val );
        bool virtual Read  ( long double    * val );
        bool virtual ReadString  ( mstring  * val );

        bool virtual Write ( bool           * val );
        bool virtual Write ( unsigned char  * val );
        bool virtual Write ( char           * val );
        bool virtual Write ( unsigned short * val );
        bool virtual Write ( short          * val );
        bool virtual Write ( unsigned int   * val );
        bool virtual Write ( int            * val );
        bool virtual Write ( __int64        * val );
        bool virtual Write ( unsigned long  * val );
        bool virtual Write ( long           * val );
        bool virtual Write ( float          * val );
        bool virtual Write ( double         * val );
        bool virtual Write ( long double    * val );
        bool virtual WriteString ( mstring  * val );

        unsigned long virtual Read  		 ( void *buf, unsigned long len );
        unsigned long virtual Write 		 ( void *buf, unsigned long len );
	unsigned long virtual ReadWithoutNewLine ( void *buf, unsigned long len );
};

//******************************************************************************
// TGameMemoryStream
//******************************************************************************

class  TGameMemoryStream : public TGameIOStream
{
protected:
        unsigned char * mem;
        unsigned long pos;
        unsigned long size;
public:
        TGameMemoryStream();
        virtual ~TGameMemoryStream();

        void * GetMemory ( );

	unsigned long virtual GetSize     ( );
	bool virtual          SetSize     ( unsigned long size   );

        unsigned long virtual GetPos      ( );
        unsigned long virtual Seek_Cur    ( unsigned long offset );
        unsigned long virtual Seek_End    ( unsigned long offset );
        unsigned long virtual Seek_Set    ( unsigned long offset );

        bool virtual Read  ( bool           * val );
        bool virtual Read  ( unsigned char  * val );
        bool virtual Read  ( char           * val );
        bool virtual Read  ( unsigned short * val );
        bool virtual Read  ( short          * val );
        bool virtual Read  ( unsigned int   * val );
        bool virtual Read  ( int            * val );
        bool virtual Read  ( __int64        * val );
        bool virtual Read  ( unsigned long  * val );
        bool virtual Read  ( long           * val );
        bool virtual Read  ( float          * val );
        bool virtual Read  ( double         * val );
        bool virtual Read  ( long double    * val );
        bool virtual ReadString  ( mstring  * val );

        bool virtual Write ( bool           * val );
        bool virtual Write ( unsigned char  * val );
        bool virtual Write ( char           * val );
        bool virtual Write ( unsigned short * val );
        bool virtual Write ( short          * val );
        bool virtual Write ( unsigned int   * val );
        bool virtual Write ( int            * val );
        bool virtual Write ( __int64        * val );
        bool virtual Write ( unsigned long  * val );
        bool virtual Write ( long           * val );
        bool virtual Write ( float          * val );
        bool virtual Write ( double         * val );
        bool virtual Write ( long double    * val );
        bool virtual WriteString ( mstring  * val );

        unsigned long virtual Read  	  ( void *buf, unsigned long len );
        unsigned long virtual Write 	  ( void *buf, unsigned long len );
};

#endif

