#include "main_functions.h"

#include <io.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <sys/stat.h>

#ifdef __BCPLUSPLUS__
        #include <filectrl.hpp>
	#include <SysUtils.hpp>
        #include <Math.hpp>
#else
	#include <atlbase.h>
	#include <winbase.h>
#endif

//******************************************************************************
// Externs
//******************************************************************************

mstring ApplicationDirectory = mGetCurrentDir ( );

//******************************************************************************
// mdatetime
//******************************************************************************

mdatetime::mdatetime ( )
{
	datetime = 0;
};

mdatetime::mdatetime ( double t )
{
	datetime = t;
};

//******************************************************************************
// String Functions
//******************************************************************************

mstring mIntToStr ( int num )
{
	char str[255];
        sprintf ( str, "%i", num );
        return mstring(str);
}

int mMixStrToInt ( mstring str )
{
	int num = 0;
        sscanf ( str.c_str(), "%i", &num );
        return num;
}

int mStrToInt ( mstring str )
{
	int num= 0;
        mstring newstring = "";

        int first = 0;
        for ( int i = 0; i< (int)str.length(); i++)
        {
        	if ( first==0 && ( str[i] == '+' || str[i] == '-' ) )
                {
                        newstring += str[i];
                }
                else
        	if ( first<=1 && str[i] == '0' )
                {
                	first = 1;
                }
                else
        	if ( str[i] >= '0' && str[i] <= '9' )
               	{
                	newstring += str[i];
                        first = 2;
               	}
                else
                if ( first>0 ) break;
        }

        sscanf ( newstring.c_str(), "%i", &num );

        return num;
}

mstring mFloatToStrF ( double num, int numofdecimals )
{
	char str[255];
        mstring decs = "%#." + mIntToStr(numofdecimals) + "lf";
        sprintf ( str, decs.c_str(), num );
        return mstring(str);
}

mstring mFloatToStr ( double num )
{
	char str[255];
        sprintf ( str, "%lf", num );
        int i, j   = -1;
        bool nonezero = false;
        for ( i = int(strlen(str))-1; i>=0; i-- )
        {
                if ( !nonezero && str[i] == '0' )
                {
                        j = i;
                }
                else
                if ( str[i] == '.' )
                {
                        if ( j != -1 )
                        {
                                if ( i+1 == j )
                                        str[i] = 0;
                                        else
                                        str[j] = 0;
                        }
                        break;
                }
                else
                nonezero = true;
        }
        return mstring(str);
}

double mStrToFloat ( mstring str )
{
	double num = 0;
        sscanf ( str.c_str(), "%lf", &num );
        return num;
}

mstring mUpperCase  ( mstring str, bool turkishfirst )
{
	mstring st;
        st.resize( str.length() );

        for ( int i = 0; i<(int)str.length(); i++ )
        {
                if ( turkishfirst && str[i] == 'i' )
                        st[i] = '';
		        else
                if ( str[i] >= 'a' && str[i] <= 'z' )
                        st[i] = str[i] - 'a' + 'A';
                        else
                if ( str[i] == '' )
                        st[i] = 'I';
		        else
                if ( str[i] == '' )
                        st[i] = '';
		        else
                if ( str[i] == '' )
                        st[i] = '';
		        else
                if ( str[i] == '' )
                        st[i] = '';
		        else
                if ( str[i] == '' )
                        st[i] = '';
		        else
                if ( str[i] == '' )
                        st[i] = '';
		        else
                        st[i] = str[i];
        }

        return st;
}

mstring mTrim  ( mstring str, bool intern, bool endonly )
{
        mstring st = "";
        if ( endonly )
        {
                int fin = str.length()-1;
                for ( ; fin>=0; fin-- )
                        if ( str[fin] != ' ' )
                                break;

                st = mSubString ( str, 0, fin+1 );
        }
        else
        if ( intern )
        {
        	for ( int i = 0; i<(int)str.length(); i++ )
                	if ( str[i] != ' ' )
                        	st = st + str[i];
        }
        else
        {
        	int start = 0;
        	for ( ; start<(int)str.length(); start++ )
                	if ( str[start] != ' ' )
                        	break;

                if ( start<(int)str.length() )
                {
                        int fin = str.length()-1;
                        for ( ; fin>=0; fin-- )
                                if ( str[fin] != ' ' )
                                        break;

                        st = mSubString ( str, start, fin-start+1 );
                }
        }
        return st;
}

mstring mSubString  ( mstring str,int start,int count )
{
        mstring st = "";
        if (start+count>(int)str.length())
                return  st;
        for ( int i =start ; i<start+count; i++ )
                st = st + str[i];
        return st;
}

//******************************************************************************
// Time Functions
//******************************************************************************

mdatetime mEncodeDateTime ( int year, int month, int day, int hour, int minutes, int seconds )
{
        static double daystomonth[12]= {0,
                                        31,
                                        31+28,
                                        31+28+31,
                                        31+28+31+30,
                                        31+28+31+30+31,
                                        31+28+31+30+31+30,
                                        31+28+31+30+31+30+31,
                                        31+28+31+30+31+30+31+31,
                                        31+28+31+30+31+30+31+31+30,
                                        31+28+31+30+31+30+31+31+30+31,
                                        31+28+31+30+31+30+31+31+30+31+30};

        bool   thisresudial = (int(year/4.0) == year/4.0);

        double yeardays  = 1 + (year-1900)*365;
        double resudials = 1 + int((year-1900-1)/4.0) + ((thisresudial && month>2)?1:0);
        double monthdays = daystomonth[month-1];
        double days      = day-1;
        double hours     = hour/24.0 + minutes/24.0/60.0 + seconds/24.0/60.0/60.0;

        return mdatetime ( yeardays + resudials + monthdays + days + hours );
}

void mDecodeDateTime ( mdatetime dt, int &year, int &month, int &day, int &hour, int &minutes, int &seconds )
{
        static double daystomonthMin[12]= {
                                        0,
                                        31,
                                        31+28,
                                        31+28+31,
                                        31+28+31+30,
                                        31+28+31+30+31,
                                        31+28+31+30+31+30,
                                        31+28+31+30+31+30+31,
                                        31+28+31+30+31+30+31+31,
                                        31+28+31+30+31+30+31+31+30,
                                        31+28+31+30+31+30+31+31+30+31,
                                        31+28+31+30+31+30+31+31+30+31+30};

        static double daystomonth[12]= {31,
                                        31+28,
                                        31+28+31,
                                        31+28+31+30,
                                        31+28+31+30+31,
                                        31+28+31+30+31+30,
                                        31+28+31+30+31+30+31,
                                        31+28+31+30+31+30+31+31,
                                        31+28+31+30+31+30+31+31+30,
                                        31+28+31+30+31+30+31+31+30+31,
                                        31+28+31+30+31+30+31+31+30+31+30,
                                        31+28+31+30+31+30+31+31+30+31+30+31};

        year = int(dt.datetime/365)+1900;
        while ( (1 + (year-1900)*365)+(1 + int((year-1900-1)/4.0)) > dt.datetime )
                year--;

        bool   thisresudial = (int(year/4.0) == year/4.0);
        double yeardays     = 1 + (year-1900)*365;
        double resudials    = 1 + int((year-1900-1)/4.0);

        dt.datetime -= yeardays+resudials;

        for ( month = 1; month<=12; month++ )
                if ( dt.datetime < daystomonth[month-1]+((thisresudial && month>=2)?1:0) )
                        break;

        dt.datetime -= daystomonthMin[month-1]+((thisresudial && month>2)?1:0);

        day = int(dt.datetime)+1;

        dt.datetime -= day-1;

        hour = int(dt.datetime*24.0);

        dt.datetime -= hour/24.0;

        minutes = int(dt.datetime*24.0*60.0);

        dt.datetime -= minutes/24.0/60.0;

        seconds = int(dt.datetime*24.0*60.0*60.0 + 0.5);
}

mstring          mDateToStr ( mdatetime dt )
{
        mstring str = "";

        int year, month, day, hour, minutes, seconds;
        mDecodeDateTime ( dt, year, month, day, hour, minutes, seconds );

        if (day >=10)
                str +=   mIntToStr(day)    + "." ;
                else
                str +=   "0" + mIntToStr(day) + "." ;

        if (month >=10)
                str +=   mIntToStr(month)  + "." ;
                else
                str +=   "0"+ mIntToStr(month)  + "." ;

        if (year >= 1000)
                str +=   mIntToStr(year) ;
                else
        if (year >= 100)
                str +=   "0" + mIntToStr(year);
                else
        if (year >= 10)
                str +=   "00" + mIntToStr(year);
                else
                str +=   "000" + mIntToStr(year);

        return str;
}

mstring          mDateTimeToStr ( mdatetime dt )
{
        mstring str = "";

        int year, month, day, hour, minutes, seconds;
        mDecodeDateTime ( dt, year, month, day, hour, minutes, seconds );

        if (day >=10)
                str +=   mIntToStr(day)    + "." ;
                else
                str +=   "0" + mIntToStr(day) + "." ;

        if (month >=10)
                str +=   mIntToStr(month)  + "." ;
                else
                str +=   "0"+ mIntToStr(month)  + "." ;

        if (year >= 1000)
                str +=   mIntToStr(year)   + " " ;
                else
        if (year >= 100)
                str +=   "0" + mIntToStr(year)   + " " ;
                else
        if (year >= 10)
                str +=   "00" + mIntToStr(year)   + " " ;
                else
                str +=   "000" + mIntToStr(year)   + " " ;

        if (hour >= 10)
                str +=   mIntToStr(hour)   + ":" ;
                else
                str +=   "0" + mIntToStr(hour)   + ":" ;

        if (minutes >= 10)
                str +=   mIntToStr(minutes)   + ":" ;
                else
                str +=   "0" + mIntToStr(minutes)   + ":" ;

        if (seconds >= 10)
                str +=   mIntToStr(seconds);
                else
                str +=   "0" + mIntToStr(seconds);

        return str;
}

mdatetime mStrToDateTime           ( mstring str  )
{
	mdatetime dt;

        str = mTrim ( str,false,false );

        int day     = mStrToInt(mSubString(str,0,2));
        int month   = mStrToInt(mSubString(str,3,2));
        int year    = mStrToInt(mSubString(str,6,4));

        int hour    = mStrToInt(mSubString(str,11,2));
        int minutes = mStrToInt(mSubString(str,14,2));
        int seconds = mStrToInt(mSubString(str,17,2));

        return mEncodeDateTime ( year, month, day, hour, minutes, seconds );
}

unsigned int     mGetTickCount   ( )
{
        return GetTickCount();
}

mdatetime mGetDateTime ( double * GTMdiff )
{
        #ifdef __linux__
                time_t t = time(NULL);
                struct tm * time = localtime(&t);
                mdatetime res = mEncodeDateTime ( time->tm_year+1900, time->tm_mon+1, time->tm_mday, time->tm_hour, time->tm_min, time->tm_sec );
        #else
                SYSTEMTIME time;
                GetLocalTime (&time);
                mdatetime res = mEncodeDateTime ( time.wYear, time.wMonth, time.wDay, time.wHour, time.wMinute, time.wSecond );
        #endif

        if ( GTMdiff )
        {
                #ifdef __linux__
                        struct tm * time2  = gmtime(&t);
                        mdatetime res2 = mEncodeDateTime ( time2->tm_year+1900, time2->tm_mon+1, time2->tm_mday, time2->tm_hour, time2->tm_min, time2->tm_sec );
                #else
                        SYSTEMTIME time2;
                        GetSystemTime(&time2);
                        mdatetime res2 = mEncodeDateTime ( time2.wYear, time2.wMonth, time2.wDay, time2.wHour, time2.wMinute, time2.wSecond );
                #endif
                *GTMdiff = (res.datetime - res2.datetime)*24;
        }

        return res;
}

double mGetJulianDay ( mdatetime datetime, double GMT )
{
        mdatetime dt ( datetime.datetime - GMT/24.0 );

        int year, month, day, hour, minute, second;
        mDecodeDateTime( dt, year, month, day, hour, minute, second );

        int a,b;

        // if month = January or February
        if ( month < 3 )
        {
                year--;
	        month += 12;
	}

	a = year / 100;

	// check for Julian or Gregorian calendar (starts Oct 4th 1582)
	if ( year > 1582 || ( year == 1582 && (month > 10 || (month == 10 && day >= 4))))
	{
	    // Gregorian calendar
	    b = 2 - a + (a / 4);
	}
	else
	{
	    // Julian calendar
	    b = 0;
	}

	// add a fraction of hours, minutes and secs to days
	double days = day + (double)(hour/24.0) + (double)(minute/1440.0) + (double)(second/86400.0);

	// now get the JD
	double JD = (int)(365.25 * (year + 4716)) +
	            (int)(30.6001 * (month + 1)) + days + b - 1524.5;

	return JD;
}

// 0 = sunday, 1 = monday, ...., 6 = saturday
unsigned int mGetDayOfWeek ( mdatetime datetime )
{
    unsigned int day;
    double JD;

    JD  = mGetJulianDay ( datetime, 0 );
    JD += 1.5;

    day = (int)JD % 7;

    return day;
}

double mComputeGMTDifference ( double lon, double lat )
{
        lon = fmod(lon+648000,1296000);
        if ( lon<0 ) lon = 1296000+lon;
        lon -= 648000;

        return float(lon) / 54000.0;
}

//******************************************************************************
// Memory Functions
//******************************************************************************

void mSetMemory( void *dest, unsigned int length, char value )
{
	#ifdef __BCPLUSPLUS__
		setmem (dest, length, value);
    	#else
		memset(dest,value,length);
	#endif
}

void mCopyMemory( void *dest, void *src, unsigned int length )
{
	#ifdef __BCPLUSPLUS__
		memcpy (dest, src, length);
    	#else
		memcpy (dest, src, length);
	#endif
}

//******************************************************************************
// File System Functions
//******************************************************************************

bool mCreateDirectory  ( mstring path )
{
	#ifdef __BCPLUSPLUS__
		return CreateDir ( AnsiString( path.c_str() ) );
    	#else
		return CreateDirectory( (LPCTSTR)path.c_str(), NULL );
	#endif
}

bool mRemoveDirectory  ( mstring path )
{
	#ifdef __BCPLUSPLUS__
		return RemoveDir ( AnsiString( path.c_str() ) );
    	#else
		return RemoveDirectory( (LPCTSTR)path.c_str() );
	#endif
}

bool mDirectoryExists  ( mstring path )
{
	#ifdef __BCPLUSPLUS__
		return DirectoryExists ( AnsiString( path.c_str() ) );
    	#else
		return _access ( (LPCTSTR)path.c_str(), 0 ) != -1;
	#endif
}

bool mForceDirectories  ( mstring path )
{
	#ifdef __BCPLUSPLUS__
        	#ifdef __linux__
		return ForceDirectories( AnsiString( path.c_str() ) );
                #else
                int    lngth = path.size ();
                char * str   = (char*)path.c_str();

                for ( int i = 0; i<lngth; i++ )
                if ( str[i] == '/' ) str[i] = '\\';
		return ForceDirectories( AnsiString( path.c_str() ) );
                #endif
    	#else
                int    lngth = path.size ();
                char * str   = (char*)path.c_str();

                bool    added  = false;
                bool    cfound = false;
                mstring ppath  = "";
                for ( int i = 0; i<lngth; i++ )
                {
                        if ( str[i] == ':' )
                        {
                                if ( !cfound )
                                {
                                        ppath = ppath + ':';
                                        i++;
                                        if ( i>=lngth || (str[i] != '\\' && str[i] != '/') )
                                                return false;
                                        cfound = true;
                                }
                                else
                                return false;
                                added = false;
                        }
                        else
                        if ( str[i] == '\\' || str[i] == '/' )
                        {
                                if ( cfound )
                                {
                                        if ( !mDirectoryExists( ppath ) )
                                                if ( !mCreateDirectory ( ppath ) ) return false;
                                }
                                else
                                return false;
                                added = false;
                        }
                        else
                        added = true;

                        ppath = ppath + str[i];
                }

                if ( added && cfound )
                        if ( !mDirectoryExists( ppath ) )
                                if ( !mCreateDirectory ( ppath ) ) return false;
                                        mCreateDirectory ( ppath );

                return cfound;
	#endif
}

void mDeleteFileFromDisk   ( mstring path )
{
	#ifdef __BCPLUSPLUS__
		DeleteFile( AnsiString( path.c_str() ) );
    	#else
		DeleteFile( (LPCTSTR)path.c_str() );
	#endif
}

mstring	mGetCurrentDir ( )
{
	#ifdef __BCPLUSPLUS__
		return mstring( GetCurrentDir ().c_str() );
	#else
		unsigned short lpBuffer[256];
		GetCurrentDirectory(256,(LPTSTR )lpBuffer);
		return  mstring ( (LPTSTR )lpBuffer );
	#endif
}

void	mSetCurrentDir ( mstring CurPath )
{
	#ifdef __BCPLUSPLUS__
		SetCurrentDir ( CurPath.c_str() );
    	#else
		SetCurrentDirectory( CurPath.c_str() );
	#endif
}

bool	mIsFileExist    ( mstring fname )
{
        if ( fname == "" ) return false;

    	FILE * file = fopen( (char*)fname.c_str(),"r");
    	if (file)
    	{
		fclose(file);
        	return true;
    	}
    	return false;
}

mdatetime    mGetFileModifyDate         ( mstring fname )
{
        unsigned short year = 1899, month = 12, day = 31;
        unsigned short hour = 0, min = 0, sec = 0;

        struct _stat buf;
        int result = _stat( fname.c_str(), &buf );
        if( result == 0 )
        {
                char dstr[30];
                mstring str;
                buf.st_mtime;
                sprintf( dstr, ctime( &buf.st_mtime ) );

                // year
                str.resize(4);
                str[0] = dstr[20];
                str[1] = dstr[21];
                str[2] = dstr[22];
                str[3] = dstr[23];
                str[4] = 0;
                year = mStrToInt(str);

                // month
                str.resize(3);
                str[0] = dstr[4];
                str[1] = dstr[5];
                str[2] = dstr[6];
                str[3] = 0;
                if ( str == "Jan" )
                        month = 1;
                        else
                if ( str == "Feb" )
                        month = 2;
                        else
                if ( str == "Mar" )
                        month = 3;
                        else
                if ( str == "Apr" )
                        month = 4;
                        else
                if ( str == "May" )
                        month = 5;
                        else
                if ( str == "Jun" )
                        month = 6;
                        else
                if ( str == "Jul" )
                        month = 7;
                        else
                if ( str == "Aug" )
                        month = 8;
                        else
                if ( str == "Sep" )
                        month = 9;
                        else
                if ( str == "Oct" )
                        month = 10;
                        else
                if ( str == "Now" )
                        month = 11;
                        else
                if ( str == "Dec" )
                        month = 12;

                // day
                str.resize(2);
                str[0] = dstr[8];
                str[1] = dstr[9];
                str[2] = 0;
                day = mStrToInt(str);

                // hour
                str.resize(2);
                str[0] = dstr[11];
                str[1] = dstr[12];
                str[2] = 0;
                hour = mStrToInt(str);

                // min
                str.resize(2);
                str[0] = dstr[14];
                str[1] = dstr[15];
                str[2] = 0;
                min = mStrToInt(str);

                // sec
                str.resize(2);
                str[0] = dstr[17];
                str[1] = dstr[18];
                str[2] = 0;
                sec = mStrToInt(str);
        }

        return mEncodeDateTime ( year, month, day, hour, min, sec );
}

mstring  mExtractPathOfFile ( mstring fname )
{
	#ifdef __BCPLUSPLUS__
		return mstring (ExtractFilePath ( fname.c_str() ).c_str());
    	#else
		char * strf = (char*)fname.c_str();
                char * str   = strrchr(strf,'\\');
		char * str2  = strrchr(strf,'/');
                if ( str && str2 && str<str2 ) str = str2;
                if (!str) str = str2;

                if (!str) return "";

		char prev   = str[0];
		str [0]     = '\0';
		mstring ret = mstring(strf);
		str [0]     = prev;
		return ret;
	#endif
}

mstring  mExtractNameOfFile ( mstring fname )
{
	#ifdef __BCPLUSPLUS__
		return mstring (ExtractFileName ( fname.c_str() ).c_str());
    	#else
		char * strf = (char*)fname.c_str();
                char * str   = strrchr(strf,'\\');
		char * str2  = strrchr(strf,'/');
                if ( str && str2 && str<str2 ) str = str2;
                if (!str) str = str2;

                if (!str) return fname;

		return mstring(&str[1]);
	#endif
}

class mOpenDirectoryData
{
public:
        HANDLE handle;
        WIN32_FIND_DATA filedata;
        bool   first;
};

void * mOpenDirectory ( mstring dirname )
{
        WIN32_FIND_DATA filedata;
        HANDLE handle = FindFirstFile( (char*)(dirname+"/*.*").c_str() , &filedata );

        if ( handle!= INVALID_HANDLE_VALUE )
        {
                mOpenDirectoryData * dt = new mOpenDirectoryData;
                dt->handle   = handle;
                dt->filedata = filedata;
                dt->first    = true;
                return dt;
        }
        else
        return NULL;
}

mstring mReadNextFile ( void * handle )
{
        if ( handle )
        {
                if ( ((mOpenDirectoryData*)handle)->first )
                {
                        ((mOpenDirectoryData*)handle)->first = false;
                        return ((mOpenDirectoryData*)handle)->filedata.cFileName;
                }
                else
                if ( FindNextFile( ((mOpenDirectoryData*)handle)->handle, &((mOpenDirectoryData*)handle)->filedata ) )
                        return ((mOpenDirectoryData*)handle)->filedata.cFileName;
                        else
                        return "";
        }
        else
        return "";
}

void         mCloseDirectory       ( void * handle )
{
        if ( handle )
        {
                FindClose ( ((mOpenDirectoryData*)handle)->handle );
                delete handle;
        }
}

double mSqr( double val , int ex )
{
        double res = 1;
        for ( int i = 0; i<ex; i++ )
                res *= val;
        return res;
}

#define COSSIN_ACCURACY         1000
#define ACOSSIN_ACCURACY        200000
#define ACOSSIN_BASE            100000
#define EXP_ACCURACY            400000
#define EXP_BASE                200000
#define LOG10_ACCURACY          400000

double costable   [360*COSSIN_ACCURACY+1];
double sintable   [360*COSSIN_ACCURACY+1];
double acostable  [ACOSSIN_ACCURACY+1];
double asintable  [ACOSSIN_ACCURACY+1];
double exptable   [EXP_ACCURACY+1];
double log10table [LOG10_ACCURACY+1];

double converter  = 1.0/COSSIN_ACCURACY/180.0*PI;
double converter2 = 1.0/ACOSSIN_BASE;
double converter3 = 50.0/EXP_BASE;
double converter4 = 10000000.0/LOG10_ACCURACY;

double PIMultTwo      = PI*2.0;
double backconverter  = COSSIN_ACCURACY*180.0/PI;
double backconverter2 = ACOSSIN_BASE/1.0;
double backconverter3 = EXP_BASE/50.0;
double backconverter4 = LOG10_ACCURACY/10000000.0;

int generatecostable ()
{
	int i;
        for ( i = 0; i<=360*COSSIN_ACCURACY; i++ )
        {
                costable[i] = cos(i*converter);
                sintable[i] = sin(i*converter);
        }
        for ( i = -ACOSSIN_BASE; i<=ACOSSIN_BASE; i++ )
        {
                acostable[i+ACOSSIN_BASE] = acos(i*converter2);
                asintable[i+ACOSSIN_BASE] = asin(i*converter2);
        }
        for ( i = -EXP_BASE; i<=EXP_BASE; i++ )
        {
                exptable [i+EXP_BASE] = exp(i*converter3);
        }
        for ( i = 1; i<=LOG10_ACCURACY; i++ )
        {
                log10table[i] = log10(i*converter4);
        }
		return 0;
}

int    doNothing = generatecostable ();

double mExp ( double val )
{
        if ( val<-50 )
                val = -50;
                else
        if ( val>50 )
                val = 50;

        double indexb = EXP_BASE+val*backconverter3;
        double weightup = indexb-int(indexb);

        return (1-weightup)*exptable [ int(indexb) ] +
                   weightup*exptable [ MinimumOf( EXP_ACCURACY, int(indexb)+1 ) ];
}

double mLog10 ( double val )
{
        if ( val<0.0000000001 )
                val = 0.0000000001;
                else
        if ( val>10000000.0 )
                val = 10000000.0;

        if ( val<1000 )
                return log10(val);

        double indexb = val*backconverter4;
        double weightup = indexb-int(indexb);

        return (1-weightup)*log10table [ int(indexb) ] +
                   weightup*log10table [ MinimumOf( LOG10_ACCURACY, int(indexb)+1 ) ];
}

double mCos           ( double val )
{
        if ( val<0 )
        {
                val = fmod(val, PIMultTwo );
                val = PIMultTwo + val;
        }
        else
        if ( val>PIMultTwo )
                val = fmod(val, PIMultTwo );

        return costable [ int(val*backconverter) ];
}

double mSin           ( double val )
{
        if ( val<0 )
        {
                val = fmod(val, PIMultTwo );
                val = PIMultTwo + val;
        }
        else
        if ( val>PIMultTwo )
                val = fmod(val, PIMultTwo );

        return sintable [ int(val*backconverter) ];
}

double mACos           ( double val )
{
        if ( val<-1 )
                val = -1;
                else
        if ( val>1 )
                val = 1;

        return acostable [ ACOSSIN_BASE+int(val*backconverter2) ];
}

double mASin           ( double val )
{
        if ( val<-1 )
                val = 0;
                else
        if ( val>1 )
                val = 1;
        return asintable [ ACOSSIN_BASE+int(val*backconverter2) ];
}

int mGetFloatDigit ( double val, int i ) // digits >> 543210.xxx
{
        return ( abs(int(val))/( int(mSqr(10,i)) ) ) % 10;
}

int mGetFloatFraction ( double val, int i ) // digits >> .12345
{
        return abs(int(val * int(mSqr(10,i)) )) % 10;
}

double mLog2 ( double val )
{
        static double log2const = log(2.0);
        return log(val)/log2const;
}

int mRound ( double val )
{
        if ( val>=0 )
                return val+0.5;
                else
                return val-0.5;
}

 extern int mRandomize()
{
        #ifdef __BCPLUSPLUS__
                Randomize();
                return RandSeed;
        #else
                // NOT COMPLETED
                return 0;
        #endif
}

 extern int mGetSeed ()
{
        #ifdef __BCPLUSPLUS__
                return RandSeed;
        #else
                // NOT COMPLETED
                return 0;
        #endif
}

 extern void mSetSeed (int seed)
{
        #ifdef __BCPLUSPLUS__
                RandSeed = seed;
        #else
                // NOT COMPLETED
        #endif
}

 extern int mRandom ( int nfrom, int nto )
{
        #ifdef __BCPLUSPLUS__
                return RandomRange ( nfrom, nto );
        #else
                // NOT COMPLETED
                return 0;
        #endif
}

 extern double mRandomGaussian( double mean, double stdev, double lowlimit, double uplimit )
{
        #ifdef __BCPLUSPLUS__
                double r = RandG(mean,stdev);
                if (r<lowlimit) r = lowlimit;
                else
                if (r>uplimit)  r = uplimit;
                return r;
        #else
                // NOT COMPLETED
                return 0;
        #endif
}


