#ifndef __GAME_ENVIRONMENT_SYSTEM_FUNCTIONS__
#define __GAME_ENVIRONMENT_SYSTEM_FUNCTIONS__

#ifdef __BCPLUSPLUS__
    #include <System.hpp>
#else
	#include <windows.h>
#endif

#include <stdlib.h>
#include <string>

//******************************************************************************
// Type Definitions
//******************************************************************************

#ifndef PI
        #define PI double(3.14159265358979323846)
#endif

#ifndef mstring
	#define mstring std::string
#endif

class mdatetime
{
public:
        double datetime;
        mdatetime ( );
        mdatetime ( double t );
};

//******************************************************************************
// Externs
//******************************************************************************

extern mstring          ApplicationDirectory;

//******************************************************************************
// Utility Functions
//******************************************************************************

#define                 MinimumOf(X,Y) (X < Y ? X : Y)
#define                 MaximumOf(X,Y) (X > Y ? X : Y)

//******************************************************************************
// String Functions
//******************************************************************************

extern mstring 	mIntToStr    ( int num );
extern int 	mStrToInt    ( mstring str );
extern int      mMixStrToInt ( mstring str );

extern mstring 	mFloatToStr  ( double num );
extern mstring 	mFloatToStrF ( double num, int numofdecimals );
extern double  	mStrToFloat  ( mstring str );

extern mstring	mTrim	     ( mstring str, bool intern = false, bool endonly = false );
extern mstring  mUpperCase   ( mstring str, bool turkishfirst = false );
extern mstring  mSubString   ( mstring str,int start,int count );

//******************************************************************************
// Date Time Functions
//******************************************************************************

extern mstring          mDateToStr 	     ( mdatetime dt );
extern mstring          mDateTimeToStr   ( mdatetime dt );
extern mdatetime        mStrToDateTime   ( mstring str  );

extern unsigned int     mGetTickCount    ( );	 // miliseconds passed since computer starts up
extern mdatetime        mGetDateTime     ( double * GTMdiff = NULL );    // days that have passed since 31.12.1899 00:00:00

extern mdatetime        mEncodeDateTime  (  int year, int month, int day,
                                                                int hour, int minute, int second );

extern void             mDecodeDateTime  (  mdatetime dt, int &year, int &month, int &day,
 						                int &hour, int &minute, int &second );

extern double           mGetJulianDay    ( mdatetime datetime, double GMT );
extern unsigned int     mGetDayOfWeek    ( mdatetime datetime ); // 0 = sunday ... 6 = saturday

extern double           mComputeGMTDifference ( double lon, double lat );

//******************************************************************************
// Memory Functions
//******************************************************************************

extern void	        mSetMemory  ( void *dest, unsigned int length, char value );
extern void	        mCopyMemory ( void *dest, void *src, unsigned int length );

//******************************************************************************
// File System Functions
//******************************************************************************

extern bool         mCreateDirectory     ( mstring path );
extern bool         mRemoveDirectory     ( mstring path );

extern bool	    mForceDirectories    ( mstring path );
extern void	    mDeleteFileFromDisk  ( mstring path );

extern mstring      mGetCurrentDir       ( );
extern void	    mSetCurrentDir       ( mstring CurPath );
extern bool	    mIsFileExist         ( mstring fname );
extern mdatetime    mGetFileModifyDate   ( mstring fname );

extern mstring      mExtractPathOfFile   ( mstring fname );
extern mstring      mExtractNameOfFile   ( mstring fname );

extern void *       mOpenDirectory       ( mstring dirname );
extern mstring      mReadNextFile        ( void * handle );
extern void         mCloseDirectory      ( void * handle );

//******************************************************************************
// Math Functions
//******************************************************************************

extern double mLog2          ( double val );
extern int    mRound         ( double val );
extern double mSqr           ( double val , int ex );

extern double mCos           ( double val ); // (fast) in radian   (0,001 degree accuracy)
extern double mSin           ( double val ); // (fast) in radian   (0,001 degree accuracy)
extern double mACos          ( double val ); // (fast) out radian  (0,00001 unit accuracy)
extern double mASin          ( double val ); // (fast) out radian  (0,00001 unit accuracy)
extern double mExp           ( double val ); // (fast) in -50..+50 (0,00001 unit accuracy)
extern double mLog10         ( double val ); // (fast) in 0..10000000 (25 unit accuracy after 1000)

extern int mGetFloatDigit    ( double val, int i ); // digits >> 543210.xx
extern int mGetFloatFraction ( double val, int i ); // digits >> xx.12345

extern int    mRandomize     ( );
extern int    mGetSeed       ( );
extern void   mSetSeed       ( int seed );
extern int    mRandom        ( int nfrom, int nto );
extern double mRandomGaussian( double mean, double stdev, double lowlimit, double uplimit );

#endif


