//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "mainf.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma link "MuhGLPanel"
#pragma resource "*.dfm"

#include "environment_visualizer.h"
#include "camera_controller.h"

#include "plugin_controller.h"
#include "simulation_objects.h"
#include "standard_objects.h"
#include "game_plugins.h"

// singleton visualizer objects ---------------------------------------------

TEnvironmentVisualizer          visualizer;
TCameraController               camcontroller;
TGameObjectLayer                layer;

// plugin create callbacks --------------------------------------------------

void * __stdcall create_TGameTerrainObject ( void * player )
{
        return new TGameTerrainObject( (TGameObjectLayer*)player );
}

void * __stdcall create_TGameRiverObject ( void * player )
{
        return new TGameRiverObject( (TGameObjectLayer*)player );
}

// borland's standard user interface object ---------------------------------

TForm1 *Form1;

//---------------------------------------------------------------------------

__fastcall TForm1::TForm1(TComponent* Owner)
        : TForm(Owner)
{
}
//---------------------------------------------------------------------------

void __fastcall TForm1::FormShow(TObject *Sender)
{
        ObjectPlugins.RegisterPlugin( "MeshObject", "TGameTerrainObject", create_TGameTerrainObject );
        ObjectPlugins.RegisterPlugin( "MeshObject", "TGameRiverObject", create_TGameRiverObject );

        visualizer.SetFilledTriangles   ( true );
        visualizer.SetLightingOn        ( true );

        visualizer.SetSunHeading        ( -90 );
        visualizer.SetSunPitch          (  45 );

        camcontroller.SetCamCoord       ( 0, -100, 30 );
        camcontroller.SetCamOrientation ( 0, -25, 0 );

        glGLResize(NULL);

        /*
        TGameTerrainObject * terrain = (TGameTerrainObject*)layer.CreateObject( "MeshObject", "TGameTerrainObject" );
        terrain->SetMeshFile( "meshes\\Terrain.3DS" );

        TGameRiverObject *   river1  = (TGameRiverObject*)layer.CreateObject( "MeshObject", "TGameRiverObject" );
        river1->SetMeshFile( "meshes\\River.3DS" );
        river1->SetBackFaceCulling( false );

        TGameRiverObject *   river2  = (TGameRiverObject*)layer.CreateObject( "MeshObject", "TGameRiverObject" );
        river2->SetMeshFile( "meshes\\Waterfall.3DS" );
        river2->SetBackFaceCulling( false );
        */
}

//---------------------------------------------------------------------------

void __fastcall TForm1::glGLResize(TObject *Sender)
{
        visualizer.SetScreenWidth   ( gl->Width );
        visualizer.SetScreenHeight  ( gl->Height );
        camcontroller.SetScreenSize ( gl->Width, gl->Height );
}
//---------------------------------------------------------------------------

void __fastcall TForm1::timerTimer(TObject *Sender)
{
        // initialize gl screen
        gl->PreRendering(0,0,gl->Width,gl->Height);

        // advance camera motion
        camcontroller.AdvanceCameraMotion ( );

        // advance environment time
        static long lasttime = GetTickCount();
        long currenttime = GetTickCount();
        layer.AdvanceTime( currenttime - lasttime );
        lasttime = currenttime;

        // initialize perspective rendering
        visualizer.InitializeRenderingPers  ( camcontroller.GetCamX(), camcontroller.GetCamY(), camcontroller.GetCamZ(),
                                              camcontroller.GetCamH(), camcontroller.GetCamP(), camcontroller.GetCamR(),
        			              camcontroller.GetCamHFov(),
                                              0.1, camcontroller.GetCamFarPlane(),
                                              true, true );

        // initialize sun light
        visualizer.InitializeLights ( );

        // render 3d objects
	visualizer.RenderObjects ( &layer );

        // finalize perspective rendering
        visualizer.FinalizeRendering ( );

        // finalize gl screen
        gl->PostRendering();

        // print frame rate
        StatusBar->Panels->Items[0]->Text = " Frame Rate = " +
                                            FloatToStrF( 1.0/visualizer.GetFrameTime(), ffFixed, 15, 0 ) +
                                            " frame/sec";
}
//---------------------------------------------------------------------------

void __fastcall TForm1::Close1Click(TObject *Sender)
{
        Close();
}
//---------------------------------------------------------------------------

void __fastcall TForm1::glMouseDown(TObject *Sender, TMouseButton Button,
      TShiftState Shift, int X, int Y)
{
        camcontroller.OnMouseDown( Button, Shift.Contains(ssShift), Shift.Contains(ssAlt), Shift.Contains(ssCtrl),
        			   Shift.Contains(ssDouble), X, Y );

}
//---------------------------------------------------------------------------

void __fastcall TForm1::glMouseMove(TObject *Sender, TShiftState Shift,
      int X, int Y)
{
        camcontroller.OnMouseMove( Shift.Contains(ssShift), Shift.Contains(ssAlt), Shift.Contains(ssCtrl), X, Y );
}
//---------------------------------------------------------------------------

void __fastcall TForm1::glMouseUp(TObject *Sender, TMouseButton Button,
      TShiftState Shift, int X, int Y)
{
        camcontroller.OnMouseUp( Button, Shift.Contains(ssShift), Shift.Contains(ssAlt), Shift.Contains(ssCtrl), X, Y );
}
//---------------------------------------------------------------------------

void __fastcall TForm1::None1Click(TObject *Sender)
{
        camcontroller.SetCameraMotion  ( cameramotionNone );
        camcontroller.SetMouseFunction ( 0, mousefunc3dRotate );                // left button
        camcontroller.SetMouseFunction ( 1, mousefunc3dForwardBackward );       // right button
        camcontroller.SetMouseFunction ( 2, mousefunc3dPan );                   // middle button
}
//---------------------------------------------------------------------------

void __fastcall TForm1::Fly1Click(TObject *Sender)
{
        camcontroller.SetCameraMotion( cameramotionFly );
        camcontroller.SetMouseFunction ( 0, mousefunc3dAccelerate );            // left button
        camcontroller.SetMouseFunction ( 1, mousefunc3dDeaccelerate );          // right button
        camcontroller.SetMouseFunction ( 2, mousefunc3dZeroSpeed );             // middle button
}
//---------------------------------------------------------------------------

void __fastcall TForm1::FilledTriangles1Click(TObject *Sender)
{
        visualizer.SetFilledTriangles ( FilledTriangles1->Checked );
}
//---------------------------------------------------------------------------

void __fastcall TForm1::LightingOn1Click(TObject *Sender)
{
        visualizer.SetLightingOn ( LightingOn1->Checked );
}
//---------------------------------------------------------------------------

void __fastcall TForm1::FromWestHalfUp1Click(TObject *Sender)
{
        visualizer.SetSunHeading ( -90 );
        visualizer.SetSunPitch   (  15 );
}
//---------------------------------------------------------------------------
void __fastcall TForm1::AtWestlow1Click(TObject *Sender)
{
        visualizer.SetSunHeading ( -90 );
        visualizer.SetSunPitch   (  45 );
}
//---------------------------------------------------------------------------

void __fastcall TForm1::AtWestHighAltitude1Click(TObject *Sender)
{
        visualizer.SetSunHeading ( -90 );
        visualizer.SetSunPitch   (  75 );
}
//---------------------------------------------------------------------------

void __fastcall TForm1::AtEastLowAltitude1Click(TObject *Sender)
{
        visualizer.SetSunHeading ( +90 );
        visualizer.SetSunPitch   (  15 );
}
//---------------------------------------------------------------------------

void __fastcall TForm1::AtEastMidAltitude1Click(TObject *Sender)
{
        visualizer.SetSunHeading ( +90 );
        visualizer.SetSunPitch   (  45 );
}
//---------------------------------------------------------------------------

void __fastcall TForm1::AtEastHighAltitude1Click(TObject *Sender)
{
        visualizer.SetSunHeading ( +90 );
        visualizer.SetSunPitch   (  75 );
}
//---------------------------------------------------------------------------


void __fastcall TForm1::Terrainvisible1Click(TObject *Sender)
{
        bool vis = Terrainvisible1->Checked;
        for ( int i = 0; i<layer.GetNumOfObjects(); i++ )
        {
                TGameObject     * std  = layer.GetObject(i);
                if ( std->IsAClass( "TGameTerrainObject" ) )
                        std->SetVisible( vis );
        }
}
//---------------------------------------------------------------------------

void __fastcall TForm1::Watervisible1Click(TObject *Sender)
{
        bool vis = Watervisible1->Checked;
        for ( int i = 0; i<layer.GetNumOfObjects(); i++ )
        {
                TGameObject     * std  = layer.GetObject(i);
                if ( std->IsAClass( "TGameRiverObject" ) )
                        std->SetVisible( vis );
        }
}
//---------------------------------------------------------------------------

void __fastcall TForm1::Entitiesvisible1Click(TObject *Sender)
{
        bool vis = Entitiesvisible1->Checked;
        for ( int i = 0; i<layer.GetNumOfObjects(); i++ )
        {
                TGameObject     * std  = layer.GetObject(i);
                if ( !std->IsAClass( "TGameTerrainObject" ) &&
                     !std->IsAClass( "TGameRiverObject" ) )
                        std->SetVisible( vis );
        }
}
//---------------------------------------------------------------------------

void __fastcall TForm1::Load1Click(TObject *Sender)
{
        if ( OpenDialog1->Execute() )
                layer.LoadFromFile( OpenDialog1->FileName.c_str() );
}
//---------------------------------------------------------------------------

void __fastcall TForm1::Save1Click(TObject *Sender)
{
        if ( SaveDialog1->Execute() )
                layer.SaveToFile( SaveDialog1->FileName.c_str() );
}
//---------------------------------------------------------------------------

