//---------------------------------------------------------------------------

#include "simulation_objects.h"

#ifdef __BCPLUSPLUS__

#else
	#define try 		{
	#define __finally 	}
	#define catch(x) 	}
#endif

//******************************************************************************
// TGameObject
//******************************************************************************

TGameObject::TGameObject  ( TGameObjectLayer * player )
{
        AddClassName ( "TGameObject" );

        id              = -1;
        name            = "Object";
	layer           = player;
        visible         = true;
        plugin          = NULL;
        parentobject    = NULL;
        Data            = NULL;
}

TGameObject::~TGameObject ( )
{
        while ( subobjects.size()>0 )
        	UnsetSubConnectedObject ( subobjects[subobjects.size()-1] );

        UnsetConnectedObject ( );


        // delete inheritance class names
        for ( int i = 0; i<classnames.size(); i++ )
                delete classnames[i];
        classnames.clear();
}

void TGameObject::AddClassName ( mstring classname )
{
        classnames.push_back( new mstring(classname) );
}

bool TGameObject::IsAClass ( mstring classname )
{
        for ( int i = 0; i<classnames.size(); i++ )
                if ( *classnames[i] == classname )
                        return true;
        return false;
}

bool TGameObject::GetVisible	   ( )
{
        return visible;
}

void TGameObject::SetVisible	   ( bool vis )
{
        visible = vis;
}

__int64 TGameObject::GetID ( )
{
        return id;
}

TGameObjectLayer * TGameObject::GetLayer ( )
{
        return layer;
}

TGameObjectPlugin * TGameObject::GetPlugin 	( )
{
        return plugin;
}

mstring TGameObject::GetName 	   ( )
{
        return name;
}

void    TGameObject::SetName 	   ( mstring pname )
{
        name = pname;
}

TGameObject * TGameObject::GetConnectedObject    ( )
{
        return parentobject;
}

bool TGameObject::SetConnectedObject   ( TGameObject * parent )
{
        if ( parent )
        {
                // check if object is in this later
                if ( parent->layer != layer )
                        return false;

                // check for circular referencing or self referencing connected objects
                TGameObject * searchparent = parent;
                while ( searchparent && searchparent != this )
                        searchparent = searchparent->GetConnectedObject();
                if ( searchparent == this )
                        return false;
        }
        UnsetConnectedObject ( );
        if ( parent )
        {
                parent->subobjects.push_back ( this );
                parentobject = parent;

                int rootobj = layer->FindRootObject( this );
                if (rootobj != -1)
                        layer->rootobjects.erase ( layer->rootobjects.begin()+rootobj );
        }
        return true;
}

void TGameObject::UnsetConnectedObject ( )
{
        if ( parentobject )
        {
                int rootobj = layer->FindRootObject( this );
                if (rootobj == -1)
                        layer->rootobjects.push_back ( this );

                for ( int i = 0; i < parentobject->subobjects.size(); i++ )
                        if ( parentobject->subobjects[i] == this )
                        {
                                parentobject->subobjects.erase ( parentobject->subobjects.begin()+i );
                                break;
                        }
                parentobject = NULL;
        }
}

int TGameObject::GetNumOfSubConnectedObjects ( )
{
        return subobjects.size();
}

TGameObject * TGameObject::GetSubConnectedObject ( int index )
{
        if ( index < 0 || index>=subobjects.size() ) return NULL;
        return subobjects[index];
}

bool TGameObject::SetSubConnectedObject ( TGameObject * subobj )
{
        if ( subobj )
                return subobj->SetConnectedObject ( this );
                else
                return false;
}

void TGameObject::UnsetSubConnectedObject ( TGameObject * subobj )
{
        if ( subobj && subobj->GetConnectedObject() == this )
                subobj->UnsetConnectedObject ( );
}

void TGameObject::AdvanceTime        ( unsigned int deltatime )
{
}

bool TGameObject::SaveToStream   ( TGameIOStream * stream )
{
        int version = 1;
        stream->Write ( &version );

        // note: ID, layer, plugin, parentobject and subobjects are not saved

        bool success = true;

        TGameMemoryStream * substream = new TGameMemoryStream () ;
        try
        {
                int i;
                unsigned long ssize;

                // save properties
                stream->WriteString ( &name );
                stream->Write       ( &visible );
        }
        __finally
        {
                delete substream;
        }

	return success;
}

bool TGameObject::LoadFromStream ( TGameIOStream * stream )
{
        int             i;
        unsigned long   ssize;
        unsigned long   baseadress;
        mstring         plugincategory;
        mstring         pluginname;

        int version = 1;
        stream->Read ( &version );

        // note: ID, layer, plugin, parentobject and subobjects are not loaded

        // assume that the file is loaded successfully
        bool success = true;

        // load properties
        stream->ReadString ( &name );
        stream->Read ( &visible );

	return success;
}

//******************************************************************************
// TGameObjectLayer
//******************************************************************************

TGameObjectLayer::TGameObjectLayer  ( )
{
        idcounter   = 1;
        GMT         = 2;
        currenttime = mStrToDateTime ( "07.04.2007 16:30:00" );
        starttime   = currenttime;
}

TGameObjectLayer::~TGameObjectLayer ( )
{
        // clear all objects
        ClearObjects ( );
}

int TGameObjectLayer::GetNumOfObjects ( )
{
        return objects.size();
}

TGameObject * TGameObjectLayer::GetObject ( int index  )
{
        if ( index<0 || index>=objects.size() ) return NULL;
        return objects[index];
}

int TGameObjectLayer::FindObject ( __int64 id )
{
        for ( int i = 0; i<objects.size(); i++ )
                if ( objects[i]->id == id )
                        return i;
        return -1;
}

int TGameObjectLayer::FindObject ( mstring name )
{
        for ( int i = 0; i<objects.size(); i++ )
                if ( objects[i]->name == name )
                        return i;
        return -1;
}

int TGameObjectLayer::FindObject ( TGameObject * obj  )
{
        for ( int i = 0; i<objects.size(); i++ )
                if ( objects[i] == obj )
                        return i;
        return -1;
}

int TGameObjectLayer::GetNumOfRootObjects ( )
{
        return rootobjects.size();
}

TGameObject * TGameObjectLayer::GetRootObject ( int index  )
{
        if ( index<0 || index>=rootobjects.size() ) return NULL;
        return rootobjects[index];
}

int TGameObjectLayer::FindRootObject ( __int64 id )
{
        for ( int i = 0; i<rootobjects.size(); i++ )
                if ( rootobjects[i]->id == id )
                        return i;
        return -1;
}

int TGameObjectLayer::FindRootObject ( TGameObject * obj )
{
        for ( int i = 0; i<rootobjects.size(); i++ )
                if ( rootobjects[i] == obj )
                        return i;
        return -1;
}

TGameObject *   TGameObjectLayer::CreateObject ( mstring plugincategory, mstring pluginname )
{
        TGameObjectPlugin * plg = ObjectPlugins.FindPlugin ( plugincategory, pluginname );
        if ( plg  )
        {
                TGameObject * obj = (TGameObject *)plg->CreateObject( this );
                objects.push_back ( obj );
                rootobjects.push_back ( obj );
                obj->layer = this;
                obj->plugin = plg;
                obj->id = idcounter++;
                return obj;
        }
        else
        return NULL;
}

void            TGameObjectLayer::DeleteObject ( int index, bool deletesubobjects )
{
        if ( index<0 || index>=objects.size() ) return;
        TGameObject * obj = objects[index];

        if ( deletesubobjects )
                while ( obj->GetNumOfSubConnectedObjects()>0 )
                        DeleteObject ( FindObject(obj->GetSubConnectedObject(0)), deletesubobjects );

        index = FindObject( obj );
        if ( index != -1 )
                objects.erase ( objects.begin()+index );

        index = FindRootObject( obj );
        if ( index != -1 )
                rootobjects.erase ( rootobjects.begin()+index );

        delete obj;

        // burda bir daha sil nk
        // delete obj komutundaki destructor rootobjects'e kendini geri atabiliyor.
        index = FindRootObject( obj );
        if ( index != -1 )
                rootobjects.erase ( rootobjects.begin()+index );
}

void TGameObjectLayer::ClearObjects ( )
{
        while ( objects.size()>0 )
        	DeleteObject ( objects.size()-1 );
        idcounter = 1;
}

double          TGameObjectLayer::GetGMTDifference ( )
{
        return GMT;
}

void            TGameObjectLayer::SetGMTDifference ( double gmt )
{
        GMT = gmt;
}

mdatetime       TGameObjectLayer::GetCurrentTime ( )
{
        return currenttime;
}
void            TGameObjectLayer::SetCurrentTime ( mdatetime pcurrenttime )
{
        currenttime = pcurrenttime;
}

mdatetime       TGameObjectLayer::GetStartTime   ( )
{
        return starttime;
}

void            TGameObjectLayer::SetStartTime   ( mdatetime pstarttime )
{
        starttime = pstarttime;
}

bool TGameObjectLayer::SaveToStream   ( TGameIOStream * stream )
{
        int magicnumber = 67070475;
        stream->Write ( &magicnumber );

        int version = 1;
        stream->Write ( &version );

        // write properties of layer
        stream->Write       ( &currenttime.datetime );
        stream->Write       ( &starttime.datetime );
        stream->Write       ( &GMT );

        int i;
        bool success = true;

        TGameMemoryStream * substream = new TGameMemoryStream () ;
        try
        {
                unsigned long ssize;

                // save number of objects
                int numofobj = objects.size();
                stream->Write ( &numofobj );

                // save objects
                for ( i = 0; i<objects.size(); i++ )
                {
                        substream->SetSize(0);
                        substream->Seek_Set(0);
                        if ( objects[i]->SaveToStream( substream ) )
                        {
                                // get size of object stream
                                ssize = substream->GetSize();

                                // write object size, ID and plugin name
                                stream->Write       ( &ssize );
                                stream->Write       ( &objects[i]->id );

                                stream->WriteString ( &objects[i]->plugin->GetCategory() );
                                stream->WriteString ( &objects[i]->plugin->GetName() );

                                // write object data
                                if ( ssize>0 )
				        stream->Write ( substream->GetMemory(), ssize );
                        }
                        else
                        {
                                success = false;
                                break;
                        }
                }
        }
        __finally
        {
                delete substream;
        }

        if ( success )
        {
                // save object links
                __int64 sid, pid;
                for ( i = 0; i<objects.size(); i++ )
                {
                        if ( objects[i]->GetConnectedObject() )
                        {
                                sid = objects[i]->GetID();
                                pid = objects[i]->GetConnectedObject()->GetID();
                                stream->Write ( &sid );
                                stream->Write ( &pid );
                        }
                }
                sid = -1;
                pid = -1;
                stream->Write ( &sid );
                stream->Write ( &pid );
        }

	return success;
}

bool TGameObjectLayer::LoadFromStream ( TGameIOStream * stream )
{
        int             i;
        __int64         sid, pid, playerid;
        unsigned long   ssize;
        unsigned long   baseadress;
        mstring         plugincategory;
        mstring         pluginname;
        TGameObject *   obj;

        // clear all the objects
        ClearObjects( );

        // check signature of file
        int magicnumber = 0;
        stream->Read  ( &magicnumber );
        if ( magicnumber != 67070475 ) return false;

        // load version number
        int version = 1;
        stream->Read ( &version );

        // read properties of layer
        stream->Read        ( &currenttime.datetime );
        stream->Read        ( &starttime.datetime );
        stream->Read        ( &GMT );

        // assume that the file is loaded successfully
        bool success = true;

        // load number of objects
        int numofobj = 0;
        stream->Read ( &numofobj );

        // load objects
        __int64 maxidcounter = 1;
        for ( i = 0; i<numofobj; i++ )
        {
                // read object size, ID and plugin name
                stream->Read       ( &ssize );
                stream->Read       ( &sid );
                stream->ReadString ( &plugincategory );
                stream->ReadString ( &pluginname );

                // compute stream position of next data block
                baseadress = stream->GetPos() + ssize;

                // create object
                obj = CreateObject( plugincategory, pluginname );

                // load object if plugin is found
                if ( obj )
                {
                        obj->id = sid;
                        if ( maxidcounter<=sid )
                                maxidcounter = sid+1;

                        if ( !obj->LoadFromStream( stream ) )
                                success = false;
                }
                else
                success = false;

                // jump to next data block
                stream->Seek_Set(baseadress);
        }
        idcounter = maxidcounter;

        // load object links
        do
        {
                if ( !stream->Read ( &sid ) ) { success = false; break; }
                if ( !stream->Read ( &pid ) ) { success = false; break; }
                if ( sid != -1 && pid != -1 )
                {
                        int sindex   = FindObject ( sid );
                        int parindex = FindObject ( pid );
                        if ( sindex != -1 && parindex != -1 )
                                objects[sindex]->SetConnectedObject( objects[parindex] );
                                else
                                success = false;
                }
        }
        while ( sid != -1 || pid != -1 );

	return success;
}

bool TGameObjectLayer::LoadFromFile ( mstring fname )
{
        TGameFileStream * file = new TGameFileStream();
        try
        {
                if ( file->OpenFileForRead(fname) )
                {
                        TGameMemoryStream memstream;
                        memstream.SetSize(file->GetSize());
                        file->Read(memstream.GetMemory(),file->GetSize());

                	return LoadFromStream ( &memstream );
                }
                else
                return false;
        }
        __finally
        {
                delete file;
        }
	return false;
}

bool TGameObjectLayer::SaveToFile ( mstring fname )
{
        TGameFileStream * file = new TGameFileStream();
        try
        {
                if ( file->OpenFileForCreate(fname) )
                        return SaveToStream ( file );
                        else
                        return false;
        }
        __finally
        {
                delete file;
        }
        return false;
}

void TGameObjectLayer::AdvanceTime        ( unsigned int deltatime )
{
        currenttime.datetime = currenttime.datetime + deltatime/1000.0/60.0/60.0/24.0;

        for ( int i = 0; i<objects.size(); i++ )
                objects[i]->AdvanceTime( deltatime );
}

