#ifndef __GAME_ENVIRONMENT_SIM_OBJECT_ACCESS__
#define __GAME_ENVIRONMENT_SIM_OBJECT_ACCESS__

//******************************************************************************
// INCLUDES
//******************************************************************************

#include "main_functions.h"
#include "vector_operations.h"
#include "io_streams.h"
 #include "plugin_controller.h"

#include <vector>

//******************************************************************************
// Classes
//******************************************************************************

class TGameObject;
class TGameObjectLayer;

//******************************************************************************
// TGameObject
//******************************************************************************

class TGameObject
{
	friend TGameObjectLayer;
protected:
        __int64                           id;
        mstring                           name;
	TGameObjectLayer * 	          layer;
        TGameObjectPlugin *               plugin;

        bool                              visible;
        TGameObject *                     parentobject;
        std::vector<TGameObject*>         subobjects;

        // class inheritance relations build in compile-time
        std::vector<mstring*>          classnames;

        void         AddClassName       ( mstring classname );

        virtual void AdvanceTime        ( unsigned int deltatime ); // in miliseconds
public:
		TGameObject  ( TGameObjectLayer * player );
        virtual ~TGameObject ( );

        void *                  Data;

        __int64			GetID	             ( );
	TGameObjectLayer * 	GetLayer 	     ( );
	TGameObjectPlugin * 	GetPlugin	     ( );
        bool                    IsAClass             ( mstring classname );

	mstring                 GetName 	     ( );
	virtual void            SetName 	     ( mstring pname );

	bool                    GetVisible   	     ( );
	virtual void            SetVisible	     ( bool vis );

        TGameObject *           GetConnectedObject   ( );
        bool                    SetConnectedObject   ( TGameObject * parent );
        void                    UnsetConnectedObject ( );

        int                     GetNumOfSubConnectedObjects ( );
        TGameObject *           GetSubConnectedObject       ( int index );
        bool                    SetSubConnectedObject       ( TGameObject * subobj );
        void                    UnsetSubConnectedObject     ( TGameObject * subobj );

        virtual bool            SaveToStream         ( TGameIOStream * stream );
        virtual bool            LoadFromStream       ( TGameIOStream * stream );
};

//******************************************************************************
// TGameObjectLayer
//******************************************************************************

class TGameObjectLayer
{
        friend TGameObject;
protected:
        __int64                         idcounter;

        double                          GMT;
        mdatetime 	                currenttime;
        mdatetime                       starttime;

	std::vector<TGameObject*>       objects;
	std::vector<TGameObject*>       rootobjects;
public:
	TGameObjectLayer  ( );
        virtual ~TGameObjectLayer ( );

        // simulation time
	double          GetGMTDifference         ( );
	void            SetGMTDifference         ( double gmt );
        mdatetime       GetCurrentTime           ( );
        void            SetCurrentTime           ( mdatetime pcurrenttime );
        mdatetime       GetStartTime             ( );
        void            SetStartTime             ( mdatetime pstarttime );
        void            AdvanceTime              ( unsigned int deltatime ); // in miliseconds

        // objects
	int 		GetNumOfObjects          ( );
        TGameObject *   GetObject		 ( int index  );
        int		FindObject	 	 ( __int64 id );
        int		FindObject	 	 ( TGameObject * obj );
        int		FindObject	 	 ( mstring name );
        void            DeleteObject             ( int index, bool deletesubobjects = true );
        void            ClearObjects             ( );

        TGameObject *   CreateObject             ( mstring plugincategory, mstring pluginname );

        // objects that are not children of any other objects
	int 		GetNumOfRootObjects 	 ( );
        TGameObject *   GetRootObject	     	 ( int index  );
        int		FindRootObject	 	 ( __int64 id );
        int		FindRootObject	 	 ( TGameObject * obj );

        // get land surface elevation (if on water, water bottom elevation given)
        float           GetLandSurfaceElevation  ( float x, float y );

        // get water surface elevation (if not on water, land surface elevation given)
        float           GetWaterSurfaceElevation ( float x, float y );

        // assignment, save and load
	virtual bool 	LoadFromStream 	         ( TGameIOStream * stream );
	virtual bool 	SaveToStream   	         ( TGameIOStream * stream );

        bool            LoadFromFile             ( mstring fname );
        bool            SaveToFile               ( mstring fname );
};

#endif

