#include "standard_objects.h"
#include <math.h>

//******************************************************************************
// TGameMeshObject
//******************************************************************************

TGameMeshObject::TGameMeshObject ( TGameObjectLayer * player ) : TGameObject ( player )
{
        AddClassName ( "TGameMeshObject" );

        // coord & ori
        objx = 0;
        objy = 0;
        objz = 0;
        objh = 0;
        objp = 0;
        objr = 0;

        // bounding volume
        entireBoundingVolume.type    = 0;
        entireBoundingVolume.centerx = 0;
        entireBoundingVolume.centery = 0;
        entireBoundingVolume.centerz = 0;
        entireBoundingVolume.sizex   = 0;
        entireBoundingVolume.sizey   = 0;
        entireBoundingVolume.sizez   = 0;
        entireBoundingVolume.minx    = 0;
        entireBoundingVolume.miny    = 0;
        entireBoundingVolume.minz    = 0;
        entireBoundingVolume.maxx    = 0;
        entireBoundingVolume.maxy    = 0;
        entireBoundingVolume.maxz    = 0;

        // mesh
        mesh         = NULL;
        uniqueMesh   = false;

        // mesh animation
        currentFrame = 0;
        targetFrame  = 0;
        useSmoothFrameTransition = true;
        frameTransitionRatio     = 0;

        // general visualization options
        backFaceCulling   = true;
        opacity         = 1.0;
}

TGameMeshObject::~TGameMeshObject ( )
{
        boundingVolumes.clear();
        DeallocateMesh();
}

// coordinate & orientation of object
float    TGameMeshObject::GetX ()
{
        return objx;
}

float    TGameMeshObject::GetY ()
{
        return objy;
}

float    TGameMeshObject::GetZ ()
{
        return objz;
}

float    TGameMeshObject::GetH ()
{
        return objh;
}

float    TGameMeshObject::GetP ()
{
        return objp;
}

float    TGameMeshObject::GetR ()
{
        return objr;
}

void     TGameMeshObject::SetCoordinate  ( float x, float y, float z )
{
        objx = x;
        objy = y;
        objz = z;
}

void     TGameMeshObject::SetOrientation ( float h, float p, float r )
{
        objh = h;
        objp = p;
        objr = r;
}

// bounding volume
void TGameMeshObject::GenerateEntireBoundingVolume()
{
        if ( boundingVolumes.size() == 0 )
        {
                entireBoundingVolume.type    = 0;
                entireBoundingVolume.centerx = 0;
                entireBoundingVolume.centery = 0;
                entireBoundingVolume.centerz = 0;
                entireBoundingVolume.sizex   = 0;
                entireBoundingVolume.sizey   = 0;
                entireBoundingVolume.sizez   = 0;
                entireBoundingVolume.minx    = 0;
                entireBoundingVolume.miny    = 0;
                entireBoundingVolume.minz    = 0;
                entireBoundingVolume.maxx    = 0;
                entireBoundingVolume.maxy    = 0;
                entireBoundingVolume.maxz    = 0;
        }
        else
        {
                bool first = true;
                float minx = 0, miny = 0, minz = 0, maxx = 0, maxy = 0, maxz = 0;

                for ( int i = 0; i<boundingVolumes.size(); i++ )
                {
                        if ( first || boundingVolumes[i].minx<minx ) minx = boundingVolumes[i].minx;
                        if ( first || boundingVolumes[i].miny<miny ) miny = boundingVolumes[i].miny;
                        if ( first || boundingVolumes[i].minz<minz ) minz = boundingVolumes[i].minz;
                        if ( first || boundingVolumes[i].maxx>maxx ) maxx = boundingVolumes[i].maxx;
                        if ( first || boundingVolumes[i].maxy>maxy ) maxy = boundingVolumes[i].maxy;
                        if ( first || boundingVolumes[i].maxz>maxz ) maxz = boundingVolumes[i].maxz;
                        first = false;
                }

                entireBoundingVolume.type    = 0;
                entireBoundingVolume.centerx = (minx+maxx)/2;
                entireBoundingVolume.centery = (miny+maxy)/2;
                entireBoundingVolume.centerz = (minz+maxz)/2;
                entireBoundingVolume.sizex   = fabs(minx-maxx);
                entireBoundingVolume.sizey   = fabs(miny-maxy);
                entireBoundingVolume.sizez   = fabs(minz-maxz);
                entireBoundingVolume.minx    = minx;
                entireBoundingVolume.miny    = miny;
                entireBoundingVolume.minz    = minz;
                entireBoundingVolume.maxx    = maxx;
                entireBoundingVolume.maxy    = maxy;
                entireBoundingVolume.maxz    = maxz;
        }
}

TGameMeshBoundingVolume   TGameMeshObject::GetEntireBoundingVolume       ( )
{
        return entireBoundingVolume;
}

int                       TGameMeshObject::GetNumberOfBoundingVolumes    ( )
{
        return boundingVolumes.size();
}

TGameMeshBoundingVolume TGameMeshObject::GetBoundingVolume            ( int index )
{
        if ( index<0 || index>=boundingVolumes.size() ) return entireBoundingVolume;
        return boundingVolumes[index];
}

// mesh
TMesh *  TGameMeshObject::GetMesh                        ( )
{
        return mesh;
}

bool     TGameMeshObject::GetUniqueMesh                   ( )
{
        return uniqueMesh;
}

void     TGameMeshObject::SetUniqueMesh           ( bool unique )
{
        if ( uniqueMesh == unique ) return;
        uniqueMesh = unique;
}

// mesh animation
int      TGameMeshObject::GetCurrentFrame                ( )
{
        return currentFrame;
}

void     TGameMeshObject::SetCurrentFrame                ( int frmNo )
{
        if ( currentFrame == frmNo ) return;
        if ( !mesh ) return;
        if ( frmNo<0 || frmNo>=mesh->GetNumOfFrames() ) return;
        currentFrame = frmNo;
}

int      TGameMeshObject::GetTargetFrame                 ( )
{
        return targetFrame;
}

void     TGameMeshObject::SetTargetFrame                 ( int frmNo )
{
        if ( targetFrame == frmNo ) return;
        if ( !mesh ) return;
        if ( frmNo<0 || frmNo>=mesh->GetNumOfFrames() ) return;
        targetFrame = frmNo;
}

bool     TGameMeshObject::GetUseSmoothFrameTransition    ( )
{
        return useSmoothFrameTransition;
}

void     TGameMeshObject::SetUseSmoothFrameTransition    ( bool smt )
{
        if ( useSmoothFrameTransition == smt ) return;
        useSmoothFrameTransition = smt;
}

float    TGameMeshObject::GetFrameTransitionRatio        ( )
{
        return frameTransitionRatio;
}

void     TGameMeshObject::SetFrameTransitionRatio        ( float ratio )
{
        if ( frameTransitionRatio == ratio ) return;
        if ( ratio<0 ) ratio = 0;
        if ( ratio>1 ) ratio = 1;
        frameTransitionRatio = ratio;
}

// general visualization options
bool     TGameMeshObject::GetBackFaceCulling                 ( )
{
        return backFaceCulling;
}

void     TGameMeshObject::SetBackFaceCulling                 ( bool bfacecul )
{
        if ( backFaceCulling == bfacecul ) return;
        backFaceCulling = bfacecul;
}

float    TGameMeshObject::GetOpacity                     ( )
{
        return opacity;
}

void     TGameMeshObject::SetOpacity                     ( float opa )
{
        if ( opacity == opa ) return;
        if ( opa<0 ) opa = 0;
        if ( opa>1 ) opa = 1;
        opacity = opa;
}

// custom 3D OpenGL visualisation methods for non-mesh parts (points are in camera coordinates)
void     TGameMeshObject::CustomView3dGL ( float camh, float camp, float camr,
                                           float camhfov, float camvfov, float camfarplane,
                                           int screenwidth, int screenheight )
{
}

void TGameMeshObject::DeallocateMesh ( )
{
       	if ( mesh )
        {
                if ( uniqueMesh )
                        delete mesh;
                        else
                        MeshLoader.UnlinkMesh ( mesh );
                mesh = NULL;
        }
}

void TGameMeshObject::LoadMeshFromStream           ( TGameIOStream * stream )
{
        bool hasmesh = false;
        stream->Read(&hasmesh);
        if ( hasmesh )
        {
                mesh = new TMesh();
                mesh->LoadFromStream ( stream );
        }
}

void TGameMeshObject::SaveMeshToStream             ( TGameIOStream * stream )
{
        bool hasmesh = mesh != NULL;
        stream->Write(&hasmesh);
        if ( mesh )
                mesh->SaveToStream ( stream );
}

// save and load
bool 		TGameMeshObject::LoadFromStream 	   ( TGameIOStream * stream )
{
	TGameObject::LoadFromStream ( stream );
        int version = 1;
        stream->Read ( &version );

        // bounding volume
        stream->Read( &entireBoundingVolume, sizeof(entireBoundingVolume) );
        TGameMeshBoundingVolume bv;
        int i;
        int numofbv = 0;
        stream->Read ( &numofbv );
        boundingVolumes.clear();
        for ( i = 0; i<numofbv; i++ )
        {
                stream->Read( &bv, sizeof(entireBoundingVolume) );
                boundingVolumes.push_back(bv);
        }

        // mesh
        DeallocateMesh ( );
        stream->Read ( &uniqueMesh );
        if ( uniqueMesh )
                LoadMeshFromStream ( stream );

        // mesh animation
        stream->Read ( &currentFrame );
        stream->Read ( &targetFrame );
        stream->Read ( &useSmoothFrameTransition );
        stream->Read ( &frameTransitionRatio );

        // general visualization options
        stream->Read ( &backFaceCulling );
        stream->Read ( &opacity );

        return true;
}

bool 		TGameMeshObject::SaveToStream   	   ( TGameIOStream * stream )
{
	TGameObject::SaveToStream ( stream );
        int version = 1;
        stream->Write ( &version );

        // bounding volume
        stream->Write( &entireBoundingVolume, sizeof(entireBoundingVolume) );
        int i;
        int numofbv = boundingVolumes.size();
        stream->Write ( &numofbv );
        for ( i = 0; i<numofbv; i++ )
                stream->Write( &boundingVolumes[i], sizeof(entireBoundingVolume) );

        // mesh
        stream->Write ( &uniqueMesh );
        if ( uniqueMesh )
                SaveMeshToStream ( stream );

        // mesh animation
        stream->Write ( &currentFrame );
        stream->Write ( &targetFrame );
        stream->Write ( &useSmoothFrameTransition );
        stream->Write ( &frameTransitionRatio );

        // general visualization options
        stream->Write ( &backFaceCulling );
        stream->Write ( &opacity );

        return true;
}


