#ifndef __GAME_ENVIRONMENT_STANDARD_OBJECT_INTERFACES__
#define __GAME_ENVIRONMENT_STANDARD_OBJECT_INTERFACES__

//******************************************************************************
// INCLUDES
//******************************************************************************

#include "simulation_objects.h"
#include "MeshLoader.h"

//******************************************************************************
// Classes
//******************************************************************************

class TGameMeshObject;

//**********************************************************************
// TGameMeshBoundingVolume
//**********************************************************************
typedef struct
{
public:
        int   type;                             // 0 = box
                                                // 1 = sphere
                                                // 2 = cylinder (facing x)
                                                // 3 = cylinder (facing y)
                                                // 4 = cylinder (facing z)
        float centerx, centery, centerz;        // center in local coordinates
        float sizex, sizey, sizez;              // size of volume in x,y,z axis
        float minx,miny,minz;                   // minimum values in local coordinates
        float maxx,maxy,maxz;                   // maksimum values in local coordinates
}
TGameMeshBoundingVolume;

//******************************************************************************
// TGameMeshObject
//******************************************************************************

class TGameMeshObject : public TGameObject
{
protected:
        // coordinate & orientation of object
        float objx, objy, objz, objh, objp, objr;

        // bounding volume
        TGameMeshBoundingVolume                 entireBoundingVolume;
        std::vector<TGameMeshBoundingVolume>    boundingVolumes;

        // mesh
        TMesh *                                 mesh;
        bool                                    uniqueMesh;

        // mesh animation
        int                                     currentFrame;
        int                                     targetFrame;
        bool                                    useSmoothFrameTransition;
        float                                   frameTransitionRatio;

        // general visualization options
        bool                                    backFaceCulling;
        float                                   opacity;

        virtual void DeallocateMesh               ( );
        virtual void LoadMeshFromStream           ( TGameIOStream * stream );
        virtual void SaveMeshToStream             ( TGameIOStream * stream );
        virtual void GenerateEntireBoundingVolume ( );
public:
                 TGameMeshObject ( TGameObjectLayer * player );
        virtual ~TGameMeshObject ( );

        // coordinate & orientation of object
        virtual float    GetX ();
        virtual float    GetY ();
        virtual float    GetZ ();
        virtual float    GetH ();
        virtual float    GetP ();
        virtual float    GetR ();
        virtual void     SetCoordinate  ( float x, float y, float z );
        virtual void     SetOrientation ( float h, float p, float r );

        // bounding volume
        virtual TGameMeshBoundingVolume   GetEntireBoundingVolume       ( );
        virtual int                       GetNumberOfBoundingVolumes    ( );
        virtual TGameMeshBoundingVolume   GetBoundingVolume             ( int index );

        // mesh
        virtual TMesh *  GetMesh                        ( );    // can be NULL
        virtual bool     GetUniqueMesh                  ( );    // if true, mesh will be internal to this object,
                                                                // and saved/load to/from stream
        virtual void     SetUniqueMesh                  ( bool unique );

        // mesh animation
        virtual int      GetCurrentFrame                ( );
        virtual void     SetCurrentFrame                ( int frmNo );
        virtual int      GetTargetFrame                 ( );
        virtual void     SetTargetFrame                 ( int frmNo );
        virtual bool     GetUseSmoothFrameTransition    ( );
        virtual void     SetUseSmoothFrameTransition    ( bool smt );
        virtual float    GetFrameTransitionRatio        ( );
        virtual void     SetFrameTransitionRatio        ( float ratio );

        // general visualization options
        virtual bool     GetBackFaceCulling             ( );
        virtual void     SetBackFaceCulling             ( bool bfacecul );
        virtual float    GetOpacity                     ( );
        virtual void     SetOpacity                     ( float opa );

        // custom 3D OpenGL visualisation methods for non-mesh parts (points are in camera coordinates)
        virtual void     CustomView3dGL                 ( float camh, float camp, float camr,
                                                          float camhfov, float camvfov, float camfarplane,
                                                          int screenwidth, int screenheight );

        // save and load
	virtual bool     LoadFromStream 	   ( TGameIOStream * stream );
	virtual bool 	 SaveToStream   	   ( TGameIOStream * stream );
};

#endif
