#include "vector_operations.h"

#include <math.h>

#ifdef __BCPLUSPLUS__

#else
	#define try 		{
	#define __finally 	}
	#define catch(x) 	}
#endif

//******************************************************************************
// Extern Instances
//******************************************************************************

TPolygonOperations 	PolygonOp;
TVectorOperations 	VectorOp;

//********************************************************
// TVertex2i
//********************************************************

TVertex2i::TVertex2i ( )
{
        x = 1;
        y = 1;
}

TVertex2i::TVertex2i ( int nx, int ny )
{
        x = nx;
        y = ny;
}

//********************************************************
// TVertex2d
//********************************************************

TVertex2d::TVertex2d ( )
{
        x = 1;
        y = 1;
}

TVertex2d::TVertex2d ( double nx, double ny )
{
        x = nx;
        y = ny;
}

//********************************************************
// TVertex3d
//********************************************************

TVertex3d::TVertex3d ( )
{
        x = 1;
        y = 1;
        z = 1;
}

TVertex3d::TVertex3d ( double nx, double ny, double nz )
{
        x = nx;
        y = ny;
        z = nz;
}

//********************************************************
// TVertex2f
//********************************************************

TVertex2f::TVertex2f ( )
{
        x = 1;
        y = 1;
}

TVertex2f::TVertex2f ( float nx, float ny )
{
        x = nx;
        y = ny;
}

//********************************************************
// TVertex3f
//********************************************************

TVertex3f::TVertex3f ( )
{
        x = 1;
        y = 1;
        z = 1;
}

TVertex3f::TVertex3f ( float nx, float ny, float nz )
{
        x = nx;
        y = ny;
        z = nz;
}

//********************************************************
// TVertex4f
//********************************************************

TVertex4f::TVertex4f ( )
{
        x = 1;
        y = 1;
        z = 1;
        w = 1;
}

TVertex4f::TVertex4f ( float nx, float ny, float nz, float nw )
{
        x = nx;
        y = ny;
        z = nz;
        w = nw;
}

//********************************************************
// TColor3i
//********************************************************

unsigned int mStructColoriToIntColor ( TColor3i &col )
{
        return ((unsigned int)col.r) |
               (((unsigned int)col.g)<<8) |
               (((unsigned int)col.b)<<16);
}

unsigned int mStructColorfToIntColor ( TColor3f &col )
{
        return ((unsigned int)(col.r*255)) |
               (((unsigned int)(col.g*255))<<8) |
               (((unsigned int)(col.b*255))<<16);
}

TColor3i mIntColorToStructColori ( unsigned int &col )
{
        return TColor3i ( (unsigned char)col,
                           (unsigned char)(col>>8),
                           (unsigned char)(col>>16) );
}

TColor3f mIntColorToStructColorf ( unsigned int &col )
{
        return TColor3f ( ((unsigned char)col)*255.0,
                           ((unsigned char)(col>>8))*255.0,
                           ((unsigned char)(col>>16))*255.0 );
}

TColor3i::TColor3i ( )
{
        r = 255;
        g = 255;
        b = 255;
}

TColor3i::TColor3i ( unsigned char rc, unsigned char gc, unsigned char bc )
{
        r = rc;
        g = gc;
        b = bc;
}

//********************************************************
// TColor4i
//********************************************************

TColor4i::TColor4i ( )
{
        r = 255;
        g = 255;
        b = 255;
        a = 255;
}

TColor4i::TColor4i ( unsigned char rc, unsigned char gc, unsigned char bc, unsigned char ac )
{
        r = rc;
        g = gc;
        b = bc;
        a = ac;
}

//********************************************************
// TColor3f
//********************************************************

TColor3f::TColor3f ( )
{
        r = 1;
        g = 1;
        b = 1;
}

TColor3f::TColor3f ( float rc, float gc, float bc )
{
        r = rc;
        g = gc;
        b = bc;
}

//********************************************************
// TColor4f
//********************************************************

TColor4f::TColor4f ( )
{
        r = 1;
        g = 1;
        b = 1;
        a = 1;
}

TColor4f::TColor4f ( float rc, float gc, float bc, float ac )
{
        r = rc;
        g = gc;
        b = bc;
        a = ac;
}

//********************************************************
// TTriangle3i
//********************************************************

TTriangle3i::TTriangle3i ( )
{
        vertices[0] = -1;
        vertices[1] = -1;
        vertices[2] = -1;
}

TTriangle3i::TTriangle3i ( int v1, int v2, int v3 )
{
        vertices[0] = v1;
        vertices[1] = v2;
        vertices[2] = v3;
}

//******************************************************************************
// TVectorOperations
//******************************************************************************

void TVectorOperations::Rotate_X( double a, double * y, double * z ) // Pitch
{
	double si = mSin(a/180.0*PI);
	double co = mCos(a/180.0*PI);
	double dxx = *y * co - *z * si ;
	double dyy = *y * si + *z * co ;
	*y  = dxx; *z = dyy;
}

void TVectorOperations::Rotate_Y( double a, double * x, double * z ) // Roll
{
	double si = mSin(a/180.0*PI);
	double co = mCos(a/180.0*PI);
	double dxx = *x * co + (*z) * si ;
	double dzz = - (*x) * si + (*z) * co ;
	*x  = dxx; *z = dzz;
}

void TVectorOperations::Rotate_Z( double a, double * x, double * y ) // Heading
{
	double si = mSin(a/180.0*PI);
	double co = mCos(a/180.0*PI);
	double dxx = *x * co - *y * si ;
	double dyy = *x * si + *y * co ;
	*x  = dxx; *y = dyy;
}

double TVectorOperations::AngleBetween2Vectors ( 	double ax, double ay, double az,
                                      			double bx, double by, double bz )
{
	double r1  = ax*ax + ay*ay + az*az;
        if (r1<=0.000000001) return 0;
	r1 = sqrt(r1);			// length of vector 1

	double r2  = bx*bx + by*by + bz*bz;
	if (r2<=0.000000001) return 0;
	r2 = sqrt(r2);			// length of vector 2

	double coss  = (ax*bx + ay*by + az*bz)/(r1*r2);
	if (coss<-1) coss = -1;
	if (coss>+1) coss = +1;

	if ( ax*by - bx*ay >= 0 )		// (z) crossproduct of two vectors
		return (mACos(coss)*180/PI);
                else
        	return 360 - (mACos(coss)*180/PI);
}

void   	TVectorOperations::NormalOfTriangle     ( double x1, double y1, double z1,
                                       double x2, double y2, double z2,
                                       double x3, double y3, double z3,
                                       double &cx, double &cy, double &cz )
{
	double ax,ay,az, bx,by,bz, nx, ny, nz;

        ax = x2-x1;
        ay = y2-y1;
        az = z2-z1;

        bx = x3-x1;
        by = y3-y1;
        bz = z3-z1;

        VectorOp.NormalOf2Vectors ( ax,ay,az, bx,by,bz, nx,ny,nz );
        VectorOp.NormalizeVector  ( nx,ny,nz, cx,cy,cz );
}

void   TVectorOperations::NormalOf2Vectors     ( 	double ax, double ay, double az,
                                      			double bx, double by, double bz,
                                      			double &cx, double &cy, double &cz )
{
        cx = ay*bz - az*by;
        cy = az*bx - ax*bz;
        cz = ax*by - ay*bx;
}

void   TVectorOperations::NormalizeVector      ( double ax, double ay, double az,
                                      		      double &cx, double &cy, double &cz )
{
    double r = ax*ax + ay*ay + az*az;
    if ( r<=0 )
    {
    	cx = 0;
        cy = 0;
        cz = 1;
    	return;
    }
    r = sqrt(r);

    double b = 1/r;
    cx = ax*b;
    cy = ay*b;
    cz = az*b;
}

int TVectorOperations::ClasifyTurn   ( double x1, double y1, double x2, double y2, double x3, double y3 )
{
	double ax = x1 - x2;
	double ay = y1 - y2;
	double bx = x3 - x2;
	double by = y3 - y2;

	double nz = ax*by - ay*bx;
	if ( nz>0 ) return  1;
		else
	if ( nz<0 ) return -1;
		else
		return 0;
}

bool   TVectorOperations::IntersectionOf2Lines (
				double x1, double y1, double x2, double y2,
				double u1, double v1, double u2, double v2,
				double &xi, double &yi )
{
		double dx = (x2-x1);
		double dy = (y2-y1);

		double du = (u2-u1);
		double dv = (v2-v1);

		if (fabs(dx)<0.0000001)	// line1 is vertical
		{
				if (fabs(du)<0.0000001) return false; // both lines vertical thus lines parallel

				xi = x1;
				yi = (xi-u1)*dv/du + v1;
		}
		else
		{
			if (fabs(du)<0.0000001)	// line2 is vertical
			{
				xi = u1;
				yi = (xi-x1)*dy/dx + y1;
			}
			else
			{
				double b1 = dy/dx;
				double b2 = dv/du;

				double a1 = y1-b1*x1;
				double a2 = v1-b2*u1;

				double db = (b1-b2);
				if (fabs(db)<0.0000001) return false;	// lines parallel

				xi = - (a1-a2)/db;
				yi = a1+b1*xi;
			}
		}

		return true;
}

void TVectorOperations::InverseMatrix3x3( double a, double b, double c,
					  double d, double e, double f,
                                          double g, double h, double i,
                            	      	  double *inva, double *invb, double *invc,
                                          double *invd, double *inve, double *invf,
                                          double *invg, double *invh, double *invi)
{
        double determinant =  a*(e*i - f*h) +
                             -b*(d*i - f*g) +
                              c*(d*h - e*g) ;

        double adja =  (e*i - f*h);
        double adjb = -(d*i - f*g);
        double adjc =  (d*h - e*g);
        double adjd = -(b*i - c*h);
        double adje =  (a*i - c*g);
        double adjf = -(a*h - b*g);
        double adjg =  (b*f - c*e);
        double adjh = -(a*f - c*d);
        double adji =  (a*e - b*d);

        double tranadja = adja;
        double tranadjb = adjd;
        double tranadjc = adjg;
        double tranadjd = adjb;
        double tranadje = adje;
        double tranadjf = adjh;
        double tranadjg = adjc;
        double tranadjh = adjf;
        double tranadji = adji;

        (*inva) = tranadja/determinant;
        (*invb) = tranadjb/determinant;
        (*invc) = tranadjc/determinant;
        (*invd) = tranadjd/determinant;
        (*inve) = tranadje/determinant;
        (*invf) = tranadjf/determinant;
        (*invg) = tranadjg/determinant;
        (*invh) = tranadjh/determinant;
        (*invi) = tranadji/determinant;
}

//******************************************************************************
// TPolygonOperations
//******************************************************************************

bool TPolygonOperations::IsPointInsideP ( std::vector<TVertex2d> * pts, double tx, double ty )
{
        int cou = 0;
        double xint;
		int i2;
		int bcount = pts->size();
        for ( int i=0; i<bcount; i++)
              {
				i2 = i+1;
				if ( i2 == bcount ) i2 = 0;

                if ( ((*pts)[i].y!=(*pts)[i2].y) &&
                     ( (((*pts)[i].y  <ty) && ((*pts)[i2].y>=ty)) ||
                       (((*pts)[i2].y<ty) && ((*pts)[i].y  >=ty)) )
                   )
                {
                  if ((*pts)[i2].y==(*pts)[i].y)
                   xint = (*pts)[i2].x;
                   else
                   xint = (*pts)[i2].x -
                        ((*pts)[i2].y-ty)*((*pts)[i2].x-(*pts)[i].x)/((*pts)[i2].y-(*pts)[i].y);
                   if ( xint<=tx ) cou++;
                }
              }
        if (cou % 2 == 1) return true; else return false;
}

bool TPolygonOperations::IsPointNearLine ( double lx1, double ly1, double lx2, double ly2, double lw, double tx, double ty )
{
        double dx,dy,dx1,dy1,dx2,dy2,r,co;
		std::vector<TVertex2d> pts;
        double angle;
		TVertex2d coord;

        if (lx1==lx2 && ly1==ly2)
                return false;

        // calc line angle

        dx = lx2-lx1;
        dy = ly2-ly1;
        r  = dx*dx+dy*dy;
        if (r<0) r = 0;
        r  = sqrt(r);
        co = dx/r;
        if (co> 1) co =  1;
        if (co<-1) co = -1;

        angle = mACos (co) * 180 / PI;
        if (dy<0)
                angle = 360 - angle;

        // find line rectangle

        dx1 =  0;
        dy1 =  lw;
        dx2 =  0;
        dy2 = -lw;

        VectorOp.Rotate_Z ( angle, &dx1, &dy1 );
        VectorOp.Rotate_Z ( angle, &dx2, &dy2 );

        coord.x = lx1 + dx1;
        coord.y = ly1 + dy1;
		pts.push_back(coord);

        coord.x = lx1 + dx2;
        coord.y = ly1 + dy2;
		pts.push_back(coord);

        coord.x = lx2 + dx2;
        coord.y = ly2 + dy2;
		pts.push_back(coord);
        
		coord.x = lx2 + dx1;
        coord.y = ly2 + dy1;
		pts.push_back(coord);

        // test if point in line rectangle

        return IsPointInsideP ( &pts, tx, ty );
}


bool TPolygonOperations::IsCounterClockWise ( std::vector<TVertex2d> & points )
{
	if (points.size()<3) return false;
	double area = 0;
    	unsigned int i,ni;
    	for ( i = 0; i<points.size(); i++ )
    	{
		if (i+1 == points.size())
			ni = 0;
            else
            ni = i+1;
		area += (points[i].x*points[ni].y) - (points[ni].x*points[i].y);
	}
    	return area > 0;
}

void TPolygonOperations::ResizePolygon ( std::vector<TVertex2d> & coords, double size )
{
	double nx, ny, nx1, ny1, nx2, ny2, r1, r2, coss, angle;

	bool ccw = IsCounterClockWise ( coords );

	int bcount = coords.size();
	for ( int i = 0; i<bcount; i++)
	{
		nx1 = coords[i].x - coords[(i+bcount-1)%bcount].x;
		ny1 = coords[i].y - coords[(i+bcount-1)%bcount].y;
		nx2 = coords[i].x - coords[(i+1)%bcount].x;
		ny2 = coords[i].y - coords[(i+1)%bcount].y;

		r1  = nx1*nx1 + ny1*ny1;
		if (r1<0) r1 = 0;
		r1 = sqrt(r1);			// length of vector 1

		r2  = nx2*nx2 + ny2*ny2;
		if (r2<0) r2 = 0;
		r2 = sqrt(r2);			// length of vector 2

		coss  = (nx1*nx2 + ny1*ny2)/(r1*r2);
		if (coss<-1) coss = -1;
		if (coss>+1) coss = +1;
		angle = (mACos(coss)*180/PI);	// angle between two vectors

		if ( nx1*ny2 - nx2*ny1 < 0 ) angle = 360 - angle;

		angle /= 2;

		if (r1>0.000001)
		{
			nx  = size*(nx1/r1);	// resize about 10 m
			ny  = size*(ny1/r1);	// resize about 10 m

			VectorOp.Rotate_Z ( angle, &nx, &ny );

			if (ccw)
			{
				nx = -nx;
				ny = -ny;
			}

			coords[i].x += nx;
			coords[i].y += ny;
		}
	}
}



