#ifndef __GAME_ENVIRONMENT_VECTOR_OPERATIONS__
#define __GAME_ENVIRONMENT_VECTOR_OPERATIONS__

#include "main_functions.h"

#include <stdlib.h>
#include <vector>

//******************************************************************************
// Classes
//******************************************************************************

class TVertex2i;
class TVertex2f;
class TVertex3f;
class TVertex4f;
class TVertex2d;
class TVertex3d;

class TColor3i;
class TColor4i;
class TColor3f;
class TColor4f;

class TTriangle3i;

class TVectorOperations;
class TPolygonOperations;

//******************************************************************************
// Extern Instances
//******************************************************************************

 extern TVectorOperations	VectorOp;
 extern TPolygonOperations 	PolygonOp;

 extern unsigned int mStructColoriToIntColor ( TColor3i &col );
 extern unsigned int mStructColorfToIntColor ( TColor3f &col );
 extern TColor3i     mIntColorToStructColori ( unsigned int &col );
 extern TColor3f     mIntColorToStructColorf ( unsigned int &col );

//******************************************************************************
// TVertex2i
//******************************************************************************

class  TVertex2i
{
public:
	int x;
        int y;
        TVertex2i ( );
        TVertex2i ( int nx, int ny );
};

//******************************************************************************
// TVertex2d
//******************************************************************************

class  TVertex2d
{
public:
	double x;
        double y;
        TVertex2d ( );
        TVertex2d ( double nx, double ny );
};

//******************************************************************************
// TVertex3d
//******************************************************************************

class  TVertex3d
{
public:
	double x;
        double y;
        double z;
        TVertex3d ( );
        TVertex3d ( double nx, double ny, double nz );
};

//******************************************************************************
// TVertex2f
//******************************************************************************

class  TVertex2f
{
public:
	float x;
        float y;
        TVertex2f ( );
        TVertex2f ( float nx, float ny );
};

//******************************************************************************
// TVertex3f
//******************************************************************************

class  TVertex3f
{
public:
	float x;
        float y;
        float z;
        TVertex3f ( );
        TVertex3f ( float nx, float ny, float nz );
};

//******************************************************************************
// TVertex4f
//******************************************************************************

class  TVertex4f
{
public:
	float x;
        float y;
        float z;
        float w;
        TVertex4f ( );
        TVertex4f ( float nx, float ny, float nz, float nw );
};

//********************************************************
// TColor3i
//********************************************************

class  TColor3i
{
public:
        unsigned char r;
        unsigned char g;
        unsigned char b;
        TColor3i ( );
        TColor3i ( unsigned char rc, unsigned char gc, unsigned char bc );
};

//********************************************************
// TColor4i
//********************************************************

class  TColor4i
{
public:
        unsigned char r;
        unsigned char g;
        unsigned char b;
        unsigned char a;
        TColor4i ( );
        TColor4i ( unsigned char rc, unsigned char gc, unsigned char bc, unsigned char ac );
};

//********************************************************
// TColor3f
//********************************************************

class  TColor3f
{
public:
        float r;
        float g;
        float b;
        TColor3f ( );
        TColor3f ( float rc, float gc, float bc );
};

//********************************************************
// TColor4f
//********************************************************

class  TColor4f
{
public:
        float r;
        float g;
        float b;
        float a;
        TColor4f ( );
        TColor4f ( float rc, float gc, float bc, float ac );
};

//********************************************************
// TTriangle3i
//********************************************************

class  TTriangle3i
{
public:
        int vertices[3];
        TTriangle3i ( );
        TTriangle3i ( int v1, int v2, int v3 );
};

//******************************************************************************
// TVectorOperations
//******************************************************************************

class  TVectorOperations
{
public:
        void 	Rotate_X( double a, double * y, double * z );                // Pitch
        void 	Rotate_Y( double a, double * x, double * z );                // Roll
        void 	Rotate_Z( double a, double * x, double * y );                // Heading

        double 	AngleBetween2Vectors ( double ax, double ay, double az,
                                       double bx, double by, double bz );

        void   	NormalOf2Vectors     ( double ax, double ay, double az,	  // for ccw >> a = v1-v0 , b = v2-v0
                                       double bx, double by, double bz,
                                       double &cx, double &cy, double &cz );

        void   	NormalOfTriangle     ( double x1, double y1, double z1,
                                       double x2, double y2, double z2,
                                       double x3, double y3, double z3,
                                       double &cx, double &cy, double &cz );

        void   	NormalizeVector      ( double ax, double ay, double az,
                                       double &cx, double &cy, double &cz );

	int    	ClasifyTurn	    ( double x1, double y1, double x2, double y2,  // 1 = right turn, 0 = parallel, -1 = left turn
        			      double x3, double y3 );

        bool   	IntersectionOf2Lines  ( double x1, double y1, double x2, double y2,
					double u1, double v1, double u2, double v2,
					double &xi, double &yi );

	void 	InverseMatrix3x3     ( double a, double b, double c,
				       double d, double e, double f,
                                       double g, double h, double i,
                            	       double *inva, double *invb, double *invc,
                                       double *invd, double *inve, double *invf,
                                       double *invg, double *invh, double *invi);
};

//******************************************************************************
// TPolygonOperations
//******************************************************************************

class  TPolygonOperations
{
public:
	bool IsCounterClockWise	  ( std::vector<TVertex2d> & points );
	bool IsPointInsideP 	  ( std::vector<TVertex2d> * pts, double tx, double ty );
	bool IsPointNearLine 	  ( double lx1, double ly1, double lx2, double ly2, double lw, double tx, double ty );
        void ResizePolygon	  ( std::vector<TVertex2d> & coords, double size );
};

#endif

